/*
 *  $Revision: 1.18 $
 *
 *  Macro translator
 *
 *  Copyright (c) 1992,1993 by Viewlogic Systems, Inc.
 *
 *
 *  This file contains routines which convert simple 4.x macros to Powerview
 *  format.
 *
 *      - Powerview is put into macro-recording mode
 *      - 4.x macro(s) is(are) parsed line-by-line
 *      - the line is either translated, ignored or added as a comment
 *
 *
 *  NOTE TO THE DEVELOPERS MODIFYING THIS CODE:
 *  The most often called functions, especially the ones called from tight
 *  loops dispense with the niceties of explicit `return' statements.  Also,
 *  if the curly braces are not needed (a function body that can be expressed
 *  as one statement), they are not used.  So, what's the point?  It might
 *  not look like the ViewScript that you are used to, but it's syntactically
 *  correct and just a bit faster than it would be otherwise.
 *
 *
 *  How does it all work?
 *  ---------------------
 *
 *  A short introduction to the workings of the 4.x to Powerview macro
 *  translator.  Feel free to modify it.  If you would like to have your
 *  changes posted on the Viewlogic's bulletin board, please contact Customer
 *  Support hotline for more information.  The entire translator is written in
 *  ViewScript, so improvements to it can be spread almost instantaneously
 *  throughout the customer base.
 *
 *  The translator uses a batch mode of processing to get its job done.  It
 *  puts the Powerview facility from which it was invoked into a macro
 *  recording mode and issues a series of vcMac.AddLine functions calls to
 *  insert appropriate lines into the Powerview macro file.  The translator
 *  can work on a single or many 4.x macro files.  The list of macros to be
 *  translated is displayed in a picklist and, internally, is stored in the
 *  .mlist field of the vcMacTrans facility data structure.
 *
 *  The translator is invoked through the `trans' command which process the
 *  list of macro files.  For each macro file, it goes through each line of it
 *  calling the `process_line' function.  That in turn calls the
 *  `process_table_entry' which determines what to do with the given 4.x macro
 *  line.  Usually, the table entry contains enough information to accomplish
 *  this task.  If a specialized process is required, it will be invoked from
 *  this function.  There is a separate section in this file containing
 *  function declarations with prefixes starting with "process_".  Those are
 *  specialized processors for 4.x functions that require more complex
 *  processing.  One of the obvious ones is the `sys' command, but there are
 *  more.  If you add a special processor function, make sure that the
 *  function accepts the appropriate arguments that will be passed in from the
 *  `process_table_entry' function.  They are:
 *
 *              (entry, x, y, cmd_str, line_no, fp)
 *
 *  `entry'     is the table entry, which is ignored for special processors.
 *  `x', `y'    are the current coordinates represented as two numbers.
 *  `cmd_str'   is the current 4.x macro line with the command name already
 *              removed.
 *  `line_no'   is the current 4.x macro line number.  It's useful when
 *              reporting errors.
 *  `fp'        is the file pointer to the current 4.x macro file.
 *
 *
 *
 *  What are the facility and region names for?
 *  -------------------------------------------
 *
 *  In Powerview every piece of real estate in a window is called a region and
 *  is controlled by a facility.  The control extends to getting the first
 *  crack at input events and processing those for which the facility has
 *  registered appropriate event handlers.  The facility name and region name
 *  are recorded in Powerview macros.  Why?  Because in the same window there
 *  can be many regions controlled by the same facility.  Therefore, both
 *  names are need to uniquely identify in which region the command is to be
 *  replayed.  Anyway, the global variables FacName and RegName are
 *  initialized when the translator is invoked for the first time and are set
 *  to the facility and the region from which the translator was invoked.
 *
 *
 *  What happens when a macro is translated?
 *  ----------------------------------------
 *
 *  The entry in the picklist representing the macro is highlighted.  You can
 *  removed that filename from the picklist by clicking on the `Remove
 *  Selected' button.  Note that this button will remove ALL the highlighted
 *  filenames from the lists.  You can now execute the translated macro using
 *  the Powerview Macro System Panel invoked from the Red Square menu.
 *
 *
 *  Macro filenames.
 *  ----------------
 *
 *  The translated macro is created in the same directory where the 4.x macro
 *  file is located with a new extension, .vmc.  If there is Powerview macro
 *  with the same name in the directory, you will be prompted whether you want
 *  to overwrite the Powerview macro or not.
 *
 *
 *  Why is it a facility?
 *  ---------------------
 *
 *  Because we can have bindings, menus, commands, macro recording and all the
 *  other good things that come to you when you become a facility.  You get
 *  your own package so that your symbols don't clash with anybody else and
 *  you get your own data structure object which can store dynamic
 *  information.  You can become incorporated into other regions and
 *  manipulated at will by other facilities.  You are basically hip and cool.
 *
 *
 *  What doesn't it do?
 *  -------------------
 *
 *  The following situations are not addressed:  window pushing & poping,
 *  drawing commands (bus, net, circle, line, etc), 4.x dialog boxes and
 *  middle button clicks that repeat the last command executed.  The first
 *  instance of the command will be translated, but the subsequent middle
 *  mouse button clicks will not.
 *
 *
 *  What does it mean "Line wasn't translated"?
 *  -------------------------------------------
 *
 *  It means that there was no entry in the translation table for the given
 *  4.x command.  Certain 4.x commands are simply ignored, but if this comment
 *  appears in the Powerview macro, it just means that there is no table
 *  entry.
 *
 */


/*
 * Global variables used by the translator.
 */
declare FacName;            /* Powerview facility name */
declare LineNo;             /* Global 4.x line counter */
declare Origin;             /* Current origin point--vcMacAbsRelVT */
declare RegName;            /* Facility's region name */
declare Table;              /* Tranlator table */
declare Units;              /* Current units--vcMacUserPhysVT */
declare Mouse;              /* Last mouse button hit */


/*
 * Constants which indicate the appropriate action to take while 
 * parsing through a cmd. scenario
 */
IGNORE  := 1;    /* Throw out the line. */
APPEND  := 2;    /* Append the line to the command. */

LEFT    := 1;
MIDDL   := 2;
RIGHT   := 3;


/*
_Create -- facility's _Create method

DESCRIPTION
     The facility's data is a structure storing the main layout table, a list
     of selected macro filenames and the name of the facility from which the
     translator was invoked.  Note that the list of selected macros is stored
     also in the second row, the picklist.

NOTE
     This _Create method does take an argument--an event.  It's needed to get
     a hold of the parent window's region to determine the facility that is
     attemting the translation.

REMINDER
     This function is called by Core whener somebody makes the following call:
         vcRegion.Create ("", vcMacConv.fac", event);
*/
function _Create(event, region)
{
    main_table := vcTable.Create ("vcMacConv", 3, 1);
    vcTable.Row (main_table, -1, vcTable.Border.Visible, 20, 20, 20,
                 vgAlign.None);

    data := struct {t               = main_table;
                    mlist           = nil;
                    parent_region   =
                        vcWin.OwnerRegion (vcEvent.Window (event))};

    /*
     * Whenever the user makes a choice in the file browser, or types a new
     * filename at the command line, the filename will be added to the
     * picklist and to a list of macro filenames.  The mlist serves as a
     * convenient data structure to have the macro filenames readily
     * accessible and always available from the facility's data object.
     */
    function update_picklist(data, event)
    {
        str := vcRegion.Get (vcTable.Get (vcTable.Get (data.t, 0, 0), 0, 1),
                             vlStringVT);

        parts41 := vscript.vsParse (str);
        fname50 := vscript.vsMerge (car (parts41), cadr (parts41),
                                    caddr (parts41), "vmc");
        if (!null (filename := vscript.vsGetini (fname50)))
        {
            function accept(d, e)
            {
                vcPicklist.Add (vcTable.Get (main_table, 1, 0), str, str);
                d.mlist = cons (str, d.mlist);
                vscript.vsDelete (filename, vscript.vsFILE);
            }
            accept_cmd := vcWv.Cmd (fac, accept);
            
            
            function skip(d, e)
                ;
            skip_cmd := vcWv.Cmd (fac, skip);
            
            buttons :=
                vcRegion.Create ("Buttons", vcButnRow.fac,
                                 list (list ("yes",
                    "Delete the existing Powerview macro and add to the list",
                                             accept_cmd,
                                             list (close_window <- TRUE)),
                                       list ("no", "Skip adding this macro",
                                             skip_cmd,
                                             list (close_window <- TRUE))));

            query_t := vcTable.Create ("Delete?", 2, 1);
            strng :=
                sprintf
                 ("The corresponding Powerview macro exists (%s).  Delete it?",
                     filename);
            text := vcRegion.Create ("", vcStaticText.fac, strng);
            vcTable.Add (query_t, text);
            vcTable.Add (query_t, buttons);

            vcWin.Popup (vcEvent.Window (event), query_t, buttons,
                         vcWin.NoMenus + vcWin.NoCmdline + vcWin.NoStatus +
                         vcWin.NoMacros);
        }
        else
        {
            vcPicklist.Add (vcTable.Get (main_table, 1, 0), str, str);
            data.mlist = cons (str, data.mlist);
        }
    }
    add_cmd := vcWv.Cmd (fac, update_picklist);
    vcBind.Set (bindings, '\enter', add_cmd);


    /*
     * Action routine to delete all selected rows in the list of 4.x macros
     * stored in the picklist.  Note that we need to check if the new list is
     * empty, which can happen of all the macros have been selected.
     */
    function delete(data, event)
    {
        index       := 0;
        picklst     := vcTable.Get (data.t, 1, 0);
        new_mlist   := nil;
        temp        := reverse (data.mlist);

        while (!null (temp))
        {
            if (!vcPicklist.IsSet (picklst, index))
                new_mlist = cons (car (temp), new_mlist);

            temp = cdr (temp);
            ++index;
        }

        if (null (data.mlist = new_mlist))
            new_picklst := vcRegion.Create ("pklist", vcPicklist.fac, nil,
                                            columns     <- 1,
                                            exclusive   <- FALSE);
        else
            new_picklst := vcRegion.Create ("pklist", vcPicklist.fac, nil,
                                            rows        <- length (new_mlist),
                                            columns     <- 1,
                                            exclusive   <- FALSE);
            
        temp = reverse (new_mlist);
        while (!null (temp))
        {
            vcPicklist.Add (new_picklst, car (temp), car (temp));
            temp = cdr (temp);
        }

        vcTable.Set (data.t, 1, 0, new_picklst);
    }
    delete_cmd := vcWv.Cmd (fac, delete);


    
    row1  := vcTable.Create ("Browser", 1, 2);
    row11 := vcRegion.Create ("", vcStaticText.fac, "4.x macro to add:");
    row12 := vcRegion.Create ("browser", vcFilePop.fac, "", filter <- "*.mac");

    vcTable.Add (row1, row11);
    vcTable.Add (row1, row12);
    vcTable.Add (main_table, row1);

    row2 := vcRegion.Create ("pklist", vcPicklist.fac, nil, columns <- 1,
                             exclusive <- FALSE);
    vcTable.Add (main_table, row2);

    row3 := 
        vcRegion.Create ("r3", vcButnRow.fac,
                         list (list ("add", "Add to list", add_cmd, nil),
                               list ("trans", "Translate", trans_cmd, nil),
                               list ("delete", "Remove Selected",
                                     delete_cmd, nil),
                               list ("cancel", "Cancel/Close", func();,
                                     list (close_window <- TRUE))));

    vcTable.Add (main_table, row3);
    vcRegion.AddChild (region, main_table);

    data;
}


/****************************************************************************/
/*                      Specialized processor functions                     */
/****************************************************************************/ 


/*
process_left_butt -- left button entry is always select

DESCRIPTION
     Selection in 4.x macros is translated into two Powerview commands.
*/
function process_left_butt(entry, x, y, cmd_str, line_no, fp)
{
    add_to_macro ("select", x, y);
    add_to_macro ("selectExecute", x, y);
    Mouse = LEFT;
}


/*
process_right_butt -- right button entry is always addselected

DESCRIPTION
     Right button selection is translated into two Powerview commands.
*/
function process_right_butt(entry, x, y, cmd_str, line_no, fp)
    if (Mouse == LEFT)
    {
        add_to_macro ("addselect", x, y);
        add_to_macro ("addselectExecute", x, y);
        Mouse = RIGHT;
    }
    

/*
process_sys_line -- process the `sys' command separately

DESCRIPTION
     Parses the sys line and strips out all unsupported options.  Returns a
     valid Powerview sys command line except for the case of plotter, where it
     returns the plotter command which, when run, will create the metafile and
     spawn the Powerview plotter executable (splplt--which both creates and
     tries to spool the plot file).

NOTE
     The substitution for the $PROJECT variable uses a syntactic escape into
     ViewScript from the command line.  The command line editor interprets
     anything enclosed between $( ) braces as an expression to be evaluated by
     the ViewScript interpreter.  The new environment variable for PROJECT is
     WDESIGN.
*/
function process_sys_line(entry, x, y, cmd_str, line_no, fp)
{
    new_str := "sys ";
    re      := vscript.re_compile (".*([$]project)([^ ]*)");

    foreach (token; string_loop (cmd_str, " "))
    {
        /*
         * Retain only supported sys options.  Note that the options can be
         * prefixed with either a hyphen or a slash.
         */
        if ((token [0] == '-') || (token [0] == '/'))
        {
            /*
             * Strip off the hyphen or the slash.  Note that the hyphen will
             * be added to the new string per Powerview requirements.
             */
            token = substring (token, 1, 0);
            if (string_eqi (token, "p")         ||
                string_eqi (token, "pause")     ||
                string_eqi (token, "nopause")   ||
                string_eqi (token, "nop")       ||
                string_eqi (token, "wait")      ||
                string_eqi (token, "w")         ||
                string_eqi (token, "nowait")    ||
                string_eqi (token, "now"))
                new_str = new_str + "-" + token + " ";
        }
        else
        {
            /*
             * Special case for plotter--substitute the plotgodefs command.
             */
            if (string_eqi (token, "plotter"))
            {
                new_str = "plotgodefs";
                return (nil);   /* This only exits the `foreach' loop. */
            }
            /*
             * Now use the regular expressions to see if a $project string is
             * not embedded somewhere in the token string.  Note that the job
             * is not done after a match if found--we need to determine
             * whether there were any other characters immediately following
             * the $project string, which is a match #1.  The extra characters
             * will be in match #2.  If there are any, they need to be
             * appended to the $WDESIGN variable.  Otherwise, a space needs
             * to be inserted.
             */
            else if (vscript.re_exec (re, token) == TRUE)
            {
                new_str = new_str + "$WDESIGN";
                if ((rest_of_token := vscript.re_getmatch (re, 2)) != nil)
                    new_str = new_str + rest_of_token;
                else
                    new_str = new_str + " ";
            }
            else
                new_str = new_str + token + " ";
        }
    }

    add_to_macro (new_str, x, y);
}


/*
process_mex -- process the `mex' command separately

DESCRIPTION
     The main reason for custom processing of the `mex' command is to strip
     off the .mac extension from the filename argument.
*/
function process_mex(entry, x, y, cmd_str, line_no, fp)
{
    declare filename;

    /*
     * See if the filename argument has been specified on the same line.
     */
    if (null (string_token (cmd_str, &filename, " ")))
    {
        cmd_str  = getline (fp);
        filename = caddr (get_coords (cmd_str));
    }

    file_parts := vscript.vsParse (filename);
    filename    = vscript.vsMerge (car (file_parts), cadr (file_parts),
                                   caddr (file_parts), "");

    add_to_macro ("mex " + filename, x, y);
}


/*
process_sgroup -- process the `sgroup' command

DESCRIPTION
     This function would be exactly the same as the process_left_butt
     function, except the need to get the starting and ending coordinates for
     the `select' and `selectExecute' commands.  Think of a rectangle that
     needs to be stretched in order to cover the selected objects--it needs
     two sets of coordinates to be placed on a schematic.
*/
function process_sgroup(entry, x, y, cmd_str, line_no, fp)
{
    if ((line1 := get_coords (getline (fp))) == nil)
        vlEx.Throw2 (vlEx.FAILURE, "vcMacTrans", vlErrSev.Error,
                     sprintf ("Unexpected EOF while parsing line %d", LineNo));

    if ((line2 := get_coords (getline (fp))) == nil)
        vlEx.Throw2 (vlEx.FAILURE, "vcMacTrans", vlErrSev.Error,
                     sprintf ("Unexpected EOF while parsing line %d", LineNo));

    add_to_macro ("select", car (line1), cadr (line1));
    add_to_macro ("selectExecute", car (line2), cadr (line2));

}

/****************************************************************************/
/****************************************************************************/


/*
 * Powerview scan codes for function keys, buttons, and cntrl, shift, alt
 * combinations of same.
 */
MIN_FKEY_VAL := 256;

LBUT_VAL    := 291;
MBUT_VAL    := 293;
RBUT_VAL    := 295;

ENTER_VAL   := 307;
SKEY_DIFF   := 64;
CKEY_DIFF   := 128;
AKEY_DIFF   := 512;


/*
is_func_key -- is it a function key?

DESCRIPTION
     Checks to see if the given string represents a 4.x function key.  If so,
     returns the Powerview scancode, otherwise nil.  Note that the F0 key is mapped
     to F10.

RETURNS
     A boolean.
*/
function is_func_key(str)
    if (str [0] == '[')
    {
        c := str [1];
        if ((c >= '1') && (c <= '9'))
            MIN_FKEY_VAL + (char2int (c) - 49);
        else if (c == '0')
            MIN_FKEY_VAL + 9;
    }
    else
        FALSE;


/*
is_button -- is it a button key?

DESCRIPTION
     Checks to see if the given string represents a 4.x button hit.  If so,
     returns the Powerview scancode, otherwise FALSE.
*/
function is_button(str)
    if (str [0] == ']')
    {
        c := str [1];

        if (c == '4')
            LBUT_VAL;
        else if (c == '2')
            MBUT_VAL;
        else if (c == '1')
            RBUT_VAL;
    }
    else
        FALSE;


/*
is_key -- is it a keystroke?

DESCRIPTION
     Determines if string represents a 4.x keystroke and returns the actual
     keycode if it does, FALSE otherwise.
 */
function is_key(str)
    if ((str_len := string_length (str)) < 2)
        FALSE;

    /* Plain function key? */
    else if ((key := is_func_key (str)) != FALSE)
        key;

    /* Plain button? */
    else if ((key = is_button (str)) != FALSE)
        key;

    /* Shift, Cntrl, or Alt combo of function or button? */
    else
    {
        if (str_len < 4)
            return (FALSE);

        if (string_eq (str, "enter"))
            return (ENTER_VAL);

        /* Strip off possible indication of Shift, Cntrl, or Alt */
        key_code := substring (str, 0, 2);

        /* Save the rest of the line */
        new_str := substring (str, 2, str_len);

        /* Cntrl-function key or button? */
        if ((string_eq (key_code, "c-")) &&
            (((key = is_func_key (new_str)) != nil) ||
             ((key = is_button (new_str)) != nil)))
            key + CKEY_DIFF;

        /* Shift-function key or button? */
        else if ((string_eq (key_code, "s-")) &&
                 (((key = is_func_key (new_str)) != nil) ||
                  ((key = is_button (new_str)) != nil)))
            key + SKEY_DIFF;

        /* Alt-function key or button? */
        else if ((string_eq (key_code, "a-")) &&
                 (((key = is_func_key (new_str)) != nil) ||
                  ((key = is_button (new_str)) != nil)))
            key + AKEY_DIFF;
        else
            FALSE;
    }


/*
valid_string -- should we bother translating?

DESCRIPTION
     Check for comment character (|), blank line or space.
*/
function valid_string(str)
    ((str [0] != '|') && (str [0] != ' ') && (str [0] != '\n'));


/*
strip_cr -- replace carriage returns

DESCRIPTION
     Checks to see if the string ends in a carriage return, and if so,
     replaces that character with the '\0' character.
*/
function strip_cr(str)
{
    last_char := string_length (str) - 1;

    if (string_ref (str, last_char) == '\n')
        string_set (str, last_char, '\0');

    str;
}


/*
getline -- read the next line from the 4.x macro

DESCRIPTION
     Aside from just reading the next line, it increments the global line
     counter, LineNo, and strips the carriage return character.  The counter
     is reset in the trans function.
*/
function getline(fp)
    if ((line := fgetline (fp)) != nil)
    {
        ++LineNo;
        strip_cr (line);
    }
    else
        nil;


/*
add_to_macro -- add a line to the Powerview macro

DESCRIPTION
     Depending on the current settings of the Units and Origin variables, an
     appropriate replay function will be chosen.
*/
function add_to_macro(command, x, y)
    vcMac.AddLine (FacName, RegName, command, x, y, '0', "", "vcMac.Win");


/*
get_coords -- process coordinates

DESCRIPTION
     Strips off the coordinates into a separate string and sets the current
     Origin and Units variables accordingly.

NOTE
     The 'v' and 'r' characters can follow one another and thus the need for
     two if statements.

RETURNS
     A three element list: x, y coordinates and the rest of the macro line.
*/
function get_coords(line)
{
    if (line [0] == 'v')
    {
        line  = substring (line, 2, string_length (line));
        if (Units != vcMacUserPhys.User)
            vcMac.ChangeUnits (Units = vcMacUserPhys.User);
    }
    
    if (line [0] == 'r')
    {
        line   = substring (line, 2, string_length (line)); /* strip away */
        if (Origin != vcMacAbsRel.Relative)
            vcMac.ChangeOrigin (Origin = vcMacAbsRel.Relative);
    }

    if ((i := lindex (line, ' ')) == FALSE)
        vlEx.Throw2 (vlEx.FAILURE, "vcMacTrans", vlErrSev.Error,
                     sprintf ("Line: %d, Coordinate format error.", LineNo));
    x := string2number (substring (line, 0, i));

    
    line = substring (line, i + 1, string_length (line));
    if ((i = lindex (line, ' ')) == FALSE)
        vlEx.Throw2 (vlEx.FAILURE, "vcMacTrans", vlErrSev.Error,
                     sprintf ("Line: %d, Coordinate format error.", LineNo));
    y := string2number (substring (line, 0, i));

    /*
     * Get command line string.  Ignore comments, if any.
     */
    if ((end := lindex (line, '|')) == FALSE)
        line = substring (line, i + 1, string_length (line));
    else        
        line = substring (line, i + 1, end);

    list (x, y, line);
}


/*
process_table_entry -- what to do with the Table entry?

DESCRIPTION
     At this point a 4.x command has an entry in the Table and we need to
     process it.  A key point is that at times only the current line will
     suffice and at times more lines will have to be read in from the 4.x
     macro file.
*/
function process_table_entry(entry, cmd_str, x, y, line_no, fp)
{
    arg_str      := "";
    args_last    := FALSE;
    opt_args_str := "";

    /*
     * First comes the new command name.  If it's nil, ignore it.
     */
    new_cmd := car (entry);     entry = cdr (entry);

    if (null (new_cmd))
        return (nil);

    /*
     * Second comes the number of required arguments.  If it's -1, it means
     * that there is a dedicated translation routine for it stored in the
     * third token of the table entry.
     */
    nargs := car (entry);   entry = cdr (entry);

    if (nargs == -1)
    {
        (car (entry)) (entry, x, y, cmd_str, line_no, fp);
        return (nil);
    }
    else
    {
        /*
         * Read in all the required 4.x arguments.  Remember that the
         * arguments can be arranged arbitrarily either all on one line or
         * each argument separately or any combination of the previous two.
         * Any key stroke found while looking for the arguments is discarded.
         */
        count := 0;
        while (count < nargs)
        {
            declare arg;

            /*
             * Note that we cannot use the string_loop iterator here, because
             * the value of the cmd_str can change if we find a key entry.
             */
            while (((cmd_str = string_token (cmd_str, &arg, " ")) != nil) &&
                   (count < nargs))
                if (!(key := is_key (arg)))
                {
                    arg_str = arg_str + " " + arg;
                    ++count;
                }
                else if (key == ENTER_VAL)
                    ++count;

            /*
             * The required arguments might be spread on many lines.  Get a
             * new one and reset the x, y coordinates.
             */
            if (count < nargs)
            {
                cmd_str = get_coords (getline (fp));
                x = car (cmd_str);
                y = cadr (cmd_str);
                cmd_str = caddr (cmd_str);
            }
        }

        /*
         * A bit of trickery here to check for file redirection.  The commands
         * that allow file redirection are defined to require 0 arguments,
         * because if there is no redirection, the text is displayed in a
         * Viewtext window.  The to this boolean expression is the fact that
         * ViewScript will stop the evaluation as soon as any of the component
         * expressions returns boolean FALSE.  So, first check if the argument
         * cound is still 0.  Second, see if there is any token left on the
         * current command line (redirection is always specified on one line
         * in 4.x macros).  Third, if we are still here, check if that token
         * is a less-than sign.  If it is, append the token and the rest of
         * the command line to the arg_str variable.
         */
        if ((count == 0)                                            &&
            (!null ((cmd_str = string_token (cmd_str, &arg, " ")))) &&
            (string_eq (arg, ">")))
            arg_str = " " + arg + cmd_str;
    }

    /*
     * The next token in the table entry are the optional arguments.  If the
     * list of the optional arguments is not empty, check if the first element
     * of the list is -1.  If it is, that means that the arguments that were
     * processed above, should come *after* the optional arguments have been
     * appended to the new_cmd string.
     */
    opt_args := car (entry);    entry = cdr (entry);

    if (!null (opt_args) && (car (opt_args) == -1))
    {
        opt_args  = cdr (opt_args);
        args_last = TRUE;
    }

    while (!null (opt_args))
    {
        opt_args_str = opt_args_str + " " + car (opt_args);
        opt_args     = cdr (opt_args);
    }

    /*
     * Now we are ready to issue the command, finally.
     */
    if (args_last)
        new_cmd = new_cmd + opt_args_str + arg_str;
    else
        new_cmd = new_cmd + arg_str + opt_args_str;

    add_to_macro (new_cmd, x, y);


    /*
     * We are passed the required 4.x arguments, time to process the optional
     * lines that might follow.
     */
    line_action := car (entry);    entry = cdr (entry);

    while (!null (line_action))
    {
        if ((action := car (line_action)) == APPEND)
            new_cmd = new_cmd + " " + caddr (get_coords (getline (fp)));
        else if (action != IGNORE)
            add_to_macro (action, x, y);
        else
            getline (fp);
        line_action = cdr (line_action);
    }
}
        

/*
process_line -- attempt to translate a single line

DESCRIPTION
     Takes a line from a 4.x macro and a line number and checks if there is an
     translation entry for this command.  If there isn't one, a comment is
     inserted indicating that the line is not translatable.  The processing of
     the table entry is handled by the process_table_entry function.

NOTE 1
     The call to the get_coords function can reset the Origin and Units
     variables, even though the line might be found not tranlatable.  This is
     a necessary nuissance without which we wouldn't have an easy access to
     the actual command name token.  Since each 4.x macro line contains
     information on which coordinate system to use, the resetting doesn't
     cause any problems.

NOTE 2
     The cmd string that is passed to process_table_entry has the command name
     stripped.  It's not needed to process the table entry.
*/
function process_line(line, line_no, fp)
{
    declare cmd_name, entry;

    c_and_c := get_coords (line);
    x       := car (c_and_c);
    y       := cadr (c_and_c);
    cmd     := caddr (c_and_c);

    if (null (cmd = string_token (cmd, &cmd_name, " ")))
        nil;
    else if (!null (entry = vlSym.Fetch (Table, cmd_name)))
        process_table_entry (entry, cmd, x, y, line_no, fp);
    else
    {
        str := sprintf ("Line #%d was not translated: %s", line_no, line);
        vcMac.InsertComment (str);
    }
}


/*
trans -- action routine for the "trans" command

DESCRIPTION
     The main engine of the translator.  Note that when this function is
     called, no checking for possible overwriting of existing Powerview macro
     files takes place.  That is handled by the update_picklist function
     defined in the _Create method.  The mlist stored in the facility's data
     object has the user's seal of approval.  WHY?  Because if we relied only
     on the vcMac.Start command, we would have no means of waiting for the
     vcMac.Start function to tell us that everything is copacetic.
     vcMac.Start returns as soon as it's done, even if it has to put up a
     dialog box asking the user for overwrite permission.  The user's response
     can come many seconds later and we have no way of knowing how long to
     wait or what was the response unless *we* put put a dialog box ourselves.
     That's what the update_picklist function does.
*/
function trans(data, event)
{
    declare line;
    declare fname41, parts41, fname50;
    declare fp;

    /*
     * First, check if the table has been loaded.  There is no need to load
     * the stuff until it actually is needed.
     */
    if (null (Table))
        load ("vcMacTable.vs");

    /*
     * Initialize the names of the facility and the region for future use.
     */
    FacName = vcFac.Name (vcRegion.Fac (data.parent_region));
    RegName = vcRegion.Name (data.parent_region);

    /*
     * Reverse the list of 4.x filenames to start processing them in the order
     * they were entered.
     */
    data.mlist = reverse (data.mlist);

    macro_count := 0;
    vcWin.Busy (vcEvent.Window (event), TRUE);
    unwind_protect
        for (tlist := data.mlist; !null (tlist); tlist = cdr (tlist))
        {
            fname41 = car (tlist);
            parts41 = vscript.vsParse (fname41);

            fname50 := vscript.vsMerge (car (parts41), cadr (parts41),
                                        caddr (parts41), "vmc");
            /*
             * Initialize the current origin and unit values.
             */
            Origin = vcMacAbsRel.Absolute;
            Units  = vcMacUserPhys.Physical;

            vcMac.Start (fname50, event, Units, Origin);

            /*
             * Check for extension on the 4.x file name.  If no extension add
             * the old .mac extension.
             */
            if (string_eq (cadddr (parts41), ""))
                fname41 = fname41 + ".mac";

            fp = fopen (fname41, "r");

            /*
             * Process each 4.x macro file line by line.  Note the counter
             * which can be useful in reporting problems.
             */
            unwind_protect
            {
                LineNo = 0;
                while ((line = fgetline (fp)) != nil)
                {
                    ++LineNo;
                    if (valid_string (line))
                        process_line (strip_cr (line), LineNo, fp);
                }
            }
            cleanup
            {
                fp = fclose(fp);
                vcMac.Close (vcEvent.Window (event));
                vcPicklist.SetPick (vcTable.Get (data.t, 1, 0), macro_count);
                ++macro_count;
            }
        }
    cleanup
        vcWin.Busy (vcEvent.Window (event), FALSE);
}


trans_cmd := vcWv.Cmd (fac, trans);
