/*
 *  $Revision: 1.93 $
 *
 *  ViewScript functions for command object creation and binding.
 *  Central location for Core utility functions.
 *
 *  Copyright (c) 1992 by Viewlogic Systems, Inc.
 */

gc_status (FALSE);


/*\vscmuserDocpage vcWv.Printable -- iterate over all printable characters
 * 
 * DESCRIPTION
 *      This is an iterator over all printable ASCI characters, from space to
 *      the tilde character ('~').
 * 
 * EXAMPLE
 *      This example is taken from Viewtext.  It will bind all the
 *      printable ASCII characters to the _InsertChar callback,
 *      which inserts the character into the current buffer.
 * 
 *          tmp_cmd := vcWv.Cmd (fac, _InsertChar, "insert_char");
 *          foreach (i; vcWv.Printable ())
 *              vcBind.Set (bindings, i, tmp_cmd);
 * 
 * SEE ALSO
 *      vcWv.AllChars, vscript.map, vscript.list_iter
 */
iterator Printable()
{
    for (i := char2int (' '); i <= char2int ('~'); ++i)
        iterate (int2char (i));
}
printable := Printable;     /* For compatibility. */


/*\vscmuserDocpage vcWv.AllChars -- iterate over all characters
 * 
 * DESCRIPTION
 *      This is an interator over all characters in the Viewlogic
 *      character set, including mouse buttons, mouse movement,
 *      function keys, and the printable characters.
 * 
 * SEE ALSO
 *      vcWv.Printable
 */
iterator AllChars()
{
    for (i := 0; i <= 1023; ++i)
        iterate (int2char (i));
}


Secondary := `secondary;

/*\vscmuserFunction vcWv.StringCmd -- Create a command object from a string
 * 
 * DESCRIPTION
 *      Given a command-line string, this function creates a command
 *      object that can be passed to a button or gesture, etc.  Bindings
 *      and menus can take strings directly, but this allows strings
 *      to be used elsewhere when commands are expected.
 *
 *      This command is given no name -- it will not be macro recordable.
 *      However, the command that it runs may be.
 */
function StringCmd(fac, str) {
    cmd := vcWv.Cmd (fac, func(data, event)
                     vcWin.RunCmdline (vcEvent.Window (event), str, event););
    vcCmd.SetHelp (cmd, str);
    cmd;
} /* function StringCmd */

/*\vscmuserDocpage vcWv.Cmd -- create a command object
 * 
 * SYNOPSIS
 *      vcWv.Cmd(facility, action_func [, name] [, dbox_title] [, param args])
 *          vcFacilityVT    facility;
 *          closureVT       action_func;    // ViewScript-accessible function
 *         [vlStringVT      name;]
 *         [vlStringVT      dbox_title;]
 *         [consVT          param args;]    // variable length list of arg
 *                                          // descriptors: vcWv.Param(man)
 * DESCRIPTION
 *      Commands provide the base data structure for the user
 *      interface.  User-generated events (keystrokes, menus,
 *      command lines) are translated by Core into commands, and are
 *      then executed.  A command object contains an action function
 *      and information about its arguments, including prompts,
 *      defaulting strategies, and types.
 * 
 *      The vcWv.Cmd function takes a variable number of arguments,
 *      of which the first two are required: the facility and an
 *      action function.  The first argument to vcWv.Cmd is the
 *      facility whose data object is to be passed as the first
 *      argument to the action function.  This is the facility that
 *      is "in charge" of this command.  That is, if you are
 *      defining a new command object which uses a Viewtrace
 *      callback as an action function, you have to specify
 *      vtrace.fac as the facility.
 * 
 *      The second argument is the action function.  This can be a
 *      ViewScript function or a C function made accessible to
 *      ViewScript.  The action function must take at least two
 *      arguments.  The first argument is the facility data object,
 *      whose type is determined by the facility in which the
 *      command is defined.  The second argument is an event object,
 *      describing the mouse location and the keystroke that
 *      triggered the command.  Both of these arguments will be
 *      provided by Core.
 * 
 *      The third argument (optional) is the command name.  Core has
 *      various uses for the command name.  It is used with any
 *      error or informational messages associated with running the
 *      command as well as with the binding display.  The name is
 *      used also to automatically register the command in the
 *      facility's command table `commands'.  Such registered
 *      commands can be invoked from the command line editor.  They
 *      also can be recorded in a macro.  Note that there is no way
 *      to macro record a command which doesn't have a name -- how do
 *      we find it?  If the action function is a C callback, then the
 *      name can be specified as nil (or omitted entirely if this is
 *      the last argument) and the command name will be taken from
 *      the callback name.
 * 
 *      The fourth argument (optional) is the title of the dialog
 *      box that Core will create if this call has any parameter
 *      descriptors.
 * 
 *      After the four arguments comes an optional and arbitrary
 *      number of parameter descriptors.  These are descriptors of
 *      the additional arguments required by the action function.
 *      Remember that an action function receives two arguments from
 *      Core by default: the facility data structure pointer and the
 *      event pointer.  Any additional arguments have to be
 *      described in the parameter descriptors.  The parameter
 *      descriptors are obtained by calling the function vcWv.Param.
 *      See that function's documentation for further details.
 * 
 * NOTE
 *      If you need to specify a command name (the third argument),
 *      but don't want the command to be recorded in macros, use the
 *      vcCmd.EnableMacro function.  This function will direct Core
 *      not to record the command when the macro recording is on.
 *      The command is NOT disabled--it's executed but not recorded
 *      in the macro file.
 * 
 * EXAMPLES
 *      The first example shows how to use the vcWv.Cmd command to
 *      create a command Id for a simple callback, which doesn't
 *      require any qualifiers or descriptors:
 * 
 *           vcWv.Cmd (fac, your_callback, "typed_command_name");
 * 
 *      The second example shows the use of the vcWv.Cmd with a
 *      callback which requires an additional argument.  The type of
 *      that argument is a string, its default value is an empty
 *      string, and it should be a sticky default, i.e. the argument
 *      should default to the last entered value:
 * 
 *           vcWv.Cmd (fac,
 *                     your_callback,
 *                     "name",
 *                     "Dialog Box Title"
 *                     vcWv.Param ("New argument", vlStringVT,
 *                                 vcWv.Sticky, ""));
 * 
 *      The third example shows the usage of a ViewScript action function
 *      together with the dynamic defaulting strategy.  Note that the action
 *      function takes two extra arguments, a name and a mode.  The command is
 *      also declared not macro recordable:
 * 
 *          function print_name(data, event, name, mode)
 *          {
 *              win := vcEvent.Window (event);
 *              vcWin.ClearStatus (win);
 *              vcWin.AddStatus (win, sprintf ("Name: %s, mode: %d",
 *                                             name, mode));
 *          }
 * 
 *          function get_name(data)
 *              "Test one";
 * 
 *          cmd := vcWv.Cmd (fac,
 *                           print_name,
 *                           "pname",
 *                           "Print Test Name",
 *                           vcWv.Param ("Test name:", vlStringVT,
 *                                       vcWv.Dynamic, get_name),
 *                           vcWv.Param ("Mode:", vlIntVT,
 *                                       vcWv.Sticky, 0));
 * 
 *          vcCmd.EnableMacro (cmd, vlOnOff.Off);
 * 
 * SEE ALSO
 *      vcWv.Param, vcCmd.EnableMacro
 * 
 * RETURNS
 *      The command object.
 */
function Cmd(args...) {
    fac   := car (args);  args = cdr (args);
    fun   := car (args);  args = cdr (args);
    name  := car (args);  args = cdr (args);
    title := car (args);  args = cdr (args);
    vcCmd.Create (fac, fun, name, title, args);
}


/*\vscmuserDocpage vcWv.Param -- declare a parameter for a command
 * 
 * SYNOPSIS
 *      vcWv.Param(prompt, arg_type, default_strategy, default_value
 *                 [, vcWv.Secondary] [, facility] [, help] [, flags....]);
 *          vlStringVT      prompt;
 *          vlTypesepcVT    arg_type;
 *          vcCmd.Strategy  default_strategy;
 *          vscmVT          default_value;
 *         [symbolVT        vcWv.Secondary;]
 *         [vcFacilityVT    facility;]
 *         [vlStringVT      help;]
 *         [consVT          flags;]
 * 
 * DESCRIPTION
 *      Used in conjunction with vcWv.Cmd to define a parameter for
 *      a command.  It takes an arbitrary number of arguments, but
 *      the first four are required.
 * 
 *      The first argument is the prompt to be used in the dialog
 *      box to enter this argument -- it has to be a string.
 * 
 *      The second argument is the value type for the argument.  You
 *      must tell Core what type of data to expect, so that it can
 *      perform the necessary type checking.
 * 
 *      The third argument is the default strategy, which can be any
 *      of the following values: vcWv.Constant, vcWv.Sticky, or
 *      vcWv.Dynamic.
 * 
 *      The fourth argument is the default value, which is dependent
 *      upon the default strategy:
 * 
 *          Dflt Strategy  Explanation                        Dflt Value
 *          -------------  ---------------------------------  -------------
 *          vcWv.Constant  The same default is always used    The value
 *          vcWv.Sticky    Default value held over from       Initial value
 *                         last invocation of command
 *          vcWv.Dynamic   Function is called each time to    Function
 *                         get default value
 * 
 *      A function used for the vcWv.Dynamic default strategy must
 *      take a single argument, the data object for the facility
 *      that owns the command, and return a value that is the same
 *      type as is expected by the parameter (as specified by the
 *      second argument to the vcWv.Param function).
 * 
 *      Additional arguments may be added to qualify the behavior of
 *      the dialog box used to gather the information.
 *      vcWv.Secondary means that the argument is of "secondary"
 *      importance -- many users will not need to change it.  It
 *      causes the prompt to be displayed in italics, and makes
 *      carriage-return skip over that entry.  If a dialog box
 *      contains a "primary" (non-secondary) parameter, followed by
 *      a series of secondary parameters, hitting carriage return
 *      after entering a value for the first parameter will dismiss
 *      the dialog box and run the command with the default values
 *      for the secondary parameters.
 * 
 *      Normally, the system automatically picks a facility to edit
 *      the parameter's value, based on the type of the parameter.
 *      However, you can override the choice by specifying a
 *      facility as the next argument to vcWv.Param.
 * 
 *      You can provide a help-string to explain a parameter in more
 *      detail.  When that parameter's editor has focus (i.e. the
 *      user is editing it now), the help string will be displayed
 *      in the status area of the dialog box.
 * 
 *      Further flags can be given to modify the behavior of the
 *      facility used to edit the parameter; any number can be
 *      specified using the form 'pref_name <- value'.  See the
 *      facility's documentation for a list of available
 *      preferences.
 * 
 * EXAMPLES
 *      This is a simple example with a dynamic callback that calls
 *      the builtin gtest function "name", which takes a gtest
 *      object and returns a string that is the name of the drawing
 *      being edited.  This is then used as the default value for
 *      the name of the new drawing to open.
 * 
 *          vcWv.Cmd (fac, open, "Open",
 *                    vcWv.Param ("Drawing", string,
 *                                vcWv.Dynamic, name));
 * 
 *      Here's an example of a command that takes two arguments of
 *      type vlYesNoVT, which can accept only a value of "Yes" or
 *      "No".  The system will, by default, use a pick-from-list
 *      editor to edit the value of those parameters.  To illustrate
 *      the use of an optional facility argument, to the parameter
 *      descriptor, we will override the editor selection to use the
 *      command line editor for the second parameter.
 * 
 *          vcWv.Cmd (fac, func(d,e,p1,p2);, "Test",
 *                    vcWv.Param ("Param 1", vlYesNoVT,
 *                                vcWv.Sticky, "Yes"),
 *                    vcWv.Param ("Param 2", vlYesNoVT,
 *                                vcWv.Sticky, "No",
 *                                vcCmdline.fac));
 * 
 * RETURNS
 *      A parameter descriptor for use by vcWv.Cmd.
 */
param := list;
Param := list;

Dynamic  := vcCmd.Strategy.Dynamic;
Constant := vcCmd.Strategy.Constant;
Sticky   := vcCmd.Strategy.Sticky;


/*\vscmuserDocpage vcWv.Interrupts -- set up keys as facility interrupts
 * 
 * SYNOPSIS
 *      vcWv.Interrupts(facility, var_args)
 *          vcFacilityVT    facility;
 *          // The var_args is a variable number of characters representing
 *          // the keys that are allowed to interrupt the given facility.
 *
 * DESCRIPTION
 *      Some facilities allow the user to interrupt them.  This
 *      function defines which keys are used to interrupt a
 *      particular facility.  Each facility responds (or ignores)
 *      interrupts in its own way.  They must actively check for
 *      interrupts from C by calling vcWinInterrupt.  From ViewScript, the
 *      call is vcWin.Interrupt.
 *
 * EXAMPLE
 *      vcWv.Interrupts (fac, '\leftup', '\rightup');
 *
 * SEE ALSO
 *      vcWin.Interrupt
 */
function Interrupts(args...) {
    if (null (cdr (args)))
        vlEx.Throw2 (vlEx.FAILURE, "vcWv", vlErrSev.Error,
                     "vcWv.Interrupts requires a facility as the first argument");
    fac := car (args);

    for (args = cdr (args); args != nil; args = cdr (args))
        vcFac.AddInterrupt (fac, car (args));
}
interrupts := Interrupts; /* For compatibility. */

/*
 * Help function for menus.
 */
function color(c) cons (`color, c);
function font(c) cons (`font, c);


default_menu_color := `cyan;
default_menu_font  := 6;


/*\vscmuserDocpage vcWv.MakeMenu -- create a menu object
 * 
 * SYNOPSIS
 *      vcMenuVT vcWv.MakeMenu(name, list_of_entries)
 *          vlStringVT name;
 *          vscmVT     list_of_entries;
 * 
 * DESCRIPTION
 *      This function creates a menu.  The first argument to this
 *      function is a string.  This is the menu name.  Then follows
 *      an arbitrary number of arguments, which define the menu
 *      entries.  Each entry is described as a list, whose first
 *      element is a visual and the second is an action entry.
 * 
 *      The visual can be either a string or a function returning a
 *      region and will be displayed as is during the menu
 *      selection.  The region can be an icon region, for example,
 *      to display a bitmap image as a visual.  See the man page for
 *      the vcWv.Icon function which is provided as a convenience to
 *      include bitmap images in menus.
 * 
 *      The action can be either of the three objects:
 *          - a command object, as generated by the vcWv.Cmd function
 *          - a command line string (with or without arguments)
 *          - another menu to create a sub-menu
 * 
 * NOTE
 *      The menu entries can be combined in a list, to create a
 *      two-dimensional menu.  The entry numbering starts at zero
 *      and goes left to right.
 * 
 * EXAMPLE
 *      // This example will create two menus: a Window and a File
 *      // menu, to be displayed as sub-menus from the main window
 *      // menu.  Function cmd provides a simpler interface to the
 *      // vcWv.Cmd command, which creates a command id.
 *      //
 *      function cmd(args...)
 *          apply (vcWv.Cmd, cons (fac, args));
 * 
 *      window_menu := vcWv.MakeMenu ("window",
 *                                    list ("Refresh", cmd (refall)),
 *                                    list ("Close",   vcWin.wclose));
 *      file_menu   := vcWv.MakeMenu ("file",
 *                                    list ("Write",   cmd (write_file)),
 *                                    list ("WriteTo", writeto_cmd));
 *      main_menu   := vcWv.MakeMenu ("vtext",
 *                                    list ("Window",  window_menu),
 *                                    list ("File",    file_menu));
 *      vcFac.SetMenus (fac, main_menu);
 * 
 * 
 *      // The file and window menus can be combined into a 2-D menu:
 *      //
 *      two_d_menu := vcWv.MakeMenu ("two-d",
 *                                   list (list ("Refresh", cmd (refall)),
 *                                         list ("Close",   vcWin.wclose)),
 *                                   list (list ("Write",   cmd (write_file)),
 *                                         list ("Writeto", writeto_cmd)));
 *      vcFac.SetMenus (fac, two_d_menu);
 * 
 * RETURNS
 *      A menu object.
 * 
 * SEE ALSO
 *      vcWv.Icon, vcWv.Separator, vcWv.Pin,
 *      vcMenu.AppendEntry, vcMenu.ReplaceEntry
 */
function MakeMenu(entries...)
    vcMenu.Create (car (entries), cdr (entries));
make_menu := MakeMenu;


/*\vscmuserDocpage vcWv.Pin -- make a pin object
 * 
 * SYNOPSIS
 *      vscmVT vcWv.Pin()
 * 
 * DESCRIPTION
 *      This function is enabled only for the OpenLook user interface
 *      which supports the notion of pinnable menus.
 * 
 * EXAMPLE
 *      menu := vcWv.MakeMenu ("Menu",
 *                             vcWv.Pin (),
 *                             list ("Alpha", cmd));
 */
function Pin() {
    if (vcWv.UI (wv) == vcUi.OpenLook)
    {
        function make_pin()
            vcRegion.Create1 ("pin", vcPin.fac, FALSE, nil);

        hlp := "Pin the menu to the display";

        function set_pin(mp, event)
        {
            vcPin.Toggle (vcMenuPop._Region (mp));
            if (vcMenuPop._Pinup (mp, event))
                vcCmd.SetHelp (pin_cmd, "Take down this menu.");
            else
                vcCmd.SetHelp (pin_cmd, hlp);
        }

        pin_cmd := vcWv.Cmd (vcMenuPop.fac, set_pin);
        vcCmd.SetHelp (pin_cmd, hlp);
        list (make_pin, pin_cmd);

    }
    else
        nil;
}


/*\vscmuserDocpage vcWv.Separator -- build a visual menu separator
 * 
 * SYNOPSIS
 *      vscmVT vcWv.Separator()
 * 
 * DESCRIPTION
 *      This function builds a visual separator for menu entries.
 *      There can be as many separators in a menu as you wish.
 * 
 * EXAMPLE
 *      menu := vcWv.MakeMenu ("Menu",
 *                             list ("Alpha", cmd1),
 *                             vcWv.Separator (),
 *                             list ("Beta", cmd2));
 * 
 * SEE ALSO
 *      vcWv.Pin, vcWv.CheckedOption
 */
function Separator() {
    function make_separator()
        vcRegion.Create1 ("sep", vcFstyle.fac, vgFill.Solid,
                          list (cons (`size,
                                      vgSize.Create (25, 1, 1, 1, 1000, 1))));
    list (make_separator);
}


/*\vscmuserDocpage vcWv.Icon -- build an icon visual for a menu entry
 * 
 * SYNOPSIS
 *      vcWv.Icon(bitmap_file)
 *          vlStringVT bitmap_file;
 * 
 * DESCRIPTION
 *      If you want your menu visuals to be bitmap pictures, use
 *      this function to create a visual region to be used in
 *      conjunction with the vcWv.MakeMenu function.  The argument
 *      to this function is the filename of the bitmap file.
 * 
 * EXAMPLE
 *      cmd := vcWv.Cmd (...);
 *      a_menu := vcWv.MakeMenu ("Icon Menu",
 *                               list (vcWv.Icon ("bitmap"), cmd));
 */
function Icon(file)
    func() vcRegion.Create1 (file, vcIcon.fac, file, nil);;


EmptyCmd := vcWv.Cmd (vcWin.fac, func(w, e););
EmptyBindings := vcBind.Create ();

/*\ONLY_ON_UNIX vscmuserFunction vcWv.CheckedOption -- build a "checked" menu entry
 * 
 * SYNOPSIS
 *      vcWv.CheckedOption(label, proc, initial_state)
 *          vlStringVT label;
 *          vscmVT     proc;
 *          boolVT     initial_state;
 * 
 * DESCRIPTION
 *      Use this function to build a "checked" menu entry.  That is, a
 *      vcToggleButton widget shows up to the left of the menu label.
 *
 *      `proc' is called every time the state of the menu entry is
 *      changed.
 *
 *      `initial_state' indicates the initial state of the
 *      vcToggleButton; TRUE for set, and FALSE for unset.
 * 
 * EXAMPLE
 *      function toggle_state(event, new_state)
 *          check_state = new_state;
 * 
 *      menu := vcWv.MakeMenu ("Menu",
 *                             list ("Alpha", cmd),
 *                             vcWv.CheckedOption ("Checked Option",
 *                                                 toggle_state, TRUE));
 * 
 * SEE ALSO
 *      vcWv.Separator
 */
function CheckedOption(label_input, cmd, state) {
    all_toggles := nil;

    toggle_cmd := vcWv.Cmd (vcCmd.Fac (cmd), func(data, event) {
        state = !state;
        foreach (t; vscript.list_iter (all_toggles)) {
            if (state)
                vcToggleButton.Set (t);
            else
                vcToggleButton.Unset (t);
            w := vcRegion.Window (t);
            if ((w != nil) && vcWin.IsAlive (w)) {
                vcRegion.QueueForRefresh (t);
                vcWin.UpdateMenus (w);
            } /* if */
        } /* foreach */
        vcCmd.Run (cmd, data, event, list (state), TRUE);
    });

    function make_check() {
        if (istype (label_input, vlStringVT)) {
            toggle := vcRegion.Create ("", vcToggleButton.fac, state,
                                       label <- label_input);
            table := toggle;
        } /* if */
        else {
            toggle := vcRegion.Create ("", vcToggleButton.fac, state);
            table := vcTable.Make ("", 1, 2, toggle, label_input);
        } /* else */
        vcRegion.SetBindings (toggle, EmptyBindings);
        all_toggles = cons (toggle, all_toggles);
        vcRegion.DestroyNotify (toggle, func(r)
            all_toggles = vscript.delete_list_element (all_toggles, toggle););
        table;
    } /* function make_check */

    list (make_check, toggle_cmd);
} /* function CheckedOption */

Interrupts (vcWv.fac, '\^C');


/*\vscmuserDocpage vcWv.PullDownMenu -- create a button that triggers a menu
 * 
 * SYNOPSIS
 *      vcRegionVT vcWv.PullDownMenu(region_name, label_string, menu)
 *          vlStringVT region_name;
 *          vlStringVT label_string;
 *          vscmVT     menu;
 * 
 * DESCRIPTION
 *      Sometimes it's necessary to locate a pull down menu
 *      separately from the rest of the menus.  This function
 *      provides a convenient interface that creates a button which
 *      triggers a menu.
 * 
 * EXAMPLE
 *      popup := vcWv.MakeMenu ("Popup",
 *                              list (list ("This", ""), list ("is", "")),
 *                              list (list ("a", ""), list ("popup", "")),
 *                              list (list ("menu", ""), list ("", "")));
 * 
 *      vcTable.Set (table, 3, 3,
 *                   vcWv.PullDownMenu ("3:3", "Pulldown Menu", popup));
 *      vcWin.Create (table);
 */
function PullDownMenu(name, str, menu) {
    arrow := vcRegion.Create ("arrow", vcArrow.fac, Orient.down,
                             triangle <- vlTrueFalse.True);
    label := vcRegion.Create ("str", vcLabel.fac, str);
    label_and_arrow := vcTable.Create ("la", 1, 2);
    vcTable.Add (label_and_arrow, label);
    vcTable.Add (label_and_arrow, arrow);
    border := vcRegion.Create (name, vcBorder.fac, label_and_arrow);
    button := vcRegion.Create (name, vcButton.fac,
                              func(ev) vcMenuPop.create (menu, ev);,
                              fire_on_down <- TRUE);
    vcRegion.AddChild (button, border);
    vcButton.HighlightRegion (button, border);
    button;
}


/*\vscmuserDocpage vcWv.CommandSequence -- run a sequence of commands
 * 
 * SYNOPSIS
 *      vcWv.CommandSequence (list_of_command_strings_or_test_functions, 
 *                            environment_list_or_nil,
 *                            func(str) output_handler (...);,
 *                            func(str) error_handler (...);,
 *                            func(mode, code) exit_handler;,
 *                            hostname_or_nil);
 *
 * DESCRIPTION
 *      If a tool wants to invoke a sequence of shell commands, it
 *      can provide a script for each of the target operating
 *      systems, or it can use vcSys.Command or vcSys.Program to run
 *      the commands.  The disadvantage of providing a script is
 *      that that script won't be portable, and it will increase the
 *      propogation of small files in the installation directory.
 *      The problem with the vcSys. functions is ensuring that the
 *      first command completes before the second is invoked, etc.
 * 
 *      There are no functions provided to spawn a command and
 *      "block" until it completes, since this would "freeze" the
 *      user interface and prevent the user from hitting cancel
 *      buttons, delay handling refresh or resize requests on the
 *      window, etc.  It is possible to write such a function as a C
 *      callback that spawns the subprocess and then runs the user
 *      interface loop until the exit handler is called, but this is
 *      inelegent, and not recommended.  It causes the stack to grow
 *      arbitrarily, and erodes the multitasking nature of the
 *      system, where different windows created from the same
 *      process can continue to operate independently while some of
 *      them are "waiting" for subprocesses to complete.
 * 
 *      This function was created to solve this problem.  It runs a
 *      sequence of shell commands, ensuring that the first one
 *      completes before the second is run.  The sequence is given
 *      as a list of strings in cmd_list.  ViewScript functions can
 *      be inserted in the list of things to call.  The return value
 *      of the function is tested to see whether to quit the
 *      sequence.
 * 
 * EXAMPLE
 *      This example runs a program, checks to see whether an output
 *      file was created in the WDIR path, and then runs another
 *      program on that output file.  In this example, the first
 *      program is "cp", and it copies the input file, foo.in, to
 *      the output file foo.out.  If the input file is present, the
 *      command succeeds, the output file is created, the test
 *      passes, and the "cat" command is run to display the contents
 *      of the output file in the log.  If foo.in does not exist,
 *      "cp" prints an error message to the log window, the output
 *      file is not created, the test fails, and the "cat" command
 *      is not run.
 * 
 *      We want any output generated by the program to be written to
 *      a logging window.  So we'll create this window with the
 *      vcTxShell facility, and make sure the output function given
 *      to vcWv.CommandSequence writes to that window.
 * 
 * 
 *      sh := vcRegion.Create ("Output Log", vcTxShell.fac, "");
 *      vcWin.Create (sh);
 * 
 *      fname := "foo";
 *      vcWv.CommandSequence (list ("rm -f " + fname + ".out",
 *                                  "cp " + fname + ".in " + fname + ".out",
 *                                  func() fexists (fname + ".out");,
 *                                  "cat " + fname + ".out"),
 *                            nil, func(str) vcTxShell.PrintString (sh, str);,
 *                            nil, nil, nil);
 * 
 * NOTE
 *      The test function can be written to do any operation, such
 *      as read a file into a tool, put up a dialog box, invoke
 *      another vcWv.CommandSequence, etc.  Just make sure it
 *      returns TRUE if you want to continue with the sequence.  It
 *      can also be used to check a flag controlled by "cancel"
 *      button or binding (such as ^C), to allow users to abort the
 *      sequence.
 * 
 * SEE ALSO
 *      vcSys.Command, vcSys.Program
 */
function CommandSequence(cmd_list, env, output_handler, error_handler,
                         exit_handler, host)
{
    function next()
        CommandSequence (cdr (cmd_list), env, output_handler,
                         error_handler, exit_handler, host);

    if (cmd_list != nil) {
        action := car (cmd_list);
        if (istype (action, vlStringVT))
            vcSys.Command (action, env, output_handler, error_handler,
                           func(mode, code)
                           {
                               if (exit_handler != nil)
                                   exit_handler (mode, code);
                               next ();
                           }, host);
        else if (action ())
            next ();
    }

} /* CommandSequence */


/*\intDocpage vcWv.Confirm -- confirm execution of a command
 *
 * SYNOPSIS
 *      vcWv.Confirm(data, event, ok_cmd, message)
 *          vscmVT     data;
 *          vcEventVT  event;
 *          vcCmdVT    ok_cmd;
 *          vlStringVT message;
 *
 * DESCRIPTION
 *      I haven't figured yet whether this functionality by itself
 *      is really of that much use to people.  vcWv.CmdWithConfirm is
 *      really the thing you want, because it causes the
 *      confirmation command not to be recorded while recording the
 *      command `ok_cmd'.
 *
 * SEE ALSO
 *      vcWv.CmdWithConfirm, which handles prompting for arguments properly.
 */
function Confirm(data, event, vcCmdVT ok_cmd, message)
{
    /*
     * Restricting this to commands that take no arguments I have
     * found is not really necessary:
     *
     * if (vcCmd.RequiredArgs (ok_cmd) != 0)
     *     error ("Commands that require arguments not allowed", ok_cmd);
     */

    /*
     * Allow for macro recordability.  We don't want to be creating
     * our dialog window if macro replaying is on -- our job was
     * taken care of during the recording process.  During
     * recording, the `ok_cmd' either gets run via vcCmd.Run or
     * not, if the user hits Cancel, at which point the command got
     * properly recorded or properly omitted.
     */
    if (vcMac.IsReplayingOn () == FALSE)
    {
        msg_region := vcRegion.Create ("OkOrCancel:msg", vcStaticText.fac,
                                       message);
        button_row :=
            vcRegion.Create ("OkOrCancel:butnRow", vcButnRow.fac,
                             list (list ("OK", "OK",
                                         func() vcCmd.Run(ok_cmd, data, event,
                                                          nil, TRUE);,
                                         list (close_window_first <- TRUE)),
                                   list ("Cancel", "Cancel",
                                         func();,
                                         list (close_window_first <- TRUE))));

        _DialogInternal (vcEvent.Window (event), "Dialog",
                         msg_region, button_row);

    } /* if not replaying macros */

} /* Confirm */


/*\vscmuserDocpage vcWv.ConfirmCmd -- create a confirmation command
 *
 * SYNOPSIS
 *      vcCmdVT vcWv.ConfirmCmd(cmd, message)
 *          vcCmdVT    cmd;
 *          vlStringVT message;
 *
 * DESCRIPTION
 *      Creates a command object which confirms with the user before
 *      executing command `cmd'.  The dialog box consists of the
 *      text `message', plus a row of two buttons, OK and Cancel.
 *
 *      `Message' should be a concise, simple question, to which
 *      "OK" is the default response meaning "go ahead and perform
 *      the action".  `Message' can be a multi-line string; in this
 *      case it should consist of descriptive text, a blank line,
 *      then the simple question described above.
 *
 * SEE ALSO
 *      vcWv.CmdWithConfirm, which deals better with commands that
 *      have parameters, and allows the user to supply a function
 *      that returns the message to put into the confirmation box,
 *      or to avoid using a confirmation box altogether, based
 *      on the parameters that the user types.
 *
 * EXAMPLE
 *      cmdDoIt := vcWv.Cmd (fac,
 *                           func(d, e) printf("Just Do It!\n");,
 *                           "do_it");
 *      cmdDoItConfirm := vcWv.ConfirmCmd (cmdDoIt, "OK to do it?");
 */
function ConfirmCmd(cmd, message)
    Cmd (vcCmd.Fac (cmd),
         func(d, e) Confirm (d, e, cmd, message););


/*\vscmuserDocpage vcWv.CmdWithConfirm -- create a command with confirmation
 *
 * SYNOPSIS
 *      vcCmdVT vcWv.CmdWithConfirm(name, action_function, confirm_function,
 *                                  enable_macros, name, title, params...)
 *      (see vcWv.Cmd, vcWv.Param for details on name, title, params...)
 *
 * DESCRIPTION
 *      Creates a command object which confirms with the user before
 *      executing `action_function'.  The dialog box consists of a
 *      visual generated by `confirm_function', plus a row of two
 *      buttons, OK and Cancel.  `confirm_function' is called to get
 *      more information (text or region) to put into the dialog box,
 *      and it is passed exactly the same information as the
 *      action_function -- that is, the facility data, the event, and
 *      all the parameters supplied by the user.  `confirm_function'
 *      should return one of the following values:
 *
 *              text string             Placed in StaticText region.
 *              region                  Placed in dialog box.
 *              FALSE                   No confirmation box is created.
 *
 *      The message generated by the confirm_function, if any, should
 *      be a concise, simple question, to which "OK" is the default
 *      response meaning "go ahead and perform the action".  It can be
 *      a multi-line string; in this case it should consist of descriptive
 *      text, a blank line, then the simple question described above.
 *
 *      The `enable_macros' argument is used to specify whether macro
 *      recording should be allowed.  Do not use vcCmd.EnableMacro on
 *      this command.
 *      
 *
 * RETURNS
 *      A command.
 *
 * EXAMPLE
 *      This is the example from the demo.  It takes a few lines to
 *      do this, but this expresses all the main concepts:
 *
 *      define_enum ("SafeOrDangerousVT", list ("Safe", "Dangerous"));
 *      
 *      function doit(data, event, arg) {
 *          rgn := vcRegion.Create ("blah", vcLabel.fac, "Did it!");
 *          vcRegion.Font (rgn, vgFont.largeb);
 *          margin := vcRegion.Create ("margin", vcMargin.fac, rgn);
 *          vcWin.Create (margin,
 *                        vcWin.NoCmdline + vcWin.NoBindings + vcWin.NoStatus);
 *      }
 *
 *      function confirm_doit(data, event, arg)
 *          if (arg == SafeOrDangerous.Safe)
 *              FALSE;
 *          else
 *      "You have selected a DANGEROUS command,
 *      so you better be damn sure you want to do it.
 *      By the way, mouse coords == " + sprintf ("(%d,%d)", vcEvent.X (event),
 *                                               vcEvent.Y (event)) + "
 *      
 *      OK to do it?";
 *      
 *      cmd := vcWv.CmdWithConfirm (vcWin.fac, doit, confirm_doit, TRUE,
 *                                  "doit", "Do It",
 *          vcWv.Param ("", SafeOrDangerousVT, vcWv.Sticky,
 *                      SafeOrDangerous.Safe));
 *      
 *      rgnButnRow := vcRegion.Create ("Example", vcButnRow.fac,
 *                                     list (list ("OK", "Press here to do it",
 *                                                 cmd, nil)));
 *      vcWin.Create (rgnButnRow,
 *                    vcWin.NoCmdline + vcWin.NoBindings + vcWin.NoStatus);
 *      
 */
function CmdWithConfirm(args...) {
    fac         := car (args);  args = cdr (args);
    fun         := car (args);  args = cdr (args);
    confirm     := car (args);  args = cdr (args);
    do_macros   := car (args);  args = cdr (args);
    name        := car (args);  args = cdr (args);
    title       := car (args);  args = cdr (args);

    just_do_it := FALSE;

    function confirm_before_action(args...) {
        if (vcMac.IsReplayingOn () || just_do_it)
            apply (fun, args);
        else {
            function do_it() {
                vcCmd.EnableMacro (cmd, do_macros);
                just_do_it = TRUE;
                vcCmd.Run (cmd, data, event, args, TRUE);
                just_do_it = FALSE;
                vcCmd.EnableMacro (cmd, vlOnOff.Off);
            } /* function do_it */

            confirm_result := (confirm == nil) ? nil : apply (confirm, args);

            data := car (args);         args = cdr (args);
            event := car (args);        args = cdr (args);

            if (confirm_result == FALSE)
                do_it ();
            else {
                if (istype (confirm_result, vcRegionVT))
                    ;
                else if (istype (confirm_result, vlStringVT))
                    confirm_result = vcRegion.Create ("", vcStaticText.fac,
                                                      confirm_result);
                else
                    error ("Invalid confirm_function return",
                           confirm_result);

                button_row :=
                    vcRegion.Create ("OkOrCancel:butnRow", vcButnRow.fac,
                             list (list ("OK", "OK", do_it,
                                         list (close_window_first <- TRUE)),
                                   list ("Cancel", "Cancel", func();,
                                         list (close_window_first <- TRUE))));

                _DialogInternal (vcEvent.Window (event), "Dialog",
                                 confirm_result, button_row);
            } /* else */
        } /* else */
    }; /* func */

    cmd := vcCmd.Create (fac, confirm_before_action, name, title, args);
    vcCmd.EnableMacro (cmd, vlOnOff.Off);
    cmd;
} /* function CmdWithConfirm */
    
    


/*\vscmuserDocpage vcWv.OkOrCancel -- create confirmation dialog box
 * 
 * SYNOPSIS
 *      vcWindowVT vcWv.OkOrCancel (win, message, ok_cmd, cancel_cmd)
 *          vcWinVT win;            // parent window
 *          vlStringVT message;     // message to user 
 *          vcCmdVT    ok_cmd;      // command invoked when "OK" is hit
 *          vcCmdVT    cancel_cmd;  // ... when "Cancel" is hit
 * 
 * DESCRIPTION
 *      vcWv.CmdWithConfirm is preferable -- it allows the U/I to be
 *      macro recordable.
 *
 *      This function is useful for verifying with the user that an
 *      action is OK before executing it.  A dialog window is created
 *      containing the text of the message specified by `message' and a
 *      button row with buttons OK, Cancel.
 *
 *      `Message' should be a concise, simple question, to which
 *      "OK" is the default response meaning "go ahead and perform
 *      the action".  `Message' can be a multi-line string; in this
 *      case it should consist of descriptive text, a blank line,
 *      then the simple question described above.
 * 
 * NOTE
 *      This function is the simplest version of vcWv.Dialog(man).
 *
 * MACRO RECORDABILITY
 *      Use vcWv.CmdWithConfirm, or, less preferably, make sure that
 *      `ok_cmd' and `cancel_cmd' destroy the dialog window, by
 *      calling vcWin.Destroy(vcEvent.Window(event)).
 * 
 * SEE ALSO
 *      vcWv.CmdWithConfirm, vcWv.Dialog
 */
function OkOrCancel(win, message, ok_cmd, cancel_cmd)
{
    if (istype (message, vlStringVT))
        message = vcRegion.Create ("OkOrCancel:msg", vcStaticText.fac,
                                    message);
    buttons := vcRegion.Create ("OkOrCancel:butnRow", vcButnRow.fac,
                                list (list ("Accept", "Accept",
                                            ok_cmd,
                                            list (close_window <- TRUE)),
                                      list ("Cancel", "Cancel",
                                            cancel_cmd,
                                            list (close_window <- TRUE))));

    _DialogInternal (win, "Dialog", message, buttons);

} /* OkOrCancel */


/*\vscmuserDocpage vcWv.Dialog -- create a dialog box
 *
 * SYNOPSIS
 *      vcWindowVT vcWv.Dialog(win, message, interrogative, button_row)
 *          vcWindowVT win;           // or nil
 *          vlStringVT title;
 *          vlStringVT message;       // or nil
 *          vlStringVT interrogative;
 *          vcRegionVT button_row;
 *
 * DESCRIPTION
 *      vcWv.CmdWithConfirm is preferable -- it allows the U/I to be
 *      macro recordable.
 *
 *      Creates a dialog box with a custom button row.  The simplest
 *      case of an "OK-or-cancel" dialog box is taken care of by
 *      vcWv.CmdWithConfirm and vcWv.OkOrCancel.
 *
 * MACRO RECORDABILITY
 *      Use vcWv.CmdWithConfirm, or, less preferably, make sure that
 *      `ok_cmd' and `cancel_cmd' destroy the dialog window, by
 *      calling vcWin.Destroy(vcEvent.Window(event)).
 *
 * SEE ALSO
 *      vcWv.CmdWithConfirm, vcWv.OkOrCancel
 */
function Dialog(win, vlStringVT title, message, vlStringVT interrogative,
                vcRegionVT button_row)
{
    declare msg;

    if (vcRegion.Fac (button_row) != vcButnRow.fac)
        vlEx.Throw2 (vlEx.FAILURE, "vcWv", vlErrSev.Error,
                     "Last parameter to vcWv.Dialog must be a vcButnRow.");

    if (message == nil)
        msg = interrogative;
    else
        if (string_ref (message, string_length (message) - 1) == '\n')
            msg := message + "\n" + interrogative;
        else
            msg := message + "\n\n" + interrogative;
    msg_region := vcRegion.Create ("Dialog msg", vcStaticText.fac, msg,
                                   width <- 0,
                                   max_width <- 1);

    _DialogInternal (win, title, msg_region, button_row);

} /* Dialog */


/*\intFunction _DialogInternal -- internal func to Dialog and OkOrCancel
 */
function _DialogInternal(win, name, msg_region, button_row)
{
        table := vcTable.Create ("Dialog table", 2, 1);
        vcTable.Row (table, -1, vcTable.Border.Invisible,
                     0, 0, 0, vgAlign.MiddleCenter);
        vcTable.Col (table, -1, vcTable.Border.Invisible,
                     0, 0, 0, vgAlign.MiddleCenter);

        table2 := vcTable.Create ("Dialog table2", 1, 2);
        vcTable.Col (table2, -1, vcTable.Border.Invisible,
                     0, 0, 0, vgAlign.MiddleRight);
        vcTable.Add (table2, vcRegion.Create ("Dialog icon", vcIcon.fac,
                                              "motif_head"));
        vcTable.Add (table2, msg_region);
        vcTable.Add (table, table2);
        vcTable.Add (table, button_row);

        margin := vcRegion.Create (name, vcMargin.fac, table);

        if (win != nil) {
            win = vcWin.Ancestor (win);
            popup := vcWin.Popup (win, margin,
                                  vcButnRow.GetDefault (button_row),
                                  vcWin.NoCmdline + vcWin.NoStatus
                                  + vcWin.NoBindings
                                  /* + vcWin.NoResize + vcWin.Transient */);
            vcWin.Forward (win, popup);
            popup;
        }
        else
            vcWin.Create (margin, vcWin.NoCmdline + vcWin.NoStatus
                          + vcWin.NoBindings /* + vcWin.NoResize */);

} /* _DialogInternal */


/*\vscmuserFunction vcWv.Delay -- Run a function at "top level"
 *
 * DESCRIPTION
 *      There are various circumstances when you want to perform
 *      some function very soon, as far as the user is concerned,
 *      but not immediately, as far as the flow of control is
 *      concerned.  One example is raising an exception from
 *      inside a notify handler -- this would prevent other
 *      notifications from occurring.  Instead you want to wait
 *      until the flow of control pops out to the "top level",
 *      and raise the exception then.  vcWv.Delay provides a
 *      handy way of doing that.
 *
 * EXAMPLE
 *      vcWv.Delay (wv, func() vlEx.Throw2 (...););
 */
function Delay(wv, zero_arg_function)
    vcWv.Timer (wv, 0, zero_arg_function, FALSE);

/*\vscmuserFunction vcWv.RequireFac -- Load a facility if not loaded already
 *
 * DESCRIPTION
 *      Loads a facility written in Viewscript into the current process,
 *      if it hasn't been loaded already.
 */
function RequireFac(wv, facility_name) {
    if (!isbound (string2symbol (facility_name), nil))
        vcFac.Create (facility_name, vscmVT);
} /* function RequireFac */
