/* This file contains the following major sections:
   - some utility functions
   - commandline with less than, etc indicators for numeric criteria
   - vertical picklist facility vfpgaVertPick
   - popup picklist facility vfpgaPopPick
   - architecture picklist facility vfpgaArchPick
   - confirm popups which may be added to core in 5.2
*/

/* This functions sets a picklist entry if the string is found in
   the in_string set of space-delimited sub-strings.  It handles both one
   dimensional and two dimensional picklists (whose list structures differ)
   by examining the picklist dimensions. */
function setpick_if_found (picklist, in_string, in_picklist) {
    declare tmp, rows, cols;

    vcPicklist.Dimensions (picklist, &rows, &cols);
    in_string = string_token(in_string, &tmp, " ()\"");
    while (!null(in_string) && !null(tmp)) {
        if (cols == 1) {
           for (temp_list:=in_picklist, count:=0; !null (temp_list); 
                temp_list = cdr (temp_list), ++count)

                if (string_eqi (cdar (temp_list), tmp)) {
                   vcPicklist.SetPick (picklist, count); 
                   break;
                } }
        else 
           for (temp_list:=in_picklist, count:=0; !null (temp_list); 
                temp_list = cdr (temp_list))

                for (temp_sub:=car (temp_list); !null (temp_sub); 
                     temp_sub = cdr (temp_sub), ++count)

                     if (string_eqi (cdar (temp_sub), tmp)) {
                         vcPicklist.SetPick (picklist, count); 
                         break;
                     }
        in_string = string_token(in_string, &tmp, " ()\"");
    }
}

/* This function concatenates or removes strings from the cmdline
   region passed in for non-exclusive picklists, or just sets the
   cmdline value for exclusive picklists */
function string_op (r_cmdline, onoff, data, exclusive) {
    if (!exclusive) {
        str := vcRegion.Get(r_cmdline, vlStringVT);
        if (onoff) {
            if ((string_eq(str, "") || (str[string_length(str)-1] == ' ')))
                vcCmdline.Set (r_cmdline, str + data + " ");
            else
                vcCmdline.Set (r_cmdline, str + " " + data + " ");
        }
        else {
            declare tmp;
            str = string_token(str, &tmp, " "); newstr := "";
            while (!null(str) && !null(tmp)) {
                if (!string_eq(data, tmp)) newstr = newstr + tmp + " ";
                str = string_token(str, &tmp, " "); 
            }
            vcCmdline.Set(r_cmdline, newstr);
        }
    }
    else if (onoff == vlOnOff.On) vcCmdline.Set (r_cmdline, data); 
}

function checkNumStr(str) {
    numstr := "";
    if (isstring(str)) {
        numstr := str;
        for (i := 0; i < string_length(numstr); ++i) {
            if (!((numstr[i] >= '0') && (numstr[i] <= '9'))) {
                numstr[i] = '\0'; 
            }
        }
    }
    numstr;
}

package vfpgaFromTo (_Create, _Get) {
   function _Create (arg, region) {
       declare num1, num2;
       arg = string_token(arg, &num1, " ");
       if (null(num1)) num1 = "";
       if (null(arg)) arg = "";
       arg = string_token(arg, &num2, " ");
       if (null(num2)) num2 = "";
       r_root := vcTable.Make("dbl_cmdline", 1, 4,
                              vcRegion.Create("", vcLabel.fac, "From:"),
                              vcRegion.Create("", vcCmdline.fac, num1,
                                              width <- 5),
                              vcRegion.Create("", vcLabel.fac, "To:"),
                              vcRegion.Create("", vcCmdline.fac, num2,
                                              width <- 5));
       vcRegion.AddChild (region, r_root);
       r_root; 
   }

   function _Get (arg, region, ts) {
       str1 := checkNumStr(vcRegion.Get(vcTable.Get(arg, 0, 1), vlStringVT));
       str2 := checkNumStr(vcRegion.Get(vcTable.Get(arg, 0, 3), vlStringVT));
       if (string_eq(str1, "") || string_eq(str2, "")) ret_str := "";
       else ret_str :=  str1 + " " + str2;
   }
}

vcFac.CreateFromPackage ("vfpgaFromTo", vscmVT, vfpgaFromTo);


/*** NUMERIC FIELD WITH COMPARATOR FACILITY ***/
/* This facility still needs to generate the "less 24" style string, and
   be able to parse a default string into the correct style.  Syntax
   checking also has to be done in case an non-integer is entered, or
   an incorrect range is entered. */
package vfpgaNumComp (_Create, _Get) {

   function _Create (arg, region) {
      r_compare := vcRegion.Create ("", vcPicklist.fac, 
                                    list (list (list ("<", "less"), 
                                                list (">", "greater"), 
                                                list ("=", "equal"),
                                                list ("Range", "range"))));
      /* Parse input argument */
      declare compare;
      num := string_token(arg, &compare, " ");
      if (null(compare)) compare = "";
      if (null(num)) num = "";
      
      /* Set correct compare picklist button */
      if (string_eqi(compare, "less")) vcPicklist.SetPick(r_compare, 0);
      else if (string_eqi(compare,"greater")) vcPicklist.SetPick(r_compare, 1);
      else if (string_eqi(compare,"equal")) vcPicklist.SetPick(r_compare, 2);
      else if (string_eqi(compare,"range")) vcPicklist.SetPick(r_compare, 3);

      /* Set command line region */
      if (string_eqi(compare, "range")) 
          r_cmdline := vcRegion.Create("", vfpgaFromTo.fac, num);
      else {
          declare trim;
          string_token(num, &trim, " ");
          if (null(trim)) trim = "";
          r_cmdline := vcRegion.Create ("", vcCmdline.fac, trim);
      }
      r_root := vcTable.Make ("", 1, 2, r_compare, r_cmdline);
      vcRegion.AddChild (region, r_root);
      
      vcPicklist.Notify(r_compare, func(data, onoff) {
          if (onoff) {
              oldstr := vcRegion.Get(vcTable.Get(r_root, 0, 1),vlStringVT);
              if (string_eqi(car(data), "range"))
                  r_new := vcRegion.Create("dbl_cmdline", vfpgaFromTo.fac,
                                           oldstr);
              else if (string_eqi(car(data), ""))
                  r_new := vcRegion.Create("one_cmdline", vcCmdline.fac, "");
              else {
                  declare firststr;
                  string_token(oldstr, &firststr, " ");
                  if (null(firststr)) firststr = "";
                  r_new := vcRegion.Create("one_cmdline", vcCmdline.fac, 
                                           firststr);
              }
              vcTable.Set(r_root, 0, 1, r_new);
          }
      });
      
      r_root; 
  }

   function _Get (arg, region, ts) {
       compare := vcRegion.Get(vcTable.Get(arg, 0, 0), vscmVT); 
       if (null(compare) || string_eq(car(compare), "")) str := "";
       else {
           num := vcRegion.Get (vcTable.Get (arg, 0, 1), vlStringVT); 
           if (string_eq(car(compare), "range")) {
               if (string_eq(num, "")) str := "";
               else str := car(compare) + " " + num;
           }
           else {
               checknum := checkNumStr(num);
               if (string_eq(checknum, "")) str := "";
               else str := car(compare) + " " + checknum;
           }
       }
   }
}


vcFac.CreateFromPackage ("vfpgaNumComp", vscmVT, vfpgaNumComp);


/*** VERTICAL PICKLIST FACILITY ***/

/* Caution: because of the way the defaults are set in the function
   setpick_if_found, you will get the wrong results if the picklist you
   pass in has spaces on the right hand side strings.  Example:
   list (list ("a", "B C"), list ("d", "e_f"));
   The first picklist entry is a problem, but the second is OK.  What will
   happen is when "a" is selected and the picklist brought up again, no
   value will be highlighted. */

package vfpgaVertPick (_Create, _Get) {

   function _Create (arg, region) {
      fl_string := "vfpga." + vcRegion.PrefGet (region, "funclist") + "();";
      picklist := vcWin.EvalPrint (fl_string, nil);
      r_pick := vcRegion.Create ("", vcPicklist.fac, picklist,
         exclusive <- vcRegion.PrefGet (region, "exclusive"));

      setpick_if_found (r_pick, arg, picklist);
      vcRegion.AddChild (region, r_pick);
      r_pick; }

   function _Get (arg, region, ts) { vcRegion.Get (arg, ts); } }


vcFac.CreateFromPackage ("vfpgaVertPick", vscmVT, vfpgaVertPick);

vlPref.Add (vfpgaVertPick.prefs, "funclist", vlStringVT,
   "function to return entries in picklist", nil);
vlPref.Add (vfpgaVertPick.prefs, "exclusive", vlTrueFalseVT,
   "restrict to single selection in list", TRUE);


/*** COMMAND LINE WITH POPUP PICKLIST FOR BROWSE ***/


package vfpgaPopPick (_Create, _Get) {

   function show_pick (e) {

      /* Get information from parents */
      r_browsebut := vcEvent.Region (e);
      r_table := vcRegion.Parent (r_browsebut);
      r_cmdline := vcTable.Get (r_table, 0, 0);
      r_parent := vcRegion.Parent (r_table);

      fl_string := "vfpga." + vcRegion.PrefGet (r_parent, "funclist") + "();";
      picklist := vcWin.EvalPrint (fl_string, nil);
      exclusive := vcRegion.PrefGet (r_parent, "exclusive");

      /* Create picklist */
      r_pick := vcRegion.Create ("r_pick", vcPicklist.fac, picklist,
                                 exclusive <- exclusive);
      setpick_if_found(r_pick, vcRegion.Get (r_cmdline, vlStringVT), picklist);
      vcPicklist.Notify (r_pick, func (data, onoff) {
         string_op (r_cmdline, onoff, data, exclusive); } ); 

      /* Create other regions */
      r_dismiss := vcRegion.Create ("r_dismiss", vcButnRow.fac,
         list(list ("Accept","Accept", func();, list (close_window <- TRUE))));
      if (length(picklist) > 8) 
          r_root := vcTable.Make ("Select from list", 2, 1, 
                                  vcPicklist.Scroll ("r_scroll", r_pick, 
                                                     5, 8, 10), 
                                  r_dismiss);
      else
          r_root := vcTable.Make ("Select from list", 2, 1, r_pick, r_dismiss);

      /* Display browse window and set up forwarding and notification */
      w_parent := vcEvent.Window (e);
      vcWin.Forward (w_parent, vcWin.Popup (w_parent, r_root, r_pick,
         vcWin.NoMacros));
      vcRegion.SetFocus (r_root, r_pick);
      vcRegion.Enable (r_browsebut, FALSE, TRUE);
      vcRegion.DestroyNotify (r_root,
         func (e) vcRegion.Enable (r_browsebut, TRUE, TRUE);); }


   function _Create (arg, region) {

      /* Create a command line region which displays the default */
      r_cmdline := vcRegion.Create ("r_cmdline", vcCmdline.fac, arg);

      /* Create a button which brings up the picklist in a window */
      r_button := vcWv.LabelButtonCreate("", "Browse...", nil, show_pick, nil);

      /* Create the root region, containing the commandline and the button */
      r_table := vcTable.Make ("r_table", 1, 2, r_cmdline, r_button);
      vcRegion.AddChild (region, r_table);

      r_table; }


   /* Get data out of 0,0 position of table, which is the command line */
   function _Get (r_table, region, ts) {
      vcRegion.Get (vcTable.Get (r_table, 0, 0), ts); } }


vcFac.CreateFromPackage ("vfpgaPopPick", vscmVT, vfpgaPopPick);
vlPref.Add (vfpgaPopPick.prefs, "funclist", vlStringVT,
   "function to return entries in picklist", nil);
vlPref.Add (vfpgaPopPick.prefs, "exclusive", vlTrueFalseVT,
   "Only one item can be selected", TRUE);


/*** Filtered Picklist facility ***/

package vfpgaFilterPick (_Create, _Get) {

    function _Create (arg, region) {

        flstr := "vfpga." + vcRegion.PrefGet (region, "funclist") + "();";
        picklist := vcWin.EvalPrint (flstr, nil);

        /* Create a command line region which displays the default */
        r_filter := vcRegion.Create ("cmdline", vcCmdline.fac, arg);

        function filterpick(str) {
            flist := CallWildcard(str, picklist, 1);
            r_pick := vcPicklist.Scroll("pickscroll",
                        vcRegion.Create ("pick", vcPicklist.fac, flist,
                                         macro_on <- TRUE),
                                        10, 10, 10);
            setpick_if_found (r_pick, vcRegion.Get(r_filter, vlStringVT), 
                              flist); 
            vcPicklist.Notify(r_pick, func (data, onoff) {
                vcCmdline.Set(r_filter, data); });
            r_pick;
        }

        /* Create the picklist */
        r_picklist := filterpick("*");
        
        /* Create the root region, containing the cmdline and the picklist */
        r_table := vcTable.Make ("list_table", 2, 1, r_filter, r_picklist);
        vcRegion.AddChild (region, r_table);

        vcCmdline.CallOnReturn (r_filter, func (e, str) { 
            if ((lindex(str, '*') != FALSE) || (lindex (str, '?') != FALSE)) 
                vcTable.Set(r_table, 1, 0, filterpick(str)); });

        r_filter; 
    }

    function _Get (arg, region, ts) { vcRegion.Get (arg, ts); } 
}
vcFac.CreateFromPackage ("vfpgaFilterPick", vscmVT, vfpgaFilterPick);
vlPref.Add (vfpgaFilterPick.prefs, "funclist", vlStringVT,
            "function to return entries in picklist", nil);

                                    


/* This facility is very similar to the vfpgaPopPick facility, except
 * for three things: the input list format, a filter region, and 
 * a scroll bar. 
 */
package vfpgaArchPick (_Create, _Get) {

   function _Create (arg, region) {
       
       fl := vcRegion.PrefGet(region, "funclist");

       /* Create a command line region which displays the default */
       r_cmdline := vcRegion.Create ("cmdline", vcCmdline.fac, arg);

       /* Create a button which brings up the picklist in a window */
       r_button := vcWv.LabelButtonCreate ("button", "Browse...",nil, 
                     vcWv.Cmd(fac, func(d, e, s) { 
                         vcCmdline.Set(r_cmdline, s);},
                              "PickArch", "Select Architecture",
                              vcWv.Param("Filter", vlStringVT, vcWv.Dynamic,
                                         func(d) arg;,
                                         vfpgaFilterPick.fac, 
                                         "Wildcard String",
                                         funclist <- fl)), nil);
       
       /* Create the root region, containing the commandline and the button */
       r_table := vcTable.Make ("cmdbut_table", 1, 2, r_cmdline, r_button);
       vcRegion.AddChild (region, r_table);

       r_cmdline; 
   }
   
   function _Get (arg, region, ts) { vcRegion.Get (arg, ts); } 

}


vcFac.CreateFromPackage ("vfpgaArchPick", vscmVT, vfpgaArchPick);
vlPref.Add (vfpgaArchPick.prefs, "funclist", vlStringVT,
   "function to return entries in picklist", nil);




/*** SPREADSHEET FACILITY ***/
/*
 * This facility builds a "spreadsheet" that matches the format 
 * of the list that is passed in via the el_list preference. There
 * is a browse button which runs the command generated by the function
 * passed in by the create routine.  The cg_arg argument for the 
 * the _Create routine is a function which takes one argument, the
 * name of element who's browse button was hit, and returns a command
 * object created by vcWv.Cmd.
 *
 * The _Create and _Get routines operate on the same type of data - 
 * an association list.  The _Create routine gets this data through
 * the el_list preference, and the _Get routine returns this data.
 * That list is of the following structure for Jedec spreadsheet:  
 *      list ( list ("Element", "Device")
 *             list ("a", "P22V10"), 
 *             list ("b", "P16L8"));
 * What is returned by the facility does not include the title list:
 *      list ( list ("a", "P22V10"), 
 *             list ("b", "P16L8"));
 *
 * That list is of the following structure for J2vhdl spreadsheet:  
 *      list ( list ("Element", "Manufacturer", "Device", "Exact Partname")
 *             list ("a", "AMD", "P22V10", "AMD PAL22V10-10PC PLD"), 
 *             list ("b", "AMD", "P16L8",  "AMD P16L8-5PC PAL"));
 *
 * There is one other routine, SetCmdline which is used to set the 
 * spreadsheet arguments for a particular element.
 *
 */
package vfpgaSpreadsheet (_Create, _Get, SetCmdline) {

    declare r_root;
    
    function _Create (cg_arg, region) {

        fl_string := "vfpga." + vcRegion.PrefGet(region, "el_list") + "();";
        elem_list := vcWin.EvalPrint (fl_string, nil);

        title_list := car(elem_list); 
        /* Count the number of elements in title list (number of columns) */
        for (t:=title_list, ncols:=1; !null(t); t=cdr(t), ncols=ncols+1);

        /* Create the root spreadsheet region */
        r_root = vcTable.Create ("", 1, ncols);
        
        /* Create the title labels and add to table */
        declare r_title;
        for (t:=title_list; !null(t); t=cdr(t)) {
            r_title = vcRegion.Create ("", vcLabel.fac, car(t));
            vcRegion.Font(r_title, vgFont.mediumb);
            vcTable.Add (r_root, r_title);
        }

        /* Add a rubber region where browse button will go */
        vcTable.Add (r_root, vcRubber.Create("", 0, 0));

        declare element;
        for (e:=cdr(elem_list); !null(e); e=cdr(e)) {
            element = car(e);
            el_name := car(element);

            /* Create a static text region which displays the element name */
            vcTable.Add (r_root, 
                         vcRegion.Create("", vcStaticText.fac, el_name)); 

            /* Create a command line region for each argument */
            for (a:=cdr(element); !null(a); a=cdr(a)) 
                vcTable.Add (r_root,
                             vcRegion.Create ("", vcCmdline.fac, car(a)));

            /* Create a button which runs the command argument */
            bb := vcWv.LabelButtonCreate (el_name, "Browse...", nil, 
                     func(e) {
                         name := vcRegion.Name(vcEvent.Region(e));
                         vcCmd.Run(cg_arg(name), nil, e, nil, TRUE); 
                     }, nil); 
            vcTable.Add (r_root, bb);
        }
        vcRegion.AddChild (region, r_root);
        r_root;
    }


    function _Get (arg, region, ts) { 
        declare nrows, ncols, r_cmd, r_txt, elem_list;
        elem_list := nil;
        vcTable.Dimensions(r_root, &nrows, &ncols);
        for (i:=1; i<nrows; i=i+1) {
            row_list := nil;
            for (j:=ncols-2; j>0; j=j-1) 
                row_list = cons(vcRegion.Get(vcTable.Get(r_root, i, j),
                                             vlStringVT), row_list);
            row_list = cons(vcRegion.Get(vcTable.Get(r_root, i, 0), 
                                         vlStringVT), row_list);
            elem_list = cons(row_list, elem_list);
        }
        elem_list;
    }
    
    function SetCmdline (namestr, cmdlist) { 
        /* Look for row with first element matching namestr */
        declare nrows, ncols, r_txt;
        if (!vcRegion.Destroyed(r_root)) {
            vcTable.Dimensions(r_root, &nrows, &ncols);
            for (i:=1; i<nrows; i=i+1) {
                r_txt = vcTable.Get(r_root, i, 0);
                if (string_eqi(namestr, vcRegion.Get(r_txt, vlStringVT))) {
                    for (j:=1; j<ncols-1; j=j+1, cmdlist=cdr(cmdlist))
                        vcCmdline.Set(vcTable.Get(r_root, i, j), car(cmdlist));
                    break;
                }
            }
        }
    } 
}


vcFac.CreateFromPackage ("vfpgaSpreadsheet", vscmVT, vfpgaSpreadsheet);
vlPref.Add (vfpgaSpreadsheet.prefs, "el_list", vlStringVT,
   "list of elements to set device for", nil);


/* Simple facility to allow creation of empty cmdline, but not allow
   the cmdline to remain empty */
package vfpgaNotEmptyString (_Create, _Get) {
   declare r_text;
   function _Create (arg, region) {
      r_text = vcRegion.Create ("", vcCmdline.fac, arg);
      vcRegion.AddChild (region, r_text);
      r_text; }

   function _Get (arg, region, ts) { 
       myerr := FALSE;
       catch (vlEx.FAILURE) 
           str := vcRegion.Get(r_text, vfpgaStringNotEmptyVT); 
       exception {
           myerr = TRUE;
       }

       if (myerr)
           error("vfpga", "String cannot be empty"); 
       str;
   }
}


vcFac.CreateFromPackage ("vfpgaNotEmptyString", vscmVT, vfpgaNotEmptyString);


/*** CONFIRM POPUPS ***/


/* Modified version of the 5.1.1 vcWv.OkOrCancel, with help pointer */
function vfpgaConfirm (helpname, text, ok_func) {
   table := vcTable.Create ("Confirm", 2, 1);
   win := vcRegion.Window (gr_log);
   vcTable.Add (table, vcRegion.Create("", vcStaticText.fac, text));
   buttons := vcRegion.Create ("", vcButnRow.fac, list (
      list ("Accept", "Accept", ok_func, list (close_window <- TRUE)),
      list ("Cancel", "Cancel", func ();, list (close_window <- TRUE)),
      list ("Help", "Help",
         func () vcAlert.SpawnHelp (win,  "vfpga", helpname);,
         list (close_window <- FALSE))));

   vcTable.Add (table, buttons);
   popup := vcWin.Popup (win, table, vcButnRow.GetDefault (buttons),
      vcWin.NoCmdline+vcWin.NoStatus+vcWin.NoBindings);
   vcWin.Forward (win, popup); }


/* Modified version of vfpgaConfirm which builds a table of filenames */
function vfpgaConfirmFiles (filelist, helpname, ok_func) {
   function dodelete () {
      for (; !null (filelist); filelist = cdr (filelist))
         vscript.vsDelete (car (filelist), vscript.vsFILE);
      cb_mainErr (0, "Status", nil, "Files deleted");
      ok_func (); }

   n_orig := length (filelist) + 1;
   n_rows := ((n_orig - 1) / 4) + 1;
   if (n_rows == 1) n_cols := n_orig;
   else n_cols :=  4;
   t_out := vcTable.Create ("", n_rows, n_cols);
   l_sort := vscript.Sort (filelist, string_gt);

   /* Form table into proper shape */
   declare i, j, k, s;
   for (i=0, k=0; i<n_rows; ++i)
      for (j=0; j<n_cols; ++j, ++k, l_sort=cdr(l_sort)) {
         if (null (s = car (l_sort))) break;
         vcTable.Add (t_out, vcRegion.Create ("", vcLabel.fac, s)); }

   /* Create popup */
   r_buttons := vcRegion.Create ("", vcButnRow.fac, list (
      list ("Accept", "Accept", dodelete, list (close_window <- TRUE)),
      list ("Cancel", "Cancel", func ();, list (close_window <- TRUE)),
      list ("Help", "Help",
         func () vcAlert.SpawnHelp (w_main,  "vfpga", helpname);,
         list (close_window <- FALSE))));
   t_main := vcTable.Make ("Confirm", 3, 1,
      vcRegion.Create ("", vcStaticText.fac,
         "OK to delete the following files?"), t_out, r_buttons);
   w_main := vcRegion.Window (gr_log);
   w_popup := vcWin.Popup (w_main, t_main, vcButnRow.GetDefault (r_buttons),
      vcWin.NoCmdline+vcWin.NoStatus+vcWin.NoBindings);
   vcWin.Forward (w_main, w_popup); }

/* Modified version of the 5.1.1 vcWv.OkOrCancel, with help pointer 
   and cmdline. */
function vfpgaConfirmCmdline (helpname, text, strname, ok_func) {

   table := vcTable.Create ("Confirm", 3, 1);
   win := vcRegion.Window (gr_log);
   vcTable.Add (table, vcRegion.Create("", vcStaticText.fac, text));
   r_cmdline := vcRegion.Create("", vfpgaNotEmptyString.fac, "");
   vcTable.Add (table, r_cmdline);
   buttons := vcRegion.Create ("", vcButnRow.fac, list (
      list ("Accept", "Accept", 
            func() ok_func(vcRegion.Get(r_cmdline, vlStringVT));,
            list (close_window <- TRUE)),
      list ("Cancel", "Cancel", func ();, list (close_window <- TRUE)),
      list ("Help", "Help",
         func () vcAlert.SpawnHelp (win,  "vfpga", helpname);,
         list (close_window <- FALSE))));

   vcTable.Add (table, buttons);
   popup := vcWin.Popup (win, table, vcButnRow.GetDefault (buttons),
      vcWin.NoCmdline+vcWin.NoBindings);
   vcWin.Forward (win, popup); }


/*** DATASHEET FACILITY ***/
/*
 * This facility builds a "datasheet" which reads the chip file using
 * the CallChiplist callback defined in pldcb.c.  When a exact part 
 * is selected in the picklist, the datasheet image is called, and
 * the sel_action function passed in to the create routine will be 
 * called with the string in the picklist.  The "name" preference
 * is the basename of the chip file.
 */
package vfpgaDatasheet (_Create, _Get) {

    declare data, t_root;

    function showdata (action) {
        for (i:=0, sl:=CallShowchip(data); !null(sl); sl=cdr(sl),i=i+1) {
            r_temp := vcTable.Get(vcTable.Get(t_root, 1, 0), i, 0);
            vcStaticText.Clear (r_temp);
            vcStaticText.Add (r_temp, car (sl));
        }
        action(data);
    }

    function _Create (sel_action, region) {
        name := vcRegion.PrefGet(region, "name");
        chlist := CallChiplist(name);
        if (chlist != nil) {
            r_picklist := vcPicklist.Scroll("dsscroll", chlist, 10, 10, 10);
            vcPicklist.Notify(r_picklist, func(d,o) {
                if (o == vlOnOff.On) { data = d; showdata(sel_action);}});
        }
        else
            r_picklist := vcRubber.Create("", 0, 0);
        t_datasheet := vcTable.Create("datasheet", 6, 1);
        for (i:=0; i<6; i=i+1) {
            r_temp := vcRegion.Create("", vcStaticText.fac, "",
                                      width <- 60);
            vcRegion.Font(r_temp, vgFont.smallf);
            vcTable.Add (t_datasheet, r_temp);
        }
        t_root = vcTable.Create("PickDevice", 2, 1);
        vcTable.Add(t_root, r_picklist);
        vcTable.Add(t_root, t_datasheet);
        vcRegion.AddChild (region, t_root);
        if (chlist != nil)
            CallChipdefault (name, r_picklist);
        t_root;
    }

    function _Get (arg, region, ts) { data; }

}
vcFac.CreateFromPackage ("vfpgaDatasheet", vscmVT, vfpgaDatasheet);
vlPref.Add (vfpgaDatasheet.prefs, "name", vlStringVT,
   "Element to build datasheet for", nil);
