/*
 *  $Revision: 1.17.1.1 $
 *
 *  Additional ViewScript library functions.
 *
 *  Copyright (c) 1992,1993 by Viewlogic Systems, Inc.
 */

/*\vscmuserSection vscript. -- ViewScript Extensions.
 *
 * DESCRIPTION
 *      These are library functions which provide additional functionality not
 *      related to the core ViewScript functionality.
 *
 *      Most of the function listed belows are C callbacks for the Viewlogic
 *      portable library functions from the vslib and vllib libraries.
 *
 * CAUTION
 *      These functions are not in the Scheme standard and neither they are in
 *      the CFI standard; they are Viewlogic extensions provided mainly for
 *      Viewlogic developers.  Usage of these function will result in
 *      non-portable extension language code to other vendors' frameworks.
 */


/*\vscmuserDocpage vscript.map -- apply a func to each element.

DESCRIPTION
     This is an equivalent of the Scheme's map function.  It applies a
     function (or callback) and applies it to each element in a list.  Unlike
     the Scheme's map, it doesn't work on multiple list arguments.  The
     results are collected in a list and returned to the caller.

NOTE
     This function is similar in its purpose to the `foreach' statement.  The
     difference between them is that `foreach' calls the procedure mainly for
     its side effects, not for its values like the vscript.map function.

EXAMPLE
     function check_numbers(arg)
          isnumber (arg);

     foo := list (1, 'a', 2.3, "no");       --> (1 'a' 2.3 "no")
     vscript.map (check_numbers, foo);      --> (TRUE FALSE TRUE FALSE)

RETURNS
     A list of the results of applying the function argument to each element
     of the argument list.

SEE ALSO
     vscript.list_iter
*/
function map(proc, a_list)
{
    declare result;

    for (; !null (a_list); a_list = cdr (a_list))
        result = cons (proc (car (a_list)), result);

    reverse_destructive (result);

}


/*\vscmuserDocpage vscript.list_iter -- list iterator.

DESCRIPTION
     This iterator is provided as a convenience when using the `foreach'
     statement with lists.  This iterator will iterater over elements of a
     list.

NOTE
     This iterator works only on the top level elements of the argument list.
     If any of the list elements is a list itself, it is not traversed.

EXAMPLE
     foo := list (1, 'b', "c", list (4, 5));

     foreach (i; vscript.list_iter (foo))
         print (i);

     --> 1
     --> 'b'
     --> "c"
     --> (4 5)

SEE ALSO
     vscript.map
*/
iterator list_iter(a_list)
{
    for (; !null (a_list); a_list = cdr (a_list))
        iterate (car (a_list));

}
    
/*\vscmuserDocpage vscript.vect_iter -- vector iterator.

DESCRIPTION
     This iterator is provided as a convenience when using the `foreach'
     statement with vectors.  This iterator will iterater over elements of a
     vector.

NOTE
     This iterator works only on the top level elements of the argument vect.
     If any of the vector elements is a vector itself, it is not traversed.

EXAMPLE
     foo := make_vector (1, 'b', "c", list (4, 5));

     foreach (i; vscript.vect_iter (foo))
         print (i);

     --> 1
     --> 'b'
     --> "c"
     --> (4 5)

SEE ALSO
     vscript.map
*/
iterator vect_iter(a_vect)
{
    n := vector_length (a_vect);
    for (i := 0; i < n; ++i)
        iterate (a_vect[i]);

}
    

/*\vscmuserDocpage vscript.iter -- list/vector iterator

DESCRIPTION
     This iterator is provided as a convenience when using the `foreach'
     statement with lists.  This iterator will iterater over elements of a
     list.

NOTE
     This iterator works only on the top level elements of the argument list.
     If any of the list elements is a list itself, it is not traversed.

EXAMPLE
     foo := list (1, 'b', "c", list (4, 5));

     foreach (i; vscript.list_iter (foo))
         print (i);

     --> 1
     --> 'b'
     --> "c"
     --> (4 5)

SEE ALSO
     vscript.map
*/
iterator iter(l_or_v)
{
    if (istype (l_or_v, vectorVT))
        vect_iter (iterate, l_or_v);
    else
        list_iter (iterate, l_or_v);
}


/*\vscmuserDocpage vscript.defstruct -- define a structure

DESCRIPTION
     This function implements ViewScript structure semantics.  Structure
     fields are implemented as an association list and the structure itself is
     a closure.  This implementation takes advantage of the first class
     property and lexical binding mechanism inherited from Scheme.
*/
function defstruct(this...)
{
    declare structlist, vals, sl, vl;


    structlist = car (this);

    /*
     * Ensure that the value list is the same length as the struct list.
     */
    vl = this;
    for (sl = structlist; !null (sl); sl = cdr (sl))
    {
        if (null (cdr (vl)))
            set_cdr (vl, cons (nil, nil));
        vl = cdr (vl);
    }

    vals = cdr (this);
    
    function get_field(field)
    {
        declare sl, vl;
        
        for (sl = structlist, vl = vals;
             !null (sl) && (field != car (sl));
             sl = cdr (sl), vl = cdr (vl))
            ;
        if (null (sl))
            vlEx.Throw2 (vlEx.FAILURE, "vscript", vlErrSev.Error,
                         sprintf ("Undefined field: %s", field));
        vl;
    }

    func (args...)
    {
        declare vl, field;

        field = car (args);

        /*
         * If there is no second argument, it's a "get".
         */
        if (null (args))
            this;
        else if (null (cdr (args)))
            car (get_field (field));
        else
            set_car (get_field (field), cadr (args));
    };
}

/*
 * Read the Viewdraw search order from the Viewdraw.ini file
 */
function read_viewdraw_search_order() {
    declare cmd, ignore, path;

    vdrawini := fopen (vscript.vsGetini ("viewdraw.ini"), "r");
    path_list := nil;
    for (s := fgetline (vdrawini); s != nil; s = fgetline (vdrawini)) {
        
        /*
         * If it's a directory line, "DIR [r] /pathname/...."
         * parse it and add it to the path list.
         */
        if (((s = string_token (s, &cmd, " \t")) != nil) &&
            (string_eqi (cmd, "dir")) &&
            ((s = string_token (s, &ignore, " \t")) != nil) &&
            (string_token (s, &path, " \t\n") != nil))
            path_list = cons (vscript.vsAbsPath (path), path_list);
    } /* for */
    fclose (vdrawini);
    reverse_destructive (path_list);
}


/*\vscmuserDocpage vscript.delete_list_element -- delete an element from a list

DESCRIPTION
     Returns the resultant list l after removing the first occurrence
     of e.  If there are no occurences of e, the list is returned intact.

CAUTION
     The original list is modified by this function, but you must use
     the return value to handle the case where the first element of the
     list was deleted.

SYNOPSIS
     consVT vscript.delete_list_elemen(l, e)
         consVT l;
         vscmVT e;
*/
function delete_list_element(l, e)
{
    function del(l) {
        if ((l != nil) && ((next := cdr (l)) != nil)) {
            if (car (next) == e)
                set_cdr (l, cdr (next));
            else
                del (next);
        }
    }
        
    if (l == nil)
        nil;
    else if (car (l) == e)
        cdr (l);
    else {
        del (l);
        l;
    }
}


string_search := nil.string_search;

