/*
 * $Revision: 2.4.1.1 $
 *
 *   vcPickEdit facility - a wrapper around vcPicklist for ease of use
 */

/*\vscmuserSection vcPickEdit. - a picklist editor facility
 *
 * DESCRIPTION
 *      vcPickEdit is a wrapper around vcPicklist which is easier to
 *      use, particularly as an editor in dialog boxes.  It allows
 *      scroll bars to be added with a preference, and allows the list
 *      of selections to be modified easily.
 *
 * CREATING A DATA LIST
 *      The data list for vcPickEdit may take the same form
 *      as that of vcPicklist, a ViewScript list of cons cells of
 *      display strings and values:
 *
 *          list (cons ("small", 10), 
 *                cons ("medium", 100),
 *                cons ("large", 1000))
 *
 *      Alternately, the data list may simply be a list of values.
 *      Display strings will be automatically generated by converting
 *      the values to strings:
 *
 *          list (vgColor.Red, vgColor.Blue, vgColor.Green)
 *
 *      These two forms may be mixed:
 *
 *          list (cons (vgColor.White, "Snow"), vgColor.Black)
 *
 * SPECIFYING THE DATA LIST
 *      The data list can be specified in one of three ways.
 *
 *      1.  The `data' preference may be used:
 *
 *          vcWv.Param ("Color", vgColorVT, vcWv.Sticky, vgColor.Red,
 *                      vcPickEdit.fac,
 *                      data <- list (vgColor.Red, ...))
 *
 *      When this method is used, the initialization data (here,
 *      vgColor.Red) is the default value of the picklist.  It may be a
 *      list of values for non-exclusive picklists.
 *
 *      2.  The `data' preference can be used to specify a function that
 *      takes no arguments, and returns the data list:
 *
 *          func () (monochrome) ? list (vgColor.Black, vgColor.White)
 *                               : list (vgColor.Red, vgColor.Blue);
 *
 *      3.  Thirdly, if the 'data' preference is not specified, the
 *      initialization data is used to provide the data list:
 *
 *          vcWv.Param ("Color", vgColorVT, vcWv.Constant, 
 *                      list (vgColor.Red, ...),
 *                      vcPickEdit.fac)
 *
 *      This method allows the list of options to be determined
 *      dynamically using vcWv.Dynamic:
 *
 *          vcWv.Param ("Size", vscmVT, vcWv.Dyanamic, get_sizes,
 *                      vcPickEdit.fac);
 *
 *      When this method is used, the default value is the first
 *      element (exclusive) or all-selections-off (non-exclusive).
 *
 *      Once the picklist is created, the elements in the list may be
 *      changed by modifying the `data' preference with
 *      vcRegion.PrefSet(man). This is true even if the data was not
 *      originally specified with the preference.
 *
 *      Because vcPickEdit is built on top of vcPicklist, a vcPickEdit
 *      region may be passed to many vcPicklist functions, including
 *      vcPicklist.SetPick(man) and vcPicklist.DeselectAll(man).  Do
 *      *not* use vcPicklist.Notify; use vcPickEdit.CallOnChange(man).
 *
 * PREFERENCES
 *      data         - vscmVT   - list of data items, or nil (see above)
 *      exclusive    - boolVT   - exlusive picklist?
 *      scroll       - boolVT   - scrolled picklist?
 *      height       - vlPositiveIntVT - ideal height in rows (if scrolled)
 *
 */

 
function data_to_list (data) {
   pickdata := nil;
   foreach (elem; vscript.list_iter (data))
      if (iscons (elem))
         pickdata = cons (elem, pickdata);
      else
         pickdata = cons (cons (convert (vlStringVT, elem), elem), pickdata);
    reverse_destructive (pickdata);
}

function _Create (dflt, rgn) {

   /* get the data list. call function, if there is one */
   /* use initialization data of 'data' is nil */
   data := (vcRegion.PrefGet (rgn, "data"));
   if (isprocedure (data))
      data = data();
   else if (data == nil) {
      data = dflt;
      dflt = nil;
  }
   
   scroll := vcRegion.PrefGet (rgn, "scroll");
   pickdata := data_to_list (data);

   exclusive := vcRegion.PrefGet (rgn, "exclusive");
   p := vcRegion.Create ("", vcPicklist.fac, pickdata, 
                      will_scroll <- scroll,
                      exclusive <- exclusive,
                      enhanced_notify <- TRUE);
   height := vcRegion.PrefGet (rgn, "height");

   if (scroll)
      s := vcPicklist.Scroll ("", p, (height > 4) ? 4 : 1, height, 100);
   else
      s := p;

   /* set default value; if none, set the first one */
   if (dflt != nil) {
      /* dflt may be a value, or may be a list of values */
      foreach (row, col, ix, isset, vis, data; vcPicklist.Items (p))
         if (dflt == data || 
                (!exclusive && iscons (dflt) && member (data, dflt)))
            vcPicklist.SetPick (p, ix);
   } else if (exclusive)
      vcPicklist.SetPick (p, 0);

   vcRegion.AddChild (rgn, s);

   vcPicklist.Notify (p, func (data, flg, ix)
      if (pe.call_on_change != nil && !pe.inhibit_notify)
         pe.call_on_change (data, flg, ix););

   pe := struct {
      pregion = p;
      call_on_change = nil;
      inhibit_notify = FALSE   /* see _Reset for explanation */
   };
}

function _Get (p, r, ts)
   vcRegion.Get (p.pregion, ts);

function _Reset (p, rgn) {
   declare x, y;

   if (vcRegion.PrefChanged (rgn, "data") &&
                    vcRegion.HasBeenReframed(rgn)) {

      vcPicklist.Dimensions (p.pregion, &x, &y);

      /* remember who has been set */
      setlist := nil;
      foreach (row, col, ix, isset, visual, data; vcPicklist.Items (p.pregion))
          if (isset)
             setlist = cons (data, setlist);
      vcPicklist.DeleteRows (p.pregion, 0, x);

      pickdata := data_to_list (vcRegion.PrefGet (rgn, "data"));

      foreach (elem; vscript.list_iter (pickdata))
         vcPicklist.Add (p.pregion, car (elem), cdr (elem));

      /* we are just resetting things that are already set, but */
      /* have been removed and replaced. Don't notify about those */
      p.inhibit_notify = TRUE;
      /* now set the ones that used to be set */
      foreach (row, col, ix, isset, visual, data; vcPicklist.Items (p.pregion))
         if (member (data, setlist))
             vcPicklist.SetPick (p.pregion, ix);
      p.inhibit_notify = FALSE;

   }
 }
      

/*\vscmuserDocpage vcPickEdit.CallOnChange - register function to call on change
 *
 * SYNOPSIS
 *      vlVoidVT vcPickEdit.CallOnChange (r, proc)
 *          vcRegionVT r;
 *          closureVT proc;
 *
 * DESCRIPTION
 *      This function allows a function to be registered that will be
 *      called whenever the selection changes.
 *
 *      The function should take the following arguments:
 *
 *          proc (vscmVT element_data, vlOnOffVT flag, vIntVT index)
 */
function CallOnChange (r, proc) {
   p := vcRegion.GetData (r);
   p.call_on_change = proc;   
}

vlPref.Replace (prefs, "data", vscmVT, "~~data for the picklist", nil);
vlPref.Replace (prefs, "height", vlPositiveIntVT,
               "~~ideal height of list", 6);
vlPref.Replace (prefs, "exclusive", boolVT,
                       "~~exclusive semantics?", TRUE);
vlPref.Replace (prefs, "scroll", boolVT,
                       "~~scrolled picklist?", TRUE);




















