/*
 * $Revision: 1.2.1.2 $
 */

/***************************************************************************/
/*  Functions dealing the invoking text editors follow                     */
/*                                                                         */
/*      Each supported editor provides an "invoke" function that takes     */
/*      the same five arguments:                                           */
/*              window          -- nil, or used for status messages, etc   */
/*              filename        -- file to edit                            */
/*              line_number     -- nil, or line number (start at 1).       */
/*              call_on_terminate -- nil, or func(mode, code) (vcSys.Command)*/
/*              read_only_flag  -- boolean to bring up file in read-only mode*/
/*                                                                         */
/*      The invoke function and a "capabilities mask" are put into a       */
/*      structure, vcWinTextEditorVT, which is used to select and invoke   */
/*      the appropriate text editor for a task.                            */
/*      A hash table maps an enumeration value (vcWvTextEditorVT) into the */
/*      editors.  The user selects the desired text editor based on a      */
/*      preference.  There's a different preference for MS Windows and     */
/*      unix, with different choices available, so the two can co-exist    */
/*      in one powerview.ini file.                                         */
/*                                                                         */
/***************************************************************************/


struct vcWinTextEditorVT {
    vlTrueFalseVT can_goto_line;
    vlTrueFalseVT can_attach_to_existing;
    vscmVT invoke_function;
};

/* maps vcWvTextEditorVT (from preference) to vcWinTextEditorVT */
text_editors := vlSym.ObjCreate (5);

/*
 * First set up the editors common to all platforms: ViewText and
 * whatever the user has in $EDITOR.
 */

/*
 * Invoke ViewText, trying to find existing window first, jumping to line
 * number if supplied, and calling call_on_terminate when the ViewText
 * region is closed, if one is supplied.  Note that we call call_on_terminate
 * with mode and code just as if it was invoked as a subprocess.
 */
function invoke_vtext(window, filename, line_number, call_on_terminate,
                      read_only_mode)
{
    function setup_viewtext(vt, update) {
        if (line_number != nil) {
            vtx.SelectLine (vt, line_number);
            if (update)
                vtx.Update (vt);
        } /* if */
        if (call_on_terminate != nil)
            vcRegion.DestroyNotify (vt, func(r) {
                call_on_terminate (vcSysMode.Exited, 0);
            });
    } /* function setup_viewtext */

    /*
     * See if there's already a ViewText window open on that file
     */
    foreach (w; vcWin.Loop ()) {
        reg := vcWin.OwnerRegion (w);
        if ((vcRegion.Fac (reg) == vtx.fac) &&
            string_eq (vcWin.Title (w), filename))
        {
            vcWin.Show (w);
            setup_viewtext (reg, TRUE);
            return (nil);
        } /* if */
    } /* foreach */

    if (window != nil)
        vcWin.AddStatus (window, "Invoking ViewText on " + filename);
    vt := vcRegion.Create (filename, vtx.fac, filename, load_file <- TRUE);
    if (read_only_mode != nil)
        vtx.ReadOnly (vt, read_only_mode);
    w := vcWin.Create (vt);
    setup_viewtext (vt, FALSE);
    w;
} /* function invoke_vtext */

/*
 * Call vcWin.System but make it invisible to the Quit dialog.  None
 * of the text editors will be using Core's I/O redirection so we don't
 * need to bug the user about quitting with them active.
 */
function sys_ignore(win, cmd, env, wait, output, error_proc, completion_proc,
                    cancel_proc, host)
{
    sys := vcWin.System (win, cmd, env, wait, output, error_proc,
                         completion_proc, cancel_proc, host);
    vcQuit.IgnoreSubprocess (sys);
    sys;
} /* function sys_ignore */

/*
 * Invoke the editor selected by the users $EDITOR environment variable.
 */
function invoke_dollar_editor(window, filename, line_number, call_on_terminate,
                              read_only_mode)
{
    editor_name := getenv ("EDITOR");
    if (editor_name == nil)
        invoke_viewtext (window, filename, line_number, call_on_terminate);
    else
        sys_ignore (window, editor_name + " " + filename, nil, FALSE,
                    vcSysOutput.TerminalPause, nil, call_on_terminate,
                    nil, nil);
} /* function invoke_dollar_editor */

declare text_editor_pref;

/*
 * Now set up the Unix-specific ones:   emacs and vi.  vi is a little
 * special cause it knows about line numbers on invocation, whereas with
 * $EDITOR we don't.
 */
if (vcWv.OS (wv) == vcOs.Unix) {
    text_editor_pref = "unix_text_editor";
    define_enum ("vcWvTextEditorVT",
                 list ("ViewText", "emacs", "vi", "$EDITOR"));
    vlPref.Replace (vcWv.prefs, "emacs_name", vlStringVT,
                    "Name executable to run to invoke emacs",
                    "emacs");
    vlPref.Replace (vcWv.prefs, "emacsclient_name", vlStringVT,
                    "Name executable to run to invoke emacsclient",
                    "emacsclient");

    /*
     * The way emacsclient works, it doesn't return until the user hits
     * ^X #, which they might not be apt to do while looping through
     * errors.  So the way I'm going to do this is, when the caller
     * gives "nil" as the call_on_terminate function, we'll only allow
     * one "emacsclient" subprocess to remain active at once.  So we
     * need a "static" variable to hold the emacsclient sys ID.
     */
    emacsclient_sys_id := nil;

    function invoke_emacs(window, filename, line_number, call_on_terminate,
                          read_only_mode)
    {
        declare client_error_text;
        cmd := filename;

        if (line_number != nil)
            cmd = sprintf ("+%d %s", line_number, cmd);

        if (emacsclient_sys_id != nil)
            vcSys.Kill (emacsclient_sys_id, vcSysKill.Normal);

        function emacsclient_failed(mode, code) {
            function cancel() {
                if (call_on_terminate != nil)
                    call_on_terminate (mode, code);
            } /* function cancel */

            function start_emacs() {
                emacs_name := vlPref.Val (vcWv.prefs, "emacs_name");
                sys_ignore (window,
                            emacs_name + " -f server-start " + cmd,
                            nil,
                            FALSE, vcSysOutput.NoOutput, nil,
                            call_on_terminate, nil, nil);
            } /* function start_emacs */

            buttons :=
                vcRegion.Create ("buttons",
                                 vcButnRow.fac,
                                 list (list ("start_emacs", "Invoke new Emacs",
                                             start_emacs,
                                             list (close_window <- TRUE)),
                                       list ("retry",
                                             "Retry connecting to existing Emacs",
                                             try_emacsclient, 
                                             list (close_window <- TRUE)),
                                       list ("cancel", "Cancel", vcWin.wclose,
                                             list (close_window <- TRUE))));

            vcWv.Dialog (window,
                         "Emacs connection failed",

"emacsclient could not connect to existing emacs:

" + client_error_text + "

This may occur because Emacs is not running, or there
may be an Emacs running already, but is not running
in \"server mode\".  You can place an existing emacs
window in server mode by doing

    Alt-X server-start

in the window, or you can enable it permanently by
putting

    (server-start)

in your .emacs file.",

                         "How would you like to proceed?",
                         buttons);
        } /* function emacsclient_failed */

        function try_emacsclient() {
            client_error_text = "";
            function error_handler(str)
                client_error_text += str;

            emacsclient_name := vlPref.Val (vcWv.prefs, "emacsclient_name");

            /*
             * Try to use emacsclient, which uses an existing emacs window.
             * If it fails with mode Exited, status 1, there was probably not
             * an emacs server, so start emacs with "-f server-start" to get
             * one going.
             */
            sys := sys_ignore (window, emacsclient_name + " " + cmd,
                               nil, FALSE, error_handler, error_handler,
                               func(mode, code) {
                                   if (sys == emacsclient_sys_id)
                                       emacsclient_sys_id = nil;
                                   if ((mode == vcSysMode.Exited) &&
                                       (code == 1))
                                       emacsclient_failed (mode, code);
                                   else if (call_on_terminate != nil)
                                       call_on_terminate (mode, code);
                               }, /* func */
                               nil, nil);
            if (call_on_terminate == nil)
                emacsclient_sys_id = sys;
            sys;
        } /* function try_emacsclient */

        try_emacsclient ();
    } /* function invoke_emacs */

    function invoke_vi(window, filename, line_number, call_on_terminate,
                       read_only_mode)
    {
        if (line_number != nil)
            filename = sprintf ("+%d %s", line_number, filename);

        sys_ignore (nil, "vi " + filename, nil, FALSE,
                    vcSysOutput.TerminalPause, nil, call_on_terminate,
                    nil, nil);
    } /* function invoke_vi */
    text_editors[vcWvTextEditor.emacs] =
        struct vcWinTextEditorVT {TRUE, TRUE, invoke_emacs};
    text_editors[vcWvTextEditor.vi] =
        struct vcWinTextEditorVT {TRUE, FALSE, invoke_vi};
} /* if */

/*
 * MS windows-specific text editors:  Notepad and Write.  I haven't
 * filled these in yet.
 */
else {
    text_editor_pref = "ms_text_editor";
    define_enum ("vcWvTextEditorVT",
                 list ("ViewText", "Notepad", "Write", "$EDITOR"));
    vlPref.Replace (vcWv.prefs, "ms_text_editor", vcWvTextEditorVT,
                    "Preferred text editor (note: not yet used universally)",
                    vcWvTextEditor.ViewText);

    function invoke_notepad(window, filename, line_number, call_on_terminate,
                            read_only_mode)
    {
        error ("How do I invoke notepad?", filename);
        /* Note, the vcSysID used to invoke notepad should be returned
           if possible */
    } /* function invoke_notepad */

    function invoke_write(window, filename, line_number, call_on_terminate,
                          read_only_mode)
    {
        error ("How do I invoke MS Write?", filename);
        /* Note, the vcSysID used to invoke write should be returned
           if possible */
    } /* function invoke_write */
    
    text_editors[vcWvTextEditor.Notepad] =
        struct vcWinTextEditorVT {FALSE, FALSE, invoke_notepad};
    text_editors[vcWvTextEditor.vi] =
        struct vcWinTextEditorVT {FALSE, FALSE, invoke_write};
} /* else */

text_editors[vcWvTextEditor.ViewText] =
    struct vcWinTextEditorVT {TRUE, TRUE, invoke_vtext};
text_editors[convert (vcWvTextEditorVT, "$EDITOR")] =
    struct vcWinTextEditorVT {FALSE, FALSE, invoke_dollar_editor};

vlPref.Replace (vcWv.prefs, text_editor_pref, vcWvTextEditorVT,
                "Preferred text editor (note: not yet used universally)",
                vcWvTextEditor.ViewText);
vlPref.Parse ("powerview.ini", wv);

/*\vscmuserDocpage vcWin.TextEdit -- bring up a text editor on a file

SYNOPSIS
     ret := vcWin.TextEdit (window,             nil or source window for status
                            filename,           Pathname of file to edit
                            line_number,        nil or number of start line
                            call_on_terminate,  nil or function to call on exit
                            attach_to_existing, boolean: require a text editor
                                                that can pop up an existing
                                                invocation and go to a new line
                                                number instead of making a new
                                                window on each call.
                            read_only_mode)     Try to bring up read only

DESCRIPTION
     Invokes a text editor on `filename', based on user preference,
     text editor capabilities, and operating system.  If `line_number'
     is non-nil, a text editor will be selected that can be positioned
     to a specific line when invoked.  If call_on_terminate is non-nil,
     it will be called with a "mode" and "code" when the editor exits
     (see vcSys.Command).  If `attach_to_existing' is TRUE, an editor
     will be selected that can be sent a message saying to go to a
     new filename and line number, without necessarily bringing up a
     new window.  The intention here is to support "Next Error"
     functionality, where you want to cycle through several lines in a
     single file, without bringing up a new window each time.

     Modulo the capabilities of the editors, we will select the editor
     based on the users preference "vcWv.ms_text_editor" or
     "vcWv.unix_text_editor", depending on platform.  Right
     now if you need either line_number or attach_to_existing
     capability, you can use either ViewText or Emacs.  If the user
     selects another editor and you need these capabilities, you'll
     get ViewText.

RETURNS
     If ViewText is used, a vcRegionVT is returned.  In all other cases,
     a vcSysID may be returned, if possible.  In some cases, that may
     not be possible.  The type of the returned object must be tested.
     It is best not to rely on this for critical functionality.

WARNING
     When implementing next_error functionality, call_on_terminate should
     be specified as nil, otherwise if the user selects emacs we will
     wind up leaving a subprocess around for each invocation of TextEdit
     until the user hits "Ctrl+X #" in emacs.  This is due to the way that
     emacsclient works.
*/
function TextEdit(window,               /* nil or source window for status */
                  filename,             /* Pathname of file to edit */
                  line_number,          /* nil or number of line to warp to */
                  call_on_terminate,    /* nil or function to call on exit */
                  attach_to_existing,   /* boolean: require a text editor
                                           that can pop up an existing
                                           invocation and go to a new line
                                           number instead of making a new
                                           window on each call. */
                  read_only_mode)       /* Try to bring up in read only mode */
{
    editor := vlPref.Val (vcWv.prefs, text_editor_pref);

    if (window != nil)
        vcWin.ClearStatus (window);

    if ((line_number != nil) && !text_editors[editor].can_goto_line) {
        if (window != nil)
            vcWin.AddStatus (window,
                             sprintf ("Editor `%s' cannot position to a line.  ",
                                      convert (vlStringVT, editor)));
        editor = vcWvTextEditor.ViewText;
    } /* if */
    else if (attach_to_existing &&
             !text_editors[editor].can_attach_to_existing)
    {
        if (window != nil)
            vcWin.AddStatus (window,
                             sprintf ("Editor `%s' cannot pop an existing window.  ",
                                      convert (vlStringVT, editor)));
        editor = vcWvTextEditor.ViewText;
    } /* else if */

    text_editors[editor].invoke_function (window, filename, line_number,
                                          call_on_terminate, read_only_mode);
} /* function TextEdit */

ViewTextCmd := vcWv.Cmd (fac, func (win, ev, fname) {
    TextEdit (win, fname, nil, nil, FALSE, FALSE);
},
                         "wtext", "Edit Text File",
                         vcWv.Param ("", vlStringVT, vcWv.Sticky, "",
                                     vcFileSelect.fac));
vcCmd.EnableApply (ViewTextCmd, vlOnOff.On);
vcCmdtable.Add (vcWv.commands, "vtext", ViewTextCmd);


