/*
 * $Revision: 1.5.2.1 $
 *
 * Functions originally written by Joshua Marantz
 * Awesome man page for vitc.CntxMake by Ken Cox.
 */

/*\vscmuserDocpage vitc.CntxMake -- utility function for creating contexts
 *
 * SYNOPSIS
 *      vitcCntxVT vitc.CntxMake(args...)
 *
 * DESCRIPTION
 *      Returns an ITC context which is completely wildcarded,
 *      except for customizations specified in the parameter list,
 *      and except for the Framework ID, which is the default.
 *
 *      The customizations are specified using the `field <- value'
 *      syntax.  Each such customization is specified as a separate
 *      parameter to vitc.CntxMake.  See the examples.
 *
 * VALID FIELD NAMES
 *      Field Name  Default           Meaning
 *      ----------  -------           -----------------
 *      tid         vitc.TidANY       any tool ID ("*")
 *      name        vitc.NameANY      any name
 *      fid         vipc.FidDFLT      default framework
 *      hnm         vipc.HnmANY       any hostname
 *      uid         vipc.UidANY       any user ID
 *      gid         vipc.GidANY       any group ID
 *      pid         vipc.PidANY       any process ID
 *      cid         vipc.CidANY       any channel ID
 *
 * EXAMPLE 1, A GENERIC MATCH CONTEXT
 *      One needs a "match context" to use during ITC registration.
 *      This context specifies the tools from which we are willing
 *      to receive messages.  One would generally want the limit to
 *      be "any tool within my framework".  In this case, one would
 *      use the generic match context with no customizations:
 *
 *          match_cntx := vitc.CntxMake ();
 *
 * EXAMPLE 2, A SIMPLE DESTINATION CONTEXT
 *      When creating a "destination context" for use with
 *      vitc.Send, one generally wants to wildcard everything except
 *      the tool ID and perhaps the name field.  To send a message
 *      to any tool called "MyTool", one would use:
 *
 *          dest_cntx := vitc.CntxMake (tid <- "MyTool");
 *
 * EXAMPLE 3, A COMPLEX DESTINATION CONTEXT
 *      Perhaps the tool "MyTool" takes full advantage of the name
 *      field, and creates one ITC tool for each design component on
 *      which it is operating.  This would mean that if "MyTool"
 *      were operating on design "amp" and design "pre-amp"
 *      simultaneously, it would have created two ITC tools.  These
 *      two tools would have the same tool ID and different name
 *      fields.  To distinguish between them, one might use:
 *
 *          dest_cntx := vitc.CntxMake (tid <- "MyTool",
 *                                      name <- "amp");
 */
function CntxMake(args...)
{
    tid  := vitc.TidANY;  /* must be vitc not vipc */
    name := vitc.NameANY; /* or nil; must also be vitc not vipc  */
    fid  := vipc.FidDFLT;
    hnm  := vipc.HnmANY;
    uid  := vipc.UidANY;
    gid  := vipc.GidANY;
    pid  := vipc.PidANY;
    cid  := vipc.CidANY;

    foreach (arg; vscript.list_iter (args))
    {
        a := car (arg);
        v := cdr (arg);
        if      (a == `gid)     gid = v;
        else if (a == `pid)     pid = v;
        else if (a == `cid)     cid = v;
        else if (a == `uid)     uid = v;
        else if (a == `fid)     fid = v;
        else if (a == `hnm)     hnm = v;
        else if (a == `tid)     tid = v;
        else if (a == `name)    name = v;
	else error("Invalid field specifier", a);
    }
    vitc.CntxCreate (gid, uid, fid, tid, name, hnm, pid, cid);

} /* CntxMake */


function CntxMake51(args...)
{
    tid  := vipc.TidANY;
    fid  := vipc.FidDFLT;
    hnm  := vipc.HnmANY;
    uid  := vipc.UidANY;
    gid  := vipc.GidANY;
    pid  := vipc.PidANY;
    cid  := vipc.CidANY;

    foreach (arg; vscript.list_iter (args))
    {
        a := car (arg);
        v := cdr (arg);
        if      (a == `gid)     gid = v;
        else if (a == `pid)     pid = v;
        else if (a == `cid)     cid = v;
        else if (a == `uid)     uid = v;
        else if (a == `fid)     fid = v;
        else if (a == `hnm)     hnm = v;
        else if (a == `tid)     tid = v;
	else error("Invalid field specifier", a);
    }
    vipc.CntxCreate (gid, uid, fid, tid, hnm, pid, cid);

} /* CntxMake51 */

_RunOnceSrcContext := nil;
_RunOnceDestContext := nil;

/*\vscmuserDocpage vitc.RunOnce -- Create a kind of window exactly once, within a context
 * 
 * SYNOPSIS
 *      vitc.RunOnce (vlStringVT tool_name, vlStringVT data_name,
 *                    closureVT create_window_function);
 * 
 * DESCRIPTION
 *      It may be desirable to be able to invoke an application from
 *      any of several executables within a framework, but to only allow
 *      one of them to exist on the screen at once.  We must use ITC
 *      to detect the existence of another invocation of that tool,
 *      and send it a request to pop itself instead of creating a new
 *      instance.
 * 
 *      `tool_name' uniquely identifies the tool within the framework,
 *      although the tool can be invoked many times with different data.
 *
 *      `data_name' uniquely identifies the data for this invocation of
 *      the tool.
 *
 *      `create_window_function' is only called if an ITC server for
 *      the specified tool cannot be found.  If it can be found, the
 *      server will be instructed to pop the window attached to it.
 *      If not, the function is called with no arguments, and is
 *      expected to return a vcWindowVT.  When the window is destroyed,
 *      the server closes down and the next call to vitc.RunOnce will
 *      cause the window to be created.
 * 
 * EXAMPLE
 *      function run_my_facility() {
 *          r := vcRegion.Create ("my_data", vcLabel.fac, "my_data");
 *          vcWin.Create (r);
 *      }
 *      
 *      vitc.RunOnce ("my_facility", "my_data", run_my_facility);
 */
function RunOnce(tool_name,
                 data_name,
                 create_window_function)
{

    if (_RunOnceSrcContext == nil) {
        _RunOnceSrcContext = SrcCntxCreate ("_RunOnceClient", "");
        _RunOnceDestContext = CntxMake (tid <- "_RunOnceClient",
                                        name <- "");
    } /* if */

    server_dest := CntxMake (tid <- tool_name, name <- data_name);
    if (Send ("_RunOnceShow", _RunOnceSrcContext, server_dest,
                   "", vipc.FlgDIRECT)
        != Success)
    {
/*        printf ("RunOnce: send failed\n"); */
        window := create_window_function ();
        server := SrcCntxCreate (tool_name, data_name);
        handler := Register ("_RunOnceShow", server, _RunOnceDestContext,
                             func(msg, src, match, data) {
/*                                 printf ("RunOnce: received show for %s\n",
                                         window); */
                                 vcWin.Show (window);
                             });
        vcWin.DestroyNotify (window, func(w) {
            window = nil;
            vitc.Unregister (handler);
            vitc.SrcCntxDestroy (server);
        });
    } /* if */
/*
    else
        printf ("RunOnce: send succeeded\n"); */
} /* function RunOnce */
