/*
 *  $Revision: 1.66.1.20 $  $Date: 1995/05/16 17:46:55 $
 *
 *  VDM encapsulation file for the RCS storage manager.
 *
 *
 *  Note on the endProc function
 *  ----------------------------
 *      Because storage manager integrators can use asynchronous functions in
 *      their implementations of the SM functions, VDM cannot presume that
 *      when a _vdm[...] method returns, the SM function has been completed.
 *      You must call the provided endProc function to signal to the VDM
 *      backplane that the SM function has indeed terminated.
 *
 *      In addition, the SM backplane has to be completely reentrant and allow
 *      many requests from different applications to be executed at the same
 *      time without tramping over each other's data.  To this end, each
 *      method is passed in an `endArg' which uniquely identifies the thread
 *      of execution for this particular method invocation.  When a method
 *      terminates and wants to notify VDM, you must call the endProc function
 *      with two arguments: the data object and the endArg:
 *
 *          endProc ( obj, endArg, user_arg );
 *
 *      If you don't call the endProc function, VDM will keep waiting for this
 *      event to occur and will eventually time out.
 *
 *  Returning a value from a _vdm method
 *  ------------------------------------
 *      Note that the endProc function takes three arguments.  The `user_arg'
 *      argument allows the integrator to pass back a return value from the
 *      integration method.  Remember, the _vdm[...] methods return
 *      immediately, but that doesn't mean that the method has finished
 *      execution.  That's why any actual return value from a _vdm function
 *      will be ignored.  But the, what if you really want to return
 *      something?  Well, it's not as straightforward, but it works.  What you
 *      do, is you specify your return value as the third argument to the
 *      endProc function.  That's the `user_arg' argument.  That argument will
 *      be converted into an attribute, which will be added to the data
 *      object.  The name of the attribute is "_vdmReturnValue".  Because
 *      attributes are generic, no restrictions are placed on the type of the
 *      "return" value.
 *
 *  Example of a _vdm method return value
 *  -------------------------------------
 *      Suppose that for a given SM encapsulation, the integrator wants the
 *      _vdmRegister method to return the name of the data object as it is
 *      stored by the storage manager.  We are assuming that the name of the
 *      data object and the name of its representation in the SM are different
 *      (like RCS, which creates ,v files).
 *
 *          function _vdmRegister(obj, endProc, endArg, smCntx)
 *          {
 *              // Do the registration
 *              ...
 *              // Get the new name
 *              new_name := some_extract_func (vdm.ObjGetPhysicalName (obj));
 *
 *              // Signal the end of the method and return the new name.
 *              endProc (obj, endArg, new_name);
 *          }
 *
 *
 *          Now, in your application, the code would look something like this:
 *
 *          foo := vdm.ObjCreate ("YourObjName", "YourClass", location,
 *                                your_vdm_cntx, nil);
 *          vdm.ObjRegister (foo);
 *          new_foo_name := vdm.ObjGetAttr (foo, "_vdmReturnValue");
 *
 *
 *      You might wander why all these complications, but the answer is quite
 *      obvious once you consider the fact that VDM is asynchrounous and that
 *      the endProc function might be called long after the method's function
 *      has returned.  It's the user that must tell VDM when the backplane
 *      methods terminate and when that happens, there's very little context
 *      information left to work with.  The `obj' and `endArg' arguments help
 *      to alleviate this problem and allow VDM to support multiple execution
 *      threads at the same time.
 *
 *  What's a storage manager context?
 *  ---------------------------------
 *      Some storage manager want to communicate state information between
 *      methods.  The _vdmCreate method returns as its value the storage
 *      manager context.  It's whatever you wish it to be.  That value is
 *      passed in to all the methods.  You make the use of it: VDM serves only
 *      as a conduit to pass it along to all the methods.
 *
 *  Currently supported methods
 *  ---------------------------
 *      _vdmInit                    - general
 *      _vdmCreate                  - general
 *      _vdmGeneric                 - general
 *
 *      _vdmCheckin                 - SM
 *      _vdmCheckoutLock            - SM
 *      _vdmCheckoutNoLock          - SM
 *      _vdmDelete                  - SM
 *      _vdmFind                    - SM
 *      _vdmGetPhysicalDescriptor   - SM
 *      _vdmLock                    - SM
 *      _vdmLoop                    - SM
 *      _vdmMetaDataName            - SM [optional]
 *      _vdmQuery                   - SM
 *      _vdmRegister                - SM
 *      _vdmStatus                  - SM
 *      _vdmUnlock                  - SM
 *
 *      _vdmOpen                    - I/O
 *      _vdmClose                   - I/O
 *      _vdmRead                    - I/O
 *      _vdmReadChar                - I/O
 *      _vdmReadLine                - I/O
 *      _vdmWrite                   - I/O
 *      _vdmWriteChar               - I/O
 *      _vdmWriteLine               - I/O
 *
 *  Copyright (c) 1994,1995 by Viewlogic Systems, Inc.
 */


/*
 * Load helpful RCS manipulation routines.
 */
load ("rcslib.vs");


/*
 * Time stamp attribute name.
 */
vdmRCS_STAMP_ATTR := "_vdmRCSStamped";



/*\Docpage _vdmInit -- Initialize the encapsulation.
 * 
 * DESCRIPTION
 *      This function is called only once by the VDM.  It's purpose is to
 *      initialize this particular storage manager encapsulation.  Note that a
 *      second initialization method in this file, _vdmCreate, is used to bind
 *      the storage manager encapsulation to a particular location.
 *      
 *      In addition, this function must return a structure which will be used
 *      to prompt and to parse for the location's string.  For more
 *      information about the ViewScript structures, see struct(man).
 *
 * NOTE
 *      The difference between this method and the _vdmCreate method is that
 *      the _vdmInit method is called only once per storage encapsulation.
 *      The _vdmCreate method is called for each location managed by this
 *      storage manager.
 *
 * SEE ALSO
 *      struct
 */
function _vdmInit(obj, endProc, endArg, ignore)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmInit.\n");

    path_struct := struct vdmRCSVT
        {
            vlStringVT  storage_area;
            vlStringVT  working_area;
            vdmDbVT     _vdmdb;
        };


    vdm.CntxCallOnError (vdm.ObjGetCntx (obj), get_the_version);

    endProc (obj, endArg, path_struct);
}



/*\Docpage _vdmCreate -- Instantiate the SM encapsulation.
 *
 * DESCRIPTION
 *      The _vdmCreate method will be called when the VDM requests the first
 *      storage operation performed on a data object in the given location.
 *      Any SM request can trigger instantiation.  The instantiation of the
 *      storage manager encapsulation is postponed until the VDM actually
 *      wants to use it in order not to slow down the initialization of
 *      applications built on top of the VDM.
 *
 *      This method returns the storage manager's context.  The context
 *      represents the particular instantiation of this storage manager
 *      encapsulation for a particular location.  The same storage manager can
 *      be used to manage another location, but VDM will create two separate
 *      instances of the storage manager encapsulation for the two lcoations.
 *
 * NOTE
 *      The difference between this method and the _vdmInit method is that the
 *      _vdmInit method is called only once per storage encapsulation.  The
 *      _vdmCreate method is called for each location managed by the given
 *      storage manager.
 *
 * CAUTION
 *      The object passed in to this function is a generic data object managed
 *      by the VDM library.  It has no relation whatsoever to any design data
 *      objects and it is used only by VDM.
 */
function _vdmCreate(obj, endProc, endArg, ignore, struct_type, location_string)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmCreate: %s.\n", location_string));

    /*
     * Parse the path string into two tokens: one for the storage path and the
     * other for the working path.  If there are any problems, rais an
     * exception.
     */
    declare storage_area, working_area;

    working_area = string_token (location_string, &storage_area, " ");

    if ((storage_area == nil) || (string_eq (storage_area, ""))  ||
        (working_area == nil) || (string_eq (working_area, "")))
        vlEx.Throw2 (vlEx.FAILURE, "vdmRCS", vlErrSev.Error,
                 sprintf ("Location string `%s' for vdmRCS must contain both a
storage and working area specification.", location_string));


    /*
     * Note the addition of the '/' slash character.  The user might or might
     * not terminate the pathnames with a slash, so we just make sure that
     * it's always there.  The vscript.vsAbsPath function will remove
     * duplicates if the pathname already ends with a slash.
     */
    path_struct :=
        struct vdmRCSVT
        {
            vscript.vsAbsPath (string_trim (storage_area) + "/"),
            vscript.vsAbsPath (string_trim (working_area) + "/"),
            vdm.DbCreate (current_package ())
        };


    if ((path_struct.storage_area == nil) || (path_struct.working_area == nil))
        vlEx.Throw2 (vlEx.FAILURE, "vdmRCS", vlErrSev.Error,
                     sprintf ("Illegal pathname specifications--`%s'.",
                              location_string));

    /*
     * The context for the RCS encapsulation is the path structure, because it
     * contains both the storage and working areas.
     */
    endProc (obj, endArg, path_struct);
}



/*\Docpage _vdmCheckoutLock -- Checkout with a lock.
 *
 * DESCRIPTION
 *      Data object is checked out from the storage manager and a "lock" is
 *      requested.
 *
 * NOTE
 *      The term "lock" is understood by the VDM as an exclusive right to
 *      modify the data object without any other user being able to
 *      _vdmCheckin any modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmCheckoutLock(obj, endProc, endArg, rcsCntx)
{
    cntx := vdm.ObjGetCntx (obj);
    vers := handle_version (obj);
    <storage_path, dest_path> := build_pathnames (obj, rcsCntx);


    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        if (is_new_meta ())
        {
            central2local_exit_handler (mode, code, obj, endProc, endArg,
                                        rcsCntx, nil);
            record_working_version (rcsCntx, obj, vers, storage_path);
        }
        else    
        {
            common_exit_handler (mode, code, obj, endProc, endArg, nil);
            record_working_version (rcsCntx, obj, vers, storage_path);
        }


    create_dest_dir (dest_path);

    /* Note that the revision label must be flush with the -l option. */
    cmd := "co -q -f -M -l" + vers;
    cmd  = cmd + " " + storage_path + " " + dest_path;


    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmCheckoutLock: %s\n", cmd));


    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmCheckoutNoLock -- Check out without a lock.
 * 
 * DESCRIPTION
 *      Retrieve a revision of the data object from the storage manager, but
 *      don't lock it.  This operation is commonly referred to as a "fetch".
 *
 *      With no other information provided by the user, the checked out
 *      revision should be the latest one.
 *
 *      Because there's no lock requested, other users can check in new
 *      versions of this data object.  For this reason, the "fetch" operation
 *      should only be done for read-only purposes.  If the user intends to
 *      modify the object, a lock should be requested via the _vdmCheckoutLock
 *      method.
 *
 * NOTE
 *      The term "lock" is understood by the VDM as an exclusive right to
 *      modify the data object without any other user being able to
 *      _vdmCheckin any modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmCheckoutNoLock(obj, endProc, endArg, rcsCntx)
{
    cntx                      := vdm.ObjGetCntx (obj);
    <storage_path, dest_path> := build_pathnames (obj, rcsCntx);
    vers                      := handle_version (obj);

    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        if (is_new_meta())
        {
            central2local_exit_handler (mode, code, obj, endProc, endArg,
                                        rcsCntx, nil);
            record_working_version (rcsCntx, obj, vers, storage_path);
        }
        else
        {
            /* Need to keep metadata objects writeable. */
            if (vdm.CntxFindClass (cntx, "vdmMetaData") ==
                vdm.ObjGetClass (obj))
            vdm.UtilChmod (dest_path, "wr", "r", "r");
            record_working_version (rcsCntx, obj, vers, storage_path);
            common_exit_handler (mode, code, obj, endProc, endArg, nil);
        }

    create_dest_dir (dest_path);

    cmd := "co -q -f -M -r" + vers;
    cmd  = cmd + " " + storage_path + " " + dest_path;

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmCheckoutNoLock: %s\n", cmd));

    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmCheckin -- Check in a data object.
 *
 * DESCRIPTION
 *      Store a new revision of the data object in the storage manager.  If
 *      the user had a lock on the object, it should be relinquished.
 *
 * NOTE
 *      The term "lock" is understood by the VDM as an exclusive right to
 *      modify the data object without any other user being able to
 *      _vdmCheckin any modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
this_new_version := 0;
this_old_version := 0;
function _vdmCheckin(obj, endProc, endArg, rcsCntx, log_message)
{
    cntx := vdm.ObjGetCntx (obj);



    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        if (is_new_meta())
        {
            local2central_exit_handler (mode, code, obj, endProc, endArg,
                                        rcsCntx, this_new_version,
                                        this_old_version, nil);
            remove_working_version (rcsCntx, obj);
        }
        else
        {
            common_exit_handler (mode, code, obj, endProc, endArg, nil);
            remove_working_version (rcsCntx, obj);
        }


    /*
     * Note that I'm using single quotes around the log message string.  This
     * forces me to escape fewer shell control characters than if I used
     * double quotes.
     */
    msg := sprintf ("-m'%s' ", escape_chars (log_message));
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    cmd := "ci -f " + msg + working_path + " " + storage_path;

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmCheckin: %s\n", cmd));

    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmLock -- Get a lock for the object.
 *
 * DESCRIPTION
 *      The storage manager is requested to establish a lock for the data
 *      object for this user.  No other user is then allowed to check out this
 *      object with a lock.
 *
 * NOTE
 *      The term "lock" is understood by the VDM as an exclusive right to
 *      modify the data object without any other user being able to
 *      _vdmCheckin any modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmLock(obj, endProc, endArg, rcsCntx)
{
    cntx := vdm.ObjGetCntx (obj);

    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    cmd := "rcs -q -l" + handle_version (obj);
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    cmd = cmd + " " + storage_path;

    /* In the RCS world, locked files should be writable. */
    vdm.UtilChmod (working_path, "wr", "r", "r");


    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmLock: %s\n", cmd));

    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmUnlock -- Cancel a lock.
 *
 * DESCRIPTION
 *      Relinquish the lock on the data object.
 *
 * NOTE
 *      The term "lock" is understood by the VDM as an exclusive right to
 *      modify the data object without any other user being able to
 *      _vdmCheckin any modifications until the "lock" is relinquished.
 *
 *      That is, while the object is locked by the storage manager, no other
 *      user can check in any new versions of the same design object.
 */
function _vdmUnlock(obj, endProc, endArg, rcsCntx)
{
    cntx := vdm.ObjGetCntx (obj);

    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    cmd          := "rcs -q -u" + handle_version (obj);
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    cmd = cmd + " " + storage_path;

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmUnlock: %s\n", cmd));

    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmRegister -- Register a new data object.
 *
 * DESCRIPTION
 *      Tell the storage manager about a brand new data object that is being
 *      added under its care and management.  Note that this operation is
 *      different from checking in, which creates a new revision for an
 *      already existing design object.
 *
 * NOTE
 *      Some storage managers don't require that you specify an initialization
 *      message.  Just ignore it in that case.
 */
function _vdmRegister(obj, endProc, endArg, rcsCntx, init_message)
{
    cntx := vdm.ObjGetCntx (obj);

    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    /*
     * Make sure that the storage area exists.  This includes creating the
     * directory. 
     */
    <device, path, nameext> := vscript.vsParse2 (storage_path);
    vscript.vsCreateWithPath (vsMerge (device, path, "", ""), vsDIR);

    cmd := "rcs -q -i " + storage_path;
    cmd  = sprintf ("echo '%s' | %s", escape_chars (init_message), cmd);

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmRegister: %s\n", cmd));

    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmDelete -- Delete the data object.
 *
 * DESCRIPTION
 *      Delete the data object from the storage manager.  No assumptions are
 *      made whether the delete operation is permanent or can be undone.
 *
 * NOTE
 *      Because RCS has a separate working and storage areas, we need to
 *      delete the copies of the file from both of these areas.  In addition,
 *      RCS changes the status of its files to be read only.  We have to
 *      change the status, otherwise the vscript.vsDelete function will fail.
 *      The previous implementation used the "rm -f" shell command, but it
 *      incurred an expensive subprocess spawning overhead.
 */
function _vdmDelete(obj, endProc, endArg, rcsCntx)
{
    cntx := vdm.ObjGetCntx (obj);
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmDelete: %s and %s\n",
                                 storage_path, working_path));

    /*
     * NOTE: we need to force a deletion of RCS files which usually have a
     * read-only user permission.
     */
    vdm.UtilChmod (storage_path, "wr", "r", "r");
    vdm.UtilChmod (working_path, "wr", "r", "r");


    /*
     * Note the order of evaluation in the boolean expression: we suffer the
     * expansive fexists function only if the deletion has failed.
     */
    if ((! vscript.vsDelete (storage_path, vsAttr.File)) &&
        fexists (storage_path))
    {
        msg := sprintf ("Unable to delete the %s file.", storage_path);
        vlErr.Out2 ("vdmRCS", vlErrSev.Error, msg);
                    
        vdm.CntxCallErrorHandlers (cntx, msg);
    }

    if ((! vscript.vsDelete (working_path, vsAttr.File)) &&
        fexists (working_path))
    {
        msg := sprintf ("Unable to delete the %s file.", working_path);
        vlErr.Out2 ("vdmRCS", vlErrSev.Error, msg);

        vdm.CntxCallErrorHandlers (cntx, msg);
    }        

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmQuery -- Get information on the data object.
 *
 * DESCRIPTION
 *      This method returns a list of object informations.  The list has the
 *      following format:
 *
 *          ("log" (("user1" . "rev1") ("user2" . "rev32"))
 *                 ("version1" "version2"))
 *
 *      The first element is a string representing the log.  More on the log
 *      below.
 *
 *      The second element is a list of the lock information.  Some storage
 *      managers that allow multiple users to have locks on multiple branches
 *      will find this format useful.  Other storage managers will return a
 *      single element list.  The lock information is a list of pairs: each
 *      pair is composed of the user name and the revision.  The pair contains
 *      information about a user which holds a lock for a particular revision.
 *
 *      The third element is a list of all the versions of this object.
 *
 * CAUTION
 *      Gathering the log information might be an expensive operation.  For
 *      this reason, this method allows the user to specify whether to get the
 *      log information or not.  The `get_log' argument will tell you that.
 *
 *      Secondly, because of the potential amount of the log information, the
 *      user might not choose to have it returned as a string, but will pass
 *      in a function pointer to be used to display the log information.  If
 *      the `log_proc' argument is not nil, you must pass the log information
 *      to this function.  It's declared to take a single argument, a string,
 *      and can be called multiple times.
 *
 *      The user will know when the log information is over, when this
 *      function returns.
 */
function _vdmQuery(obj, endProc, endArg, rcsCntx, get_log, log_proc)
{
    re_vers      := vscript.re_compile ("revision ([0-9.]+)");
    re_lock      := vscript.re_compile ("(.*)locked by: (.*);");   
    re_symb      := vscript.re_compile ("\t[a-zA-Z0-9_]+: [0-9\.]+\n");   
    revision     := nil;
    versions     := nil;
    locks        := nil;
    log          := "";
    name         := vdm.ObjGetPhysicalName (obj) + ",v";
    storage_path := vscript.vsMerge ("", rcsCntx.storage_area, name, "");
    proc_syms    := FALSE;
    sym_str      := "";


    cntx := vdm.ObjGetCntx (obj);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
    {
        if ((code != 0) || (mode != vcSysMode.Exited))
            vlErr.Out2 ("vdmRCS", vlErrSev.Error,
                        sprintf ("RCS exit code: %s, mode: %s", code,
                                 mode));

        endProc (obj, endArg, list (log, locks,
                                    reverse_destructive (versions)));
    }

    /*
     * The output handler is really processing the log information.  If the
     * log_proc is not nil, send it to the user's function.  Independent of
     * the log, gather the version and the lock information.  Both are picked
     * up by regular expressions: it makes it easier to change if the log
     * format changes somehow.
     */
    function out_handler(str)
    {
        if (get_log)
            if (log_proc == nil)
                log = log + str;
            else
                log_proc (str);

        if (!null (string_search (str, "symbolic names:")))
        {
            proc_syms = TRUE;
            sym_str   = sym_str + str;
        }

        if (proc_syms && vscript.re_exec (re_symb, str))
            sym_str = sym_str + str;

        if (!null (string_search (str, "comment leader:")))
        {
            declare token;   
            proc_syms = FALSE;
            
            /* Remove the "symbolic names:" bit. */
            sym_str = string_token (sym_str, &token, ":");
            while (!null (sym_str = string_token (sym_str, &token, " :;\n\t")))
            {
                versions = cons (token, versions);
                sym_str  = string_token (sym_str, &token, " :;\n\t");
            }
        }

        if (vscript.re_exec (re_vers, str))
        {
            revision = vscript.re_getmatch (re_vers, 1);
            versions = cons (revision, versions);

            if (vscript.re_exec (re_lock, str))
                locks = cons (cons (vscript.re_getmatch (re_lock, 2),
                                    revision),
                              locks);
        }
    }
        
    cmd := "rlog " + storage_path;

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmQuery: %s\n", cmd));

    /*
     * Shouldn't really call the rlog function on an object that doesn't
     * exist.
     */
    <wk_status, sm_status> := vdm.ObjGetStatus (obj);

    if (sm_status == vdm.SMStatus.Uncontrolled)
        endProc (obj, endArg, list (nil, nil, nil));
    else
        vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                         exit_handler, nil));
}



/*\Docpage _vdmFind -- Find the object.
 *
 * DESCRIPTION
 *      Is this object under the control of this storage manager in the given
 *      location?
 */
function _vdmFind(obj, endProc, endArg, rcsCntx)
{
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmFind: %s and %s\n", storage_path,
                                 working_path));

    endProc (obj, endArg, (fexists (storage_path) || fexists (working_path)));
}



/*\Docpage _vdmGetPhysicalDescriptor -- Get the physical descriptor.
 * 
 * DESCRIPTION
 *      Each persistent data object is stored somewhere on disk, and thus has
 *      a physical descriptor.  The form of the descriptor depends on the
 *      storage manager.  For file-based storage managers, it might be a
 *      filename,  For databases, it might be a combination of tokens
 *      describing the name and the location of the database plus some
 *      information about where the given data is located in that database.
 */
function _vdmGetPhysicalDescriptor(obj, endProc, endArg, rcsCntx)
{
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmGPD: %s\n", working_path));

    endProc (obj, endArg, working_path);
}



/*\Docpage _vdmGetPhysicalSMDescriptor -- Get the physical descriptor.
 * 
 * DESCRIPTION
 *      Return a string which describes how the object is physically stored
 *      in the SM. This may be a path, but on the other hand it might be.
 *
 *      The intention of this routine is to give users the ability to
 *      identify multiple vdm objects which are pointeing to the same
 *      physical object.
 */
function _vdmGetPhysicalSMDescriptor(obj, endProc, endArg, rcsCntx)
{
    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmGPD: %s\n", storage_path));

    endProc (obj, endArg, storage_path);
}



/*\Docpage _vdmLoop -- Build a list of objects for a VDM loop.
 * 
 * DESCRIPTION
 *      Returns a list of data objects of the given class in this location.
 *      If there are no objects of that type, it returns nil.  This function
 *      is called by the vdm.ClassLoopObj function.
 */
function _vdmLoop(obj, endProc, endArg, rcsCntx, class, loc, cntx, user_filter)
    loop_aux (obj, endProc, endArg, rcsCntx, class, loc, cntx, user_filter,
              TRUE, TRUE);



/*\Docpage _vdmLoopWorking -- Build a list of objects for a VDM loop.
 * 
 * DESCRIPTION
 *      Returns a list of data objects of the given class in this location,
 *      but only those objects that are in the working area of the location.
 *      If there are no objects of that type, it returns nil.  This function
 *      is called by the vdm.ClassLoopObjWorking function.
 */
function _vdmLoopWorking(obj, endProc, endArg, rcsCntx, class, loc, cntx,
                         user_filter)
    loop_aux (obj, endProc, endArg, rcsCntx, class, loc, cntx, user_filter,
              TRUE, FALSE);



/*\Docpage _vdmLoopStorage -- Build a list of objects for a VDM loop.
 * 
 * DESCRIPTION
 *      Returns a list of data objects of the given class in this location
 *      but only those objects that are in the storage area of the location.
 *      If there are no objects of that type, it returns nil.  This function
 *      is called by the vdm.ClassLoopObjStorage function.
 */
function _vdmLoopStorage(obj, endProc, endArg, rcsCntx, class, loc, cntx,
                         user_filter)
    loop_aux (obj, endProc, endArg, rcsCntx, class, loc, cntx, user_filter,
              FALSE, TRUE);



/*\Docpage loop_aux -- Build a list of objects for a VDM loop.
 * 
 * DESCRIPTION
 *      Returns a list of data objects of the given class in this location.
 *      If there are no objects of that type, it returns nil.
 */
function loop_aux(obj, endProc, endArg, rcsCntx, class, loc, cntx,
                  user_filter, do_working, do_storage)
{
    file_template   := vdm.ClassBuildFileTemplate (class);
    regexp_template := vdm.ClassBuildRegexpTemplate (class, user_filter);
    path            := vscript.vsMerge ("", rcsCntx.storage_area,
                                        file_template + ",v", "");
    path_length     := string_length (rcsCntx.storage_area);
    regexp          := vscript.re_compile (regexp_template);
    objects         := vlSym.StrCreate (111);
    matched_attrs   := nil;
    match_str       := nil;
    index           := 0;

    vlSym.Dynamic (objects, 2.0, 2.0);

    /*
     * The algorithm for the RCS encapsulation is as follows:
     *
     * 1. Substitute all the attribute references in the template with '*'.
     * 2. Append the "wildcarded" template to the storage area's path.
     * 3. Loop through all the files that match the template.  Then apply the
     *    regular expression template to narrow down the match.
     * 4. If the regular expression template matches, extract the attribute
     *    values and build an object handle.  Then, store the object in a hash
     *    table.
     * 5. Repeat 3,4 for the working area's path, but this time checkin if the
     *    object has already been found in the storage area.
     * 6. Create a list from all the objects in the hash table.
     */
    if (do_storage)
    {
        if (vdmDebug)
            debug_out (obj, sprintf ("vdmRCS _vdmLoop: %s\n", path));

        foreach (pathname; vscript.TraverseDir (path, vsAttr.Both))
        {
            if ((index = rindex (pathname, ',')) != FALSE)
            {
                pathname  = substring (pathname, 0, index);
                match_str = substring (pathname, path_length, 0);
                if (vscript.re_exec (regexp, match_str))
                {
                    matched_attrs =
                        vdm.ClassGetMatchedAttributes (class, regexp,
                                                       match_str);
                    vlSym.Store (objects, match_str,
                                 vdm.ObjCreate (match_str, vdm.VersLatest,
                                                class, loc, cntx,
                                                matched_attrs));
                }
            }
        }
    }

    /*
     * Now loop through the working area, but don't include the objects that
     * have already been picked up from the storage area.
     */
    path = vscript.vsMerge ("", rcsCntx.working_area, file_template, "");
    path_length = string_length (rcsCntx.working_area);

    if (do_working)
    {
        if (vdmDebug)
            debug_out (obj, sprintf ("vdmRCS _vdmLoop: %s\n", path));

        foreach (pathname; vscript.TraverseDir (path, vsAttr.Both))
        {
            match_str = substring (pathname, path_length, 0);
            if ((vlSym.Fetch (objects, match_str) == nil) &&
                (vscript.re_exec (regexp, match_str)))
            {
                matched_attrs = vdm.ClassGetMatchedAttributes (class, regexp,
                                                               match_str);
                vlSym.Store (objects, match_str,
                             vdm.ObjCreate (match_str, vdm.VersLatest, class,
                                            loc, cntx, matched_attrs));
            }
        }
    }

    /*
     * Move the objects from the table into a list, because that's what this
     * method is supposed to return.
     */
    object_list := nil;
    foreach (entity, obj; vlSym.Loop (objects))
        object_list = cons (obj, object_list);

    endProc (obj, endArg, object_list);
}



/*\Function _vdmStatus -- What's the storage status of this object.
 * 
 * DESCRIPTION
 *      The sole purpose of this method is to determine the status of the
 *      given object with respect to the storage manager.  This method must
 *      return two values, representing the possible storage manager state
 *      (vdm.SMStatusVT) and the possible working area state (vdm.WorkAreaVT).
 *
 *      If the object is under the storage manager's control, the object's
 *      status can be one of the following:
 *
 *          vdm.SMStatus.Unlocked       - object is not locked
 *          vdm.SMStatus.LockedByYou    - object is locked by you
 *          vdm.SMStatus.LockedByOther  - object is locked by another user
 *          vdm.SMStatus.Uncontrolled   - object is not yet in SM
 *
 *      If the storage manager allows the data object to persist outside of
 *      the storage manager's storage area, then such an object is said to
 *      reside in the user's working area.  Its status can then be one of the
 *      following:
 *
 *          vdm.WorkStatus.New       - object is brand new.
 *          vdm.WorkStatus.Modified  - object is modified
 *          vdm.WorkStatus.Outdated  - object is not the same as object in SM
 *          vdm.WorkStatus.UpToDate  - object is the same as that in SM
 *          vdm.WorkStatus.Empty     - object is not in the work area
 *
 * SEE ALSO
 *      vdm.SMStatusVT, vdm.WorkStatusVT
 */
function _vdmStatus(obj, endProc, endArg, rcsCntx)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmStatus.\n");


    /* Handler functions. */
    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    rcs := nil;
    <storage_path, local_path> := build_pathnames (obj, rcsCntx);


    /*
     * The algorithm for the RCS encapsulation is as follows:
     *
     * If the file (everything is a file in RCS) doesn't exist in the local
     * area, it's in storage, case closed.
     *
     * If the file cannot be found in the storage area, that means that it's
     * brand new and has never been registered.
     *
     * If the file is found in the storage area, get the timestamps of both
     * the local and the storage files and compare them.  There can be only
     * three outcomes to the comparison: the files are the same, the local
     * file is newer or older.
     */
    storage_time := vscript.vsFGMtime (storage_path);
    local_time   := vscript.vsFGMtime (local_path);

    sm_status := nil;
    wk_status := nil;

    if (storage_time == 0)
        sm_status = vdm.SMStatus.Uncontrolled;
    else
    {
        rcs = rcs_parse (storage_path);

        /* No deltas, but its registered ... so say we have it locked. */
        if (vlSym.Size (rcs.deltas) == 0)
            sm_status = vdm.SMStatus.LockedByYou;

        /* Deltas, but no locks so its unlocked. */
        else if (vlSym.Size (rcs.locks) == 0)
            sm_status = vdm.SMStatus.Unlocked;


        /* Deltas and locks. Is it me or you? */
        else
        {
            user := vscript.vsGetLoginName ();

            if (null (user))
                user = "";

            vers := vdm.ObjGetVersion (obj);
            if (string_eq (vers, vdm.VersLatest))
                if (local_time == 0)
                    vers = rcs_default_version (rcs);
                else
                    vers = wvers_fetch_vers (rcsCntx, obj);

            u := vlSym.Fetch (rcs.locks, vers);
            if (null (u))
                sm_status = vdm.SMStatus.Unlocked;
            else if (string_eq (u, user))
                sm_status = vdm.SMStatus.LockedByYou;
            else                        
                sm_status = vdm.SMStatus.LockedByOther;
        }
    }


    /* Working area status. */
    if (local_time == 0)
        wk_status = vdm.WorkStatus.Empty;

    else if (sm_status == vdm.SMStatus.Uncontrolled)
        wk_status = vdm.WorkStatus.New;

    else if (vlSym.Size (rcs.deltas) == 0)
        wk_status = vdm.WorkStatus.Modified;

    else
    {
        /* Figure out what version we have. */
        vers := handle_version (obj);
        if (vers == "")
        {
            vers = wvers_fetch_vers (rcsCntx, obj);
            if (null (vers))
                vers = rcs_default_version (rcs);
        }

        delta := vlSym.Fetch (rcs.deltas, vers);
        if (null (delta))
            vlEx.Throw2 (vlEx.FAILURE, "vdmRCS", vlErrSev.Error,
                         sprintf (
            "Unable to find version information for `%s' version `%s'.\n" +
            "\n" +
            "If you are seeing this message after first installing\n" +
            "Powerview 5.3.2, remove all data from the working\n" +
            "area and perform a Checkout operation.\n" +
            "\n" +
            "See the Powerview 5.3.2 Release Notes for more\n" +
            "information.", vdm.ObjGetName (obj), vers));

        storage_time = rcs_rcstime_to_longtime (delta.date);


        wvers := wvers_fetch_vers (rcsCntx, obj);
        if (!null (wvers) && !string_eq (vers, wvers))
            wk_status = vdm.WorkStatus.Empty;
        else if (local_time > storage_time)
            wk_status = vdm.WorkStatus.Modified;
        else if (local_time < storage_time)
            wk_status = vdm.WorkStatus.Outdated;
        else
            wk_status = vdm.WorkStatus.UpToDate;
    }


    endProc (obj, endArg, list (wk_status, sm_status));
}



/*\Docpage _vdmMetaDataName -- Create a data name for a meta-data object.
 * 
 * DESCRIPTION
 *      This is an optional method.  It should be provided only if the given
 *      storage manager does not handle the meta-data itself.
 *
 *      The sole purpose of this method is to return a string that will
 *      represent the data name of the meta-data object.  If the storage
 *      manager does not handle meta-data, VDM will create a meta-data object
 *      which shadows the data object.  But, given a data object, VDM has to
 *      be able to find its corresponding meta-data object.  That's where this
 *      method enters the picture.
 *
 *      The object passed into this method is the data object.  Use its data
 *      name to create a unique meta-data name.  For file-based storage
 *      managers, it's recommended that you prepend the name of a
 *      subdirectory, like `vdmmeta' to the data object's name and return it.
 *      In this way, the name of the meta-data object is the same as that of
 *      the data object's that it's supposed to shadow, but in a special
 *      subdirectory.  You can make that subdirectory anything you want, make
 *      it visible or not to the user, just as long as it's valid for this
 *      storage manager.
 */
function _vdmMetaDataName(obj, endProc, endArg, rcsCntx)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmMetaDataName.\n");

    endProc (obj, endArg, vscript.vsMerge ("", "vdmmeta",
                                           vdm.ObjGetPhysicalName (obj), ""));
}



/*\Docpage _vdmSetTag -- Set a tag on the object.
 * 
 * DESCRIPTION
 *      Attaches a given string tag to the object.
 */
function _vdmSetTag(obj, endProc, endArg, rcsCntx, tag)
{
    cntx := vdm.ObjGetCntx (obj);

    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    rev := "-n" + tag + ":" + handle_version (obj);
    cmd := "rcs -q " + rev + " " + storage_path;

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmSetTag: %s\n", cmd));

    vcSys.LineBuffer (vcSys.Command (cmd, nil, out_handler, err_handler,
                                     exit_handler, nil));
}



/*\Docpage _vdmGetDefaultVersion -- Get the object's default version.
 * 
 * DESCRIPTION
 *      What is the object's latest default version?
 *
 * NOTE
 *      The version number must be returned as a string value.
 */
function _vdmGetDefaultVersion(obj, endProc, endArg, rcsCntx)
{
    cntx    := vdm.ObjGetCntx (obj);
    version := nil;

    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    <storage_path, working_path> := build_pathnames (obj, rcsCntx);

    if (fexists (storage_path))
        version = rcs_default_version (rcs_parse (storage_path));
    else
        version = nil;


    endProc (obj, endArg, version);
}



/*\Docpage _vdmGetWorkingVersion -- Get the object's working version.
 * 
 * DESCRIPTION
 *      What is the object's latest working version?
 *
 * NOTE
 *      The version number must be returned as a string value.
 */
function _vdmGetWorkingVersion(obj, endProc, endArg, rcsCntx)
{
    function out_handler(str)
        vdm.CntxCallOutputHandlers (cntx, str);

    function err_handler(str)
        vdm.CntxCallErrorHandlers (cntx, str);

    function exit_handler(mode, code)
        common_exit_handler (mode, code, obj, endProc, endArg, nil);


    version :=  wvers_fetch_vers (rcsCntx, obj);


    endProc (obj, endArg, version);
}



/****************************************************************************/
/*                              I/O methods                                 */
/****************************************************************************/


/*
 * The RCS I/O encapsulation needs to hang the file pointer somewhere: the
 * object is the most obvious and convenient object.  The file pointer will be
 * stored as the value of the _vdmFp attribute.
 */
FILE_ATTR_STR := "_vdmFp";



/*\Docpage _vdmOpen -- Open a design object for I/O operations.
 *
 * DESCRIPTION
 *      Opens the the object for a particular mode of I/O operations.  If the
 *      object has already been opened and not closed, an exception will be
 *      raised.
 */
function _vdmOpen(obj, endProc, endArg, rcsCntx, mode)
{
    path      := vscript.vsMerge ("", rcsCntx.working_area,
                                  vdm.ObjGetPhysicalName (obj), "");
    fp        := nil;
    file_mode := nil;


    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmOpen: %s, mode: %s.\n", path,
                                 mode));

    /*
     * Verify that the _vdmFp attribute is not already taken.  If so, somebody
     * is trying to reopen this object.
     */
    if (! (null (vdm.ObjGetAttr (obj, FILE_ATTR_STR))))
        vlEx.Throw2
            (vlEx.FAILURE, "vdmRCS", vlErrSev.Err,
             sprintf
             ("Attempted to re-open %s when it hasn't been closed yet.",
              vdm.ObjGetName (obj)));
    

    if (mode == vdm.IOMode.Read)
        file_mode = "r";
    else
    {
        /*
         * Make sure any sub-directories exist when we are performing "write"
         * operations.  Ideally, we would wait for an operation to fail before
         * doing this ... but that's more involved.
         */
        <device, subdir, filename> := vscript.vsParse2 (path);

        vscript.vsCreateWithPath (subdir, vsAttr.Dir);
    
        if (mode == vdm.IOMode.Write)
            file_mode = "w";
        else if (mode == vdm.IOMode.Append)
            file_mode = "a";
        else
            file_mode = "r+";
    }
    
    fp = fopen (path, file_mode);
    vdm.ObjSetAttr (obj, FILE_ATTR_STR, fp, FALSE);


    /* !!!  We have modified the file, so there is no need for
     *      us to muck with the timestamp.
     * !!!  Uh, there's an optimization to be made here. Look above.
     */
    if ((mode == vdm.IOMode.Write) || (mode == vdm.IOMode.Append))
        vdm.ObjRemoveAttr (obj, vdmRCS_STAMP_ATTR);


    endProc (obj, endArg, nil);
}



/*\Docpage _vdmClose -- Close a design object for I/O operations.
 *
 * DESCRIPTION
 *      If the object has been closed already, just ignore it.  Otherwise,
 *      close the object in such a way that any subsequent I/O request will
 *      fail, with the exception of open.
 */
function _vdmClose(obj, endProc, endArg, rcsCntx)
{
    path := vscript.vsMerge ("", rcsCntx.working_area,
                             vdm.ObjGetPhysicalName (obj), "");

    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmClose %s.\n", path));

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    /*
     * If the object has already been closed, ignore it.  Otherwise, close the
     * file pointer and set the attribute value to nil.
     */
    if (! null (fp))
    {
        fclose (fp);
        vdm.ObjSetAttr (obj, FILE_ATTR_STR, nil, FALSE);
    }

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmReadChar -- Read a single character.
 */
function _vdmReadChar(obj, endProc, endArg, rcsCntx)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmReadChar.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    endProc (obj, endArg, fgetchar (fp));
}



/*\Docpage _vdmReadLine -- Read a single line from the design.
 */
function _vdmReadLine(obj, endProc, endArg, rcsCntx)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmReadLine.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    endProc (obj, endArg, fgetline (fp));
}



/*\Docpage _vdmRead -- Read a number of characters from a design.
 */
function _vdmRead(obj, endProc, endArg, rcsCntx, byte_no)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmRead.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    endProc (obj, endArg, fread (fp, byte_no));
}



/*\Docpage _vdmWriteChar -- Write a single character into a design object.
 */
function _vdmWriteChar(obj, endProc, endArg, rcsCntx, char)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmWriteChar: %c.\n", char));


    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    fputchar (fp, char);

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmWriteLine -- Write a single line into a design object.
 */
function _vdmWriteLine(obj, endProc, endArg, rcsCntx, line)
{
    if (vdmDebug)
        debug_out (obj, "vdmRCS _vdmWriteLine.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    fputline (fp, line);

    endProc (obj, endArg, nil);
}



/*\Docpage _vdmWrite -- Write a number of character into a design object.
 */
function _vdmWrite(obj, endProc, endArg, rcsCntx, block)
{
    debug_out (obj, "vdmRCS _vdmWrite.\n");

    fp := vdm.ObjGetAttr (obj, FILE_ATTR_STR);

    endProc (obj, endArg, fwrite (fp, block));
}



/****************************************************************************/
/*                              Generic method                              */
/****************************************************************************/


/*\Docpage _vdmGeneric -- Generic method.
 *
 * DESCRIPTION
 *      The VDM backplane has a fixed number of methods.  The generic method
 *      is provided as a means of extending the backplane to allow custom
 *      integrations that require more storage management capabilities than
 *      the predefined methods allow.  This function will be called with a
 *      keyword, which should serve as a dispatch key to call other functions.
 *      This method will be called as a result of a call to the
 *      vdm.ObjCallGeneric(man) API function.  That function accepts the
 *      dispatch key, represented as a string, and an addition argument of any
 *      type.
 *
 * EXAMPLE
 *      arg := nil;
 *      vdm.ObjCallGeneric (obj, "action1", arg);
 */
function _vdmGeneric(obj, endProc, endArg, rcsCntx, dispatch_key, arg)
{
    if (vdmDebug)
        debug_out (obj, sprintf ("vdmRCS _vdmGeneric: %s\n", dispatch_key));

    endProc (obj, endArg, nil);
}


/****************************************************************************/
/*                          AUXILIARY FUNCTIONS                             */
/****************************************************************************/


/*\Docpage common_exit_handler -- Common RCS exit handler.
 *
 * DESCRIPTION
 *      Because the endProc and endArg arguments passed in to each method are
 *      different every time, RCS encapsulation cannot have one single exit
 *      handler which is passed in to every vcSys.Command call.  Instead, each
 *      method has to declare its own exit handler which creates a closure of
 *      all the variable bindings in effect at that moment.  That exit handler
 *      can then call this function, the common exit handler, with the
 *      appropriate arguments.
 *
 *      If the subprocess terminated with the exit status not equal to zero or
 *      was interrupted, this function will display an error message.
 *
 * CAUTION
 *      You must call the endProc function to signal to the VDM that the
 *      subprocess has terminated.  Do not raise any exceptions in this
 *      function; any errors should just be reported.
 */
function common_exit_handler(mode, code, obj, endProc, endArg, ret_value)
{
    if (mode == vcSysMode.Signaled)
    {
        str := sprintf
            ("An interrupt (#%d) was detected during an RCS command.\n",
             code);
        vdm.CntxCallErrorHandlers (vdm.ObjGetCntx (obj), str);
    }
    else if (code != 0)
    {
        str := sprintf ("RCS command exited with error code: %d\n", code);
        vdm.CntxCallErrorHandlers (vdm.ObjGetCntx (obj), str);
    }

    endProc (obj, endArg, ret_value);
}


/*\Docpage stamping_exit_handler -- Locking RCS exit handler.
 *
 * DESCRIPTION
 *      This routine is identical to the common_exit_handler() except that
 *      it updates the timestamp of the file in the working area. This is
 *      how we are getting around the problem of files in the local area
 *      looking modified immediately after being checked out.
 */
function stamping_exit_handler(mode, code, obj, endProc, endArg, ret_value)
{
    common_exit_handler (mode, code, obj, endProc, endArg, ret_value);

    vdm.ObjSetAttr (obj, vdmRCS_STAMP_ATTR, nil, FALSE);
}



/*\Docpage handle_version -- Translate version.
 *
 * DESCRIPTION
 *      Translate the "internal" version to the RCS version.
 */
function handle_version(obj)
{
    version := vdm.ObjGetVersion (obj);

    if (null (version) ||
        string_eq (version, "vdmVersLATEST") ||
        string_eq (version, "vdmVersDEFAULT"))
        version = "";

    version;
}



/*\Docpage local2central_exit_handler -- Exit handler
 *
 * DESCRIPTION
 *      Move the local metadata database into the central database.
 */
function
local2central_exit_handler(mode, code, obj, endProc, endArg, cntx,
                           new_vers, old_vers, ret_value)
{
    if (mode == vcSysMode.Signaled)
    {
        str := sprintf
            ("An interrupt (#%d) was detected during an RCS command.\n",
             code);
        vdm.CntxCallErrorHandlers (vdm.ObjGetCntx (obj), str);
    }
    else if (code != 0)
    {
        str := sprintf ("RCS command exited with error code: %d\n", code);
        vdm.CntxCallErrorHandlers (vdm.ObjGetCntx (obj), str);
    }
    else
        /* Move metadata to central database. */
        LocalToCentral (cntx, obj, old_vers, new_vers);


    endProc (obj, endArg, ret_value);
}



/*\Docpage central2local_exit_handler -- Exit handler
 *
 * DESCRIPTION
 *      Move metadata from the central database to the local one.
 */
function
central2local_exit_handler(mode, code, obj, endProc, endArg, cntx, ret_value)
{
    if (mode == vcSysMode.Signaled)
    {
        str := sprintf
            ("An interrupt (#%d) was detected during an RCS command.\n",
             code);
        vdm.CntxCallErrorHandlers (vdm.ObjGetCntx (obj), str);
    }
    else if (code != 0)
    {
        str := sprintf ("RCS command exited with error code: %d\n", code);
        vdm.CntxCallErrorHandlers (vdm.ObjGetCntx (obj), str);
    }
    else
        /* Move metadata from central database. */
        CentralToLocal (cntx, obj);


    endProc (obj, endArg, ret_value);
}



/*\Docpage escape_chars -- Escape shell characters.
 *
 * DESCRIPTION
 *      Backslash escape any newline characters in the log message because
 *      otherwise, the shell will interpret the newline as the end of the
 *      command.  Also, escape any shell control characters that might cause
 *      trouble if the string is enclosed in single quotes.
 */
function escape_chars(str)
{
    new_str := "";
    str_len := string_length (str);

    for (i := 0; i < str_len; ++i)
        if (str [i] == '\n')
            new_str += "\\\n";
        else if (str [i] == '!')
            new_str += "\\!";
        else if (str [i] == '\'')
            new_str += "'\\''";
        else
            new_str += string_var (str [i]);

    new_str;
}



/*\Docpage debug_out -- Handy output function.
 * 
 * DESCRIPTION
 *      Takes an object and a string and if the vdmDebug variable is set, it
 *      calls the VDM's context output handlers with the string.
 *
 *      When running ViewData, the output will show up in the session log
 *      window.
 */
function debug_out(obj, str)
    vdm.CntxCallOutputHandlers (vdm.ObjGetCntx (obj), str);



/*\Docpage build_pathnames -- Build storage and work area pathnames.
 *
 * DESCRIPTION
 *      This is needed so frequently that it was useful to simply have a
 *      single function which returns the expanded pathnames.
 *
 * NOTE
 *      The location of the object (its storage_path) is the rcsCntx.  The
 *      full pathname of the object is the storage_path + the object's data
 *      name + the ,v extension.  Note that the ,v extension cannot be added
 *      in the call to the vsMerge function, because then it would appended as
 *      a ".,v".
 */
function build_pathnames(obj, rcsCntx)
{
    name         := vdm.ObjGetPhysicalName (obj);
    storage_path := vscript.vsMerge ("", rcsCntx.storage_area, name, "");
    storage_path  = storage_path + ",v";

    dest_path    := vscript.vsMerge ("", rcsCntx.working_area, name, "");
    list (storage_path, dest_path);
}



/*\Docpage create_dest_dir -- Make sure the destination path exists.
 *
 * DESCRIPTION
 *      This is needed to make sure that the path to which an object is
 *      checkouted exists. If we do not do this, trouble arises when the
 *      user attempts to checkout data from an existing storage area
 *      into a new working area.
 */
function create_dest_dir(dest_path)
{
    <device, path, name> := vscript.vsParse (dest_path);
    vscript.vsCreateWithPath (vsMerge (device, path, "", ""), vsDIR);
}



/*\Docpage is_new_meta -- Figure out if the new metadata is being used.
 */
function
is_new_meta()
    vlPref.Val (vlPref.Create ("vdm"), "new_meta");



/*\Docpage get_the_version -- Get the objects new and old version.
 *
 * DESCRIPTION
 *      This function is registered as an error handler. It is used by the
 *      VDM 2.x metadata scheme to figure out which version is being
 *      checkined, and which version it is going to become.
 *
 *      It gets called many times, so we make sure it only sets the
 *      version variables when it gets a match, so don't init them to
 *      nil, or anything, or "things will break".
 */
function
get_the_version (str)
{
    nre_str := "new revision: ([0-9.]+); previous revision: ([0-9.]+)";
    ire_str := "initial revision: ([0-9.]+)";

    ire := vscript.re_compile (ire_str);
    nre := vscript.re_compile (nre_str);


    if (vscript.re_exec (ire, str))
    {
        this_new_version = vscript.re_getmatch (ire, 1);
        this_old_version = vdm.VersLatest;
    }
    else if (vscript.re_exec (nre, str))
    {
        this_new_version = vscript.re_getmatch (nre, 1);
        this_old_version = vscript.re_getmatch (nre, 2);
    }
}



/*
 *  Working version determination.
 */
function
record_working_version (cntx, obj, vers, path)
{
    if (string_eq (vers, ""))
        vers = rcs_default_version (rcs_parse (path));
    wvers_store_vers(cntx, obj, vers);
}



function
remove_working_version (cntx, obj)
    wvers_remove_vers(cntx, obj);




struct
wvers_file_s
{
    vlIntVT         time;
    vlSymTableVT    cache;
};


struct
wvers_cache_t
{
    vlSymTableVT    cache;    
};


wvers_cache     := struct wvers_cache_t { vlSym.StrCreate (11) };
vlSym.Dynamic (wvers_cache.cache, 1.0, 2.0);


/*\intDocpage rcs_cache_store -- Store this record in the cache.
 *
 *  DESCRIPTION
 *      Store the rcs record in the cache. Boot somebody out if
 *      the cache is already full. The record with the fewest hits
 *      is booted.
 */
function
wvers_store_vers(cntx, obj, vers)
{
    file := wvers_fetch_file (cntx.working_area);
    vlSym.Store (file.cache, vdm.ObjGetVerboseLogicalName (obj), vers);
    wvers_store_file (cntx.working_area, file);
}



function
wvers_fetch_vers(cntx, obj)
{
    file := wvers_fetch_file (cntx.working_area);
    vlSym.Fetch (file.cache, vdm.ObjGetVerboseLogicalName (obj));
}



function
wvers_remove_vers(cntx, obj)
{
    file := wvers_fetch_file (cntx.working_area);
    vlSym.Remove (file.cache, vdm.ObjGetVerboseLogicalName (obj));
    wvers_store_file (cntx.working_area, file);
}



function
wvers_fetch_file (path)
{
    file := vlSym.Fetch (wvers_cache.cache, path);

    if (null (file))
    {
        file = struct wvers_file_s
                {
                    0,
                    vlSym.StrCreate (11)
                };
        vlSym.Dynamic (file.cache, 1.0, 2.0);
    };

    filespec := vscript.vsMerge ("", path, ".vdmwvdb", "vvv");
    finfo := vscript.vsInfo (filespec);

    if ((!null (finfo)) && (file.time < car (finfo)))
    {
        declare l;
        declare key;
        declare vers;

        vlSym.Clear (file.cache);
        
        fp := fopen (filespec, "r");
        while (!null (l = fgetline (fp)))
        {
            l = string_token (l, &key, "|");
            l = string_token (l, &vers, "|");
            vlSym.Store (file.cache, string_trim (key), string_trim (vers));
        }

        fclose (fp);
    }

    file;
}



function
wvers_store_file (path, file)
{
    filespec := vscript.vsMerge ("", path, ".vdmwvdb", "vvv");

    fp := fopen (filespec, "w");
    foreach (key, vers; vlSym.Loop (file.cache))
        fprintf (fp, "%s|%s\n", key, vers);
    fclose (fp);

    finfo     := vscript.vsInfo (filespec);
    file.time  = car (finfo);
}
