/*
 * LENWID.C:     Sample user DRC extension #2
 *
 * This extension adds a rule that looks for LENGTH and WIDTH component
 * attributes and reports all occurrences that don't fall in the range
 * [.25,4].  It also reports errors if it can't parse the length and
 * width attributes as numbers.
 * 
 * This example is more complex than count.c because:
 *      1.      It has its own structure to hold its error classes,
 *              and the legal bounds.
 *      2.      It defines a preferences to define the allowable
 *              range of length/width ratios.
 */

#include "vdrc.h"

/*
 * This could be a static structure, rather than a typedef, but
 * for various reasons, including image-size, shared-library and
 * PC-DOS static memory limitations, it is better to use an allocated 
 * structure.
 */ 
typedef struct {
    vdrcErrorID out_of_range;
    vdrcErrorID invalid_syntax;
    double min_ratio;
    double max_ratio;
} LengthWidthDataT;

/*
 * Try to get a double-precision attribute value, returning TRUE if
 * successful, and reporting an error if not
 */
static vlBoolT get_attr_val(lwd, drc, com, name, val)
    LengthWidthDataT *lwd;
    vdrcID drc;
    COMPONENTS *com;
    const char *name;
    double *val;
{
    ATTRIBUTES *attr = ifattatt (name, igcomatt (com));
    if (attr != NULL) {
        char *attval = igattval (attr);

        if ((sscanf (attval, "%lf", val) != 1) ||
            (*val <= 0))
        {
            char *buf;
            buf = vlSprintfAlloc ("Invalid length/width attribute: %s=%s",
                                   name, attval);
            vdrcReport (drc, lwd->invalid_syntax, (vlPointerT) com, vdrcCOM,
                        buf);
            vlStrFree(buf);
        } /* if */
        else
            return (TRUE);
    } /* if */
    return (FALSE);
} /* static vlBoolT get_attr_val */

static void check_length_width(lwd, drc, com)
    LengthWidthDataT *lwd;
    vdrcID drc;
    COMPONENTS *com;
{
    double length, width;

    if (get_attr_val (lwd, drc, com, "LENGTH", &length) &&
        get_attr_val (lwd, drc, com, "WIDTH", &width))
    {
        double ratio = length / width;

        if ((ratio < lwd->min_ratio) || (ratio > lwd->max_ratio)) {
            char *buf;
            buf = vlSprintfAlloc ("Length/Width %lf out of range %lf..%lf",
                     ratio, lwd->min_ratio, lwd->max_ratio);
            vdrcReport (drc, lwd->out_of_range, (vlPointerT) com, vdrcCOM,
                        buf);
            vlStrFree(buf);
        } /* if */
    } /* if */
} /* static void check_length_width */

/*
 * This routine is called whenever the user changes preferences
 * interactively.  It simply reads the new preference values from
 * the preference database and copies them into the lwd structure.
 */
static void read_prefs(lwd, drc, prefs)
    LengthWidthDataT *lwd;
    vdrcID drc;
    vlPreferencesID prefs;
{
    lwd->min_ratio = 
        (* ((double *) (vlPrefVal (prefs, "min_size_ratio").ptr_data)));
    lwd->max_ratio =
        (* ((double *) (vlPrefVal (prefs, "max_size_ratio").ptr_data)));
} /* static void read_prefs */

/*
 * This function is called by drcuser.c to initialize the test.  Note
 * that in order to report an error, an error class must be created
 * with vdrccErrorCreate.  This error class is passed as client data
 * through to the test function, count_components.
 */
void vdrcLenWidInit(drc, argcp, argv)
    vdrcID drc;
    int *argcp;
    char **argv;
{
    /*
     * Allocate a structure needed to hold the information to perform
     * this check, using vlMemMalloc, which is guaranteed never to
     * return NULL.
     */
    LengthWidthDataT *lwd = (LengthWidthDataT *)
        vlMemMalloc (sizeof (LengthWidthDataT), "LengthWidthData");

    /*
     * Set up the error classes required for this check
     */
    lwd->invalid_syntax = 
        vdrcErrorCreate (drc, "LenWidSyntax", vdrcTestAUTO_ASSIGN, vlErrERROR,
                        "Syntax of length or width attribute is invalid");
    lwd->out_of_range = 
        vdrcErrorCreate (drc, "OutOfRange", vdrcTestAUTO_ASSIGN,
                         vlErrWARNING,
                         "Length/Width ratio is out of range");

    /*
     * Set up preferences to that the user can change the default constraints
     */
    lwd->min_ratio = .25;
    lwd->max_ratio = 4;

    {vlPreferencesID prefs = vdrcPreferences (drc);
     vlPrefAdd (prefs, "min_size_ratio", &vlPositiveFloatTS,
                "Minimum ratio of length/width attributes on components",
                vlTypePTR (&lwd->min_ratio));
     vlPrefAdd (prefs, "max_size_ratio", &vlPositiveFloatTS,
                "Maximum ratio of length/width attributes on components",
                vlTypePTR (&lwd->max_ratio));}

    /*
     * Register the routine that is run on each component in the design,
     * and also a routine that is run whenever the user interactively
     * changes preferences.
     */
    vdrcAddObjTest (drc, vdrcCOM, check_length_width, (vlPointerT) lwd);
    vdrcAddPrefsTest (drc, read_prefs, (vlPointerT) lwd);
} /* void vdrcCountInit */
