/*--------------------------------------------------------------------------*/
/* Library.h - Macros for "inline expansion" of Viewsim builtins.           */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Notes:                                                                   */
/*    1) These macros are invoked by the compiled code generator when       */
/*       the corresponding viewsim builtin is instantiated.                 */
/*    2) Each macro parameter is surrounded by parenthesis to allow         */
/*       the argument to be an arbitrary expression.                        */
/*    3) The macro parameters appear in the order defined by the pinorder   */
/*       attribute of the corresponding builtin.                            */
/*    4) THE INPUTS TO THE MACROS CAN ONLY ASSUME THE VALUES 0, 1, or X.    */
/*       The compiled code generator will guarantee that there are no Z's.  */
/*       To achieve this, it will determine from circuit topology whether   */
/*       a Z can appear at an input, and will invoke a mapping table to     */
/*       convert Z's to X's.                                                */
/*    5) The BUF and DELAY macro's should be recognized as special          */
/*       cases by the compiled code generator (since they perform no        */
/*       function) and should not be generated. They are included here      */
/*       anyway.                                                            */
/*    6) This is the only file that "knows" about the LXZH state encodings  */
/*       and is the only one that needs to change if the encodings change.  */
/*       Except for library.c.                                              */
/*    7) This file replaces the original lib.h developed for the            */
/*       original prototype.                                                */
/*--------------------------------------------------------------------------*/
/* History:                                                                 */
/*                                                                          */
/* WHO   WHEN        WHAT                                                   */
/*--------------------------------------------------------------------------*/
/* RAM   2  Dec 91   Added CCS_RISING and CCS_FALLING defines for find_edge */
/* DBS   25 Nov 91   Fixed udfdl source generation and unknown handling     */
/* RAM   20 Nov 91   Prefixed all routines with CCS_                        */
/* DBS   20 Nov 91   Fixed xnor3                                            */
/* RAM   20 Nov 91   Added init_global_val routine since not provided in VSC*/
/* RAM   12 Nov 91   Added NOT                                              */
/* RAM   12 Nov 91   Fixed params to all jkffre prims to !include CLR,PRE   */
/* RBS   12 Nov 91   Added FLOWTHRU attribute support.                      */
/* RBS   11 Nov 91   Added "sim_val" and "logic_lev"   extern declarations. */
/* DMS    6 Nov 91   Add MUX select                                         */
/* RBS    6 Nov 91   Added assertion checking.                              */
/* RBS    4 Nov 91   Fixed some minor typos                                 */
/* RBS   29 Oct 91   Added ANDs, ORs, NOTs, XORs, TRIs                      */
/* RBS   27 Feb 92   Moved static to library.c                              */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* STATE ENCODINGS                                                          */
/* The dependencies on these encodings are listed below:                    */
/*     1) Table lookup scheme in UJKFF (and decoders and muxes) depends     */
/*        on these assignments. (See vsimlkup.h.) Since Viewsim also        */
/*        uses this table lookup scheme, these assignments must match       */
/*        viewsim's assignments.                                            */
/*     2) The logical operations in the boolean builtin rotines (&, |)     */
/*        depend on these assignments.                                      */
/*     3) All the "state logic and mapping tables" depend on these          */
/*        assignments.                                                      */
/*--------------------------------------------------------------------------*/
#ifndef __library_h__
#define __library_h__

#define DATAFILEEXT "ccs"

typedef struct ccs_timing {
    long delays[4][4];
} CCS_TIMING;

extern int ccs_init_instance();
extern vsc_keyword_handler ccs_keyword_handler();
extern void ccs_free_instance();
extern void ccs_init_oblivious();
extern void ccs_init_event();
extern void ccs_init_pin();
extern void ccs_schedule();
extern void ccs_current_instance();

/*--------------------------------------------------------------------------*/
/* STATE LOGIC and MAPPING TABLES                                           */
/*--------------------------------------------------------------------------*/

/* from xdefs.c */
extern char sim_val[16][4]; /* STR01,VAL->28 state */
extern char smerge[31][32]; /* bus resolution */

extern vsc_pin_4values vsc_31to4_state[31];
extern vsc_pin_4values vsc_31to3_state[31];

extern int z2x[4];                                   /* Converts Z to X */
extern int function_not[4];                          /* Inversion (~Z->X) */
extern int function_xor[4][4];                       /* Xor */
extern int function_xnor[4][4];                      /* Xnor */
extern int mux_select[4][4];                         /* selects out [S1][S0] */

extern void CCS_UBDEC38();
extern void CCS_MUX41();
extern void CCS_UJKFF();
extern void CCS_UDFDL();
extern void CCS_FT_UDFDL();
extern void CCS_TRIAND2();
extern void CCS_TRIOR2();
extern void CCS_TRINAND2();
extern void CCS_TRINOT();
extern void CCS_TRIBUF();
extern void CCS_TRIAND3();
extern void CCS_TRIAND4();
extern void CCS_TRINOR4();
extern void CCS_TRIOR4();
extern void CCS_TRINAND4();
extern void CCS_TRINOR3();
extern void CCS_TRIOR3();
extern void CCS_TRINAND3();
extern void CCS_TRINOR2();

extern int nchan_4[4][4];
extern int nchan_4d0[4][4];
extern int nchan_4d1[4][4];
extern int nchan_4d01[4][4];
extern int pchan_4[4][4];
extern int pchan_4d0[4][4];
extern int pchan_4d1[4][4];
extern int pchan_4d01[4][4];

extern int nchan_28[28][4];
extern int nchan_28d0[28][4];
extern int nchan_28d1[28][4];
extern int nchan_28d01[28][4];
extern int pchan_28[28][4];
extern int pchan_28d0[28][4];
extern int pchan_28d1[28][4];
extern int pchan_28d01[28][4];

/*--------------------------------------------------------------------------*/
/* The following macros implement the inline evaluation of simple boolean   */
/* functions.                                                               */
/*--------------------------------------------------------------------------*/
		  
                    
#define CCS_ANDNOR22(q, a, b, c, d) \
        (*(q)) = function_not[((a) & (b)) | ((c) & (d))]
        
#define CCS_BUF(q, a) (*(q)) = a

#define CCS_NOT(q, a) (*(q)) = function_not[a]

#define CCS_DELAY(q, a) (*(q)) = a

#define CCS_DAND8(q, qn, a, b, c, d, e, f, g, h) \
        (*(q)) = (a) & (b) & (c) & (d) & (e) & (f) & (g) & (h);\
        (*(qn)) = function_not[(*(q))]
        
#define CCS_DAND4(q, qn, a, b, c, d) \
        (*(q)) = (a) & (b) & (c) & (d);\
        (*(qn)) = function_not[(*(q))]
                
#define CCS_DAND3(q, qn, a, b, c) \
        (*(q)) = (a) & (b) & (c);\
        (*(qn)) = function_not[(*(q))]
                
#define CCS_DAND2(q, qn, a, b) \
        (*(q)) = (a) & (b);\
        (*(qn)) = function_not[(*(q))]
        
#define CCS_NAND8(qn, a, b, c, d, e, f, g, h) \
        (*(qn)) = function_not[(a) & (b) & (c) & (d) & (e) & (f) & (g) & (h)]
        
#define CCS_NAND4(qn, a, b, c, d) \
        (*(qn)) = function_not[(a) & (b) & (c) & (d)]
                
#define CCS_NAND3(qn, a, b, c) \
        (*(qn)) = function_not[(a) & (b) & (c)]
                
#define CCS_NAND2(qn, a, b) \
        (*(qn)) = function_not[(a) & (b)]

#define CCS_AND8(q, a, b, c, d, e, f, g, h) \
        (*(q)) = (a) & (b) & (c) & (d) & (e) & (f) & (g) & (h)

#define CCS_AND4(q, a, b, c, d) \
        (*(q)) = (a) & (b) & (c) & (d)

#define CCS_AND3(q, a, b, c) \
        (*(q)) = (a) & (b) & (c)
                
#define CCS_AND2(q, a, b) \
        (*(q)) = (a) & (b)
        
#define CCS_DOR8(q, qn, a, b, c, d, e, f, g, h) \
        (*(q)) = (a) | (b) | (c) | (d) | (e) | (f) | (g) | (h);\
        (*(qn)) = function_not[(*(q))]
        
#define CCS_DOR4(q, qn, a, b, c, d) \
        (*(q)) = (a) | (b) | (c) | (d);\
        (*(qn)) = function_not[(*(q))]
                
#define CCS_DOR3(q, qn, a, b, c) \
        (*(q)) = (a) | (b) | (c);\
        (*(qn)) = function_not[(*(q))]
                
#define CCS_DOR2(q, qn, a, b) \
        (*(q)) = (a) | (b);\
        (*(qn)) = function_not[(*(q))]

#define CCS_NOR8(qn, a, b, c, d, e, f, g, h) \
        (*(qn)) = function_not[(a) | (b) | (c) | (d) | (e) | (f) | (g) | (h)]
        
#define CCS_NOR4(qn, a, b, c, d) \
        (*(qn)) = function_not[(a) | (b) | (c) | (d)]
                
#define CCS_NOR3(qn, a, b, c) \
        (*(qn)) = function_not[(a) | (b) | (c)]
                
#define CCS_NOR2(qn, a, b) \
        (*(qn)) = function_not[(a) | (b)]

#define CCS_OR8(q, a, b, c, d, e, f, g, h) \
        (*(q)) = (a) | (b) | (c) | (d) | (e) | (f) | (g) | (h)
        
#define CCS_OR4(q, a, b, c, d) \
        (*(q)) = (a) | (b) | (c) | (d)

                
#define CCS_OR3(q, a, b, c) \
        (*(q)) = (a) | (b) | (c)
                
#define CCS_OR2(q, a, b) \
        (*(q)) = (a) | (b)
        
#define CCS_DXOR8(q, qn, a, b, c, d, e, f, g, h) \
        {\
        int r1,r2,r3,r4;\
        r1 = function_xor[(a)][(b)];\
        r2 = function_xor[(c)][(d)];\
        r3 = function_xor[(e)][(f)];\
        r4 = function_xor[(g)][(h)];\
        (*(q)) = function_xor[function_xor[r1][r2]][function_xor[r3][r4]];\
        (*(qn)) = function_not[(*(q))];\
        }
        
#define CCS_DXOR4(q, qn, a, b, c, d) \
        (*(q)) = function_xor[function_xor[(a)][(b)]][function_xor[(c)][(d)]];\
        (*(qn)) = function_not[(*(q))]
        
#define CCS_DXOR3(q, qn, a, b, c) \
        (*(q)) = function_xor[function_xor[(a)][(b)]][(c)];\
        (*(qn)) = function_not[(*(q))]
        
#define CCS_DXOR2(q, qn, a, b) \
        (*(q)) = function_xor[(a)][(b)];\
        (*(qn)) = function_not[(*(q))]
        
#define CCS_XOR8(q, a, b, c, d, e, f, g, h) \
        {\
        int r1,r2,r3,r4;\
        r1 = function_xor[(a)][(b)];\
        r2 = function_xor[(c)][(d)];\
        r3 = function_xor[(e)][(f)];\
        r4 = function_xor[(g)][(h)];\
        (*(q)) = function_xor[function_xor[r1][r2]][function_xor[r3][r4]];\
        }
        
#define CCS_XOR4(q, a, b, c, d) \
        (*(q)) = function_xor[function_xor[(a)][(b)]][function_xor[(c)][(d)]];


#define CCS_XOR3(q, a, b, c) \
        (*(q)) = function_xor[function_xor[(a)][(b)]][(c)];
        
#define CCS_XOR2(q, a, b) \
        (*(q)) = function_xor[(a)][(b)];
        
#define CCS_XNOR8(qn, a, b, c, d, e, f, g, h) \
        {\
        int r1,r2,r3,r4;\
        r1 = function_xnor[(a)][(b)];\
        r2 = function_xnor[(c)][(d)];\
        r3 = function_xnor[(e)][(f)];\
        r4 = function_xnor[(g)][(h)];\
        (*(qn)) = function_xnor[function_xnor[r1][r2]][function_xnor[r3][r4]];\
        }
        
#define CCS_XNOR4(qn, a, b, c, d) \
      (*(qn)) = function_xnor[function_xnor[(a)][(b)]][function_xnor[(c)][(d)]];


#define CCS_XNOR3(qn, a, b, c) \
        (*(qn)) = function_not[function_xor[function_xor[(a)][(b)]][(c)]];

        
#define CCS_XNOR2(qn, a, b) \
        (*(qn)) = function_xnor[(a)][(b)];

/*--------------------------------------------------------------------------*/
/* The following MACROs expand into the appropriate flip flop calls for     */
/* the various different combinations of active input levels for UDFDL      */
/* and UJKFF. In all cases, the default active input level is HIGH. If      */
/* an attribute specifically says it is low, then the inversion function is */
/* applied to that input before it's value is passed into the evaluation    */
/* routine.                                                                 */
/*                                                                          */
/* The naming conventions for UDFDL, UJKFFRE, and UJKFF are:                */
/*   <FT_>UDFDL_AA<H | L>AC<H | L>AP<H | L>                                 */
/*   <FT_>UJKFFRE_AA<H | L>                                                 */
/*   <FT_>UJKFF_AA<H | L>AC<H | L>                                          */
/*                                                                          */
/*   where AA is AASYNCLEV (Active Asynchronous Input Level)                */
/*         AC is ACLKEDGE (Active clock edge)                               */
/*         AP is APELEV (Active parallel enable level)                      */
/*         FT is FLOWTHRU (Latch works in "transparent mode")               */
/*   and H indicates the active level is high, L indicate low.              */
/*                                                                          */
/* So for example, a UDFDL with AASYNCLEV=0, ACLKEDGE=1, APELEV=0, and      */
/* FLOWTHRU=ON                                                              */
/* would cause the compiled code generator to generate a reference to:      */
/*                                                                          */
/* FT_UDFDL_AALACHAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE)     */
/*                                                                          */
/* And the macros defined below would expand this macro into the call:      */
/*                                                                          */
/* FT_UDFDL(Q,QN,D,function_not[S],function_not[R],CLK,function_not[PRE],   */
/*       function_not[CLR],PD,function_not[PE],O_QN,O_Q,O_CLK,O_PE);        */
/*                                                                          */
/*--------------------------------------------------------------------------*/

#define CCS_FT_UDFDL_AALACLAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         function_not[(CLK)],function_not[(PRE)],function_not[(CLR)],\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AALACLAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         function_not[(CLK)],function_not[(PRE)],function_not[(CLR)],\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AALACHAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         (CLK),function_not[(PRE)],function_not[(CLR)],\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AALACHAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         (CLK),function_not[(PRE)],function_not[(CLR)],\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AAHACLAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),(S),(R),\
         function_not[(CLK)],(PRE),(CLR),\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AAHACLAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),(S),(R),\
         function_not[(CLK)],(PRE),(CLR),\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AAHACHAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),(S),(R),\
         (CLK),(PRE),(CLR),\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_FT_UDFDL_AAHACHAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_FT_UDFDL((Q),(QN),(D),(S),(R),\
         (CLK),(PRE),(CLR),\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))

#define CCS_UDFDL_AALACLAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         function_not[(CLK)],function_not[(PRE)],function_not[(CLR)],\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AALACLAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         function_not[(CLK)],function_not[(PRE)],function_not[(CLR)],\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AALACHAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         (CLK),function_not[(PRE)],function_not[(CLR)],\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AALACHAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),function_not[(S)],function_not[(R)],\
         (CLK),function_not[(PRE)],function_not[(CLR)],\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AAHACLAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),(S),(R),\
         function_not[(CLK)],(PRE),(CLR),\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AAHACLAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),(S),(R),\
         function_not[(CLK)],(PRE),(CLR),\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AAHACHAPL(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),(S),(R),\
         (CLK),(PRE),(CLR),\
         (PD),function_not[(PE)],(O_QN),(O_Q),(O_CLK),(O_PE))
         
#define CCS_UDFDL_AAHACHAPH(Q,QN,D,S,R,CLK,PRE,CLR,PD,PE,O_QN,O_Q,O_CLK,O_PE) \
   CCS_UDFDL((Q),(QN),(D),(S),(R),\
         (CLK),(PRE),(CLR),\
         (PD),(PE),(O_QN),(O_Q),(O_CLK),(O_PE))

#define CCS_JKFFRE_AAL(Q,QN,J,K,S,R,CLK,O_CLK,O_Q,O_QN) \
  CCS_UJKFF((Q),(QN),(J),(K),function_not[(S)],function_not[(R)],(CLK), \
         HIGH4,HIGH4,(O_CLK),(O_Q),(O_QN))
   
#define CCS_JKFFRE_AAH(Q,QN,J,K,S,R,CLK,O_CLK,O_Q,O_QN) \
  CCS_UJKFF((Q),(QN),(J),(K),(S),(R),(CLK),LOW4,LOW4,(O_CLK),(O_Q),(O_QN))

#define CCS_UJKFF_AALACL(Q,QN,J,K,S,R,CLK,PRE,CLR,O_CLK,O_Q,O_QN) \
  CCS_UJKFF((Q),(QN),(J),(K),function_not[(S)],\
        function_not[(R)],function_not[(CLK)],\
        function_not[(PRE)],function_not[(CLR)],(O_CLK),(O_Q),(O_QN))
         
#define CCS_UJKFF_AALACH(Q,QN,J,K,S,R,CLK,PRE,CLR,O_CLK,O_Q,O_QN) \
  CCS_UJKFF((Q),(QN),(J),(K),function_not[(S)],\
        function_not[(R)],(CLK),\
        function_not[(PRE)],function_not[(CLR)],(O_CLK),(O_Q),(O_QN))
         
#define CCS_UJKFF_AAHACL(Q,QN,J,K,S,R,CLK,PRE,CLR,O_CLK,O_Q,O_QN) \
  CCS_UJKFF((Q),(QN),(J),(K),(S),\
        (R),function_not[(CLK)],\
        (PRE),(CLR),(O_CLK),(O_Q),(O_QN))
         
#define CCS_UJKFF_AAHACH(Q,QN,J,K,S,R,CLK,PRE,CLR,O_CLK,O_Q,O_QN) \
  CCS_UJKFF((Q),(QN),(J),(K),(S),\
        (R),(CLK),\
        (PRE),(CLR),(O_CLK),(O_Q),(O_QN))

#define CHAN(DOUT,DIN,GATE,TABLE) *(DOUT) = TABLE[(DIN)][(GATE)];

#define CCS_NCHAN_4(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_4)
#define CCS_NCHAN_4D0(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_4d0)
#define CCS_NCHAN_4D1(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_4d1)
#define CCS_NCHAN_4D01(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_4d01)
#define CCS_PCHAN_4(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_4)
#define CCS_PCHAN_4D0(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_4d0)
#define CCS_PCHAN_4D1(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_4d1)
#define CCS_PCHAN_4D01(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_4d01)

#define CCS_NCHAN_28(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_28)
#define CCS_NCHAN_28D0(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_28d0)
#define CCS_NCHAN_28D1(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_28d1)
#define CCS_NCHAN_28D01(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,nchan_28d01)
#define CCS_PCHAN_28(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_28)
#define CCS_PCHAN_28D0(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_28d0)
#define CCS_PCHAN_28D1(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_28d1)
#define CCS_PCHAN_28D01(DOUT,DIN,GATE) CHAN(DOUT,DIN,GATE,pchan_28d01)

#endif
