/***************************************************************************
*
*
*   Name: vdfubs.vs
*
*   Facility:  Viewdraw FUBS.
*
*   Functional Description:	 This contains viewscript code for implementing
*				 FUBs facility in Core.
*
*   Environment:  UNIX
*
*   Authors:  Murali Kudlugi 
*             Viewlogic Consulting Services
*
*   Creation Date : July 1992
*
*   Modified by:
*   Who            Version   When      What
*   VFE            5.2       6-94      Restructured to be nested; support file
*                                      locking and macro recording.
*
*   Copyright (c) 1987-1992 by Viewlogic Systems, Inc.
*
***************************************************************************/


/***************************************************************************
   create_fub - This function creates a FUB object in ViewDraw.  The user
                interface to create a FUB is the same as that for creating
                a box (i.e. user specifies two diagonally opposing points).
                In addition to that the user must specify a unique name to
                be assigned to the FUB.
***************************************************************************/
function create_fub()
{

FubName := "";

   /************************************************************************
      end_fub_box - The user has specified the second point for the FUB so
                    now go create the FUB given these locations.
    ***********************************************************************/
   function end_fub_box(display, event)
   {

      /* If we are not creating a FUB then do not continue.  This protects us
         from running the FubEnd command outside of the fub context.  It also
         protects us from the same FUB macro being run more than once in the
         same session and thus inadvertently modifying existing FUBs. */

      if (string_eq (FubName, "") == FALSE) {

         /* Finish dragging the FUB box. */
         finbox (display, event); 

         /* Create the FUB now that the points are finalized */
         status := fub(display, event, FubName);

         unbind (display);
         clear_binding_table ();

         win := vcEvent.Window (event);
         vcWin.ClearStatus(win);

         /* If FUB sucessfully created then allow user to move it */
         if (status == 0) {

            vcWin.AddStatus (win, "Move cursor to place the FUB");
            move_created_object(display, event);
            }
         else {
            vcWin.AddStatus(win,"Error: Unable to create FUB '" + FubName + "' ");
            }

         /* go back to fast_draw mode */
         vcCmd.Run(fast_draw_cmd, display, event, nil, TRUE); 

         /* We are done creating the FUB so reset the name */
         FubName = "";
         }
   }

   end_fub_cmd := vcWv.Cmd (fac, end_fub_box, "FubEnd");

   /************************************************************************
      start_fub_box - The user has entered the first FUB box point.  
                      Initialize the FUB box and bind the mouse buttons so
                      that the second point can be entered.
   ************************************************************************/
   function start_fub_box (display, event)
   {

      /* If we are not creating a FUB then do not continue.  This protects us
         from running the FubAdd command outside of the fub context.  It also
         protects us from the same FUB macro being run more than once in the
         same session and thus inadvertently modifying existing FUBs. */

      if (string_eq (FubName, "") == FALSE) {

         /* Initialize the box */
         if (initbox (display, event, getcolor (vqObject.Box),
                      getlstyle (vqObject.Box), SNAP) == 0) {

            status_msg (vcEvent.Window (event), "Specify points for FUB box");

            clear_binding_table ();

            /* Use the same bindings that are used for adding boxes. */
            vcBind.Set (bt, '\move',       drag_box_nosnap_cmd);
            vcBind.Set (bt, '\middledown', end_fub_cmd);
            vcBind.Set (bt, '\leftdown',   end_fub_cmd);
            vcBind.Set (bt, '\^[',         cancel_drag_box_cmd); 

            if (move_mode == 4)
                vcBind.Set (bt, '\rightdown',  cancel_drag_box_cmd);
            else
                vcBind.Set (bt, '\rightdown',  popup_box_menu_cmd);

            add_drag_bindings(bt, func(d, e) { contbox (d, e, NOSNAP); }, YES);

            bind (display, bt);
            }
         }
   }

   fub_box_cmd := vcWv.Cmd (fac, start_fub_box, "FubAdd");    

   /***********************************************************************
      checkSheetNum() - See if the incoming FUB name has a sheet number
                        other than 1.  If so generate a warning and default
                        to 1.  We do not want FUBs with different sheet 
                        numbers because those are only supposed to be used
                        for alternate graphics and FUB graphics are always
                        the same.  In addition, parts of the system require
                        at least a sheet 1 so we would not want a sheet 2
                        without a sheet 1.
   ***********************************************************************/
   function checkSheetNum(event, name)
   {

      /* Look for a dot character.  The extension follows. */
      dotIndex := rindex(name, '.');

      /* If a dot was found then check the extension against 1. */
      if (dotIndex != FALSE) {

         extIndex := dotIndex + 1;
         extStr := substring(name, extIndex, 0);

         /* If there is no extension then do nothing.  If the extension is not
            1 then warn the user and change it to 1. */

         if ((string_eq(extStr, "") == FALSE) && (string_eq(extStr, "1") == FALSE)) {

            vcErr.ReportFromString(vcEvent.Window(event), nil,
                                   "vdraw: Warning: FUBs cannot have an extension other than 1. Converting extension to 1.");

            newname := substring(name, 0, extIndex) + "1";
            name = string_copy(newname);
            }
         }

   name;
   }

   /***********************************************************************
      create_fub_box - This function initializes the creation of the FUB.
                       It validates the name of the FUB and binds the middle
                       mouse button to the fub_box_cmd.
   ************************************************************************/
   function create_fub_box(display, event, name)
   {   

      /* Make sure the name string passed in is not empty. */
      if (string_eq (name, "") == FALSE) {

         win := vcEvent.Window (event);  
         vcWin.ClearStatus (win);

         /* Make sure the sheet number is 1, else warn and default to 1. */
         name = checkSheetNum(event, name);

         /* Try to load the symbol. If found the user cannot create another 
            FUB by that name. */

         if (loadblock(name, vqDataType.Symbol) == 0) {
            vcWin.ClearStatus (win);
            vcWin.AddStatus (win, "Error: Symbol '" + name + "' already exists or cannot be created.");
            /* Set FubName to empty string so that we know we are not creating
               a legitimate FUB. */

            FubName = "";
            }

         else {  
      
            vcWin.ClearStatus (win);
            vcWin.AddStatus (win, "Draw the FUB box"); 

            /* Initiate fub_box_cmd to draw fub box */
            FubName = name; 
            bind_middle_button (fub_box_cmd); 
            }
         }
   }

   /*************************   start create_fub *************************/

   vcWv.Cmd (fac, FILE_CHECK1(create_fub_box), "fub", "Create FUB", 
             vcWv.Param("FUB Name", vlStringRestOfLine1VT, 
                        vcWv.Constant, "") ); 
}

fub_cmd := create_fub();

