/* Main file for ViewDatapath */
/* Table of contents:
   - types plus a couple of global sticky values
   - preference database and error functions
   - functions to shorten command declarations
   - functions to write the viewcell.cfg file
   - command definitions (long!)
   - add => cell command
   - change => cell command
   - prop => cell command
*/

/*** Types plus a couple of global sticky values ***/


vcWv.MakeEnum ("tidePinDirVT", list ("Input", "Output", "Bidir"));
vcWv.MakeEnum ("tideReqPinPolVT", list ("ActiveHigh", "ActiveLow"));
vcWv.MakeEnum ("tideOptPinPolVT", list ("ActiveHigh", "ActiveLow", "None"));
vcWv.MakeEnum ("tideBothPinPolVT", list ("ActiveHigh", "ActiveLow", "Both"));
vcWv.MakeEnum ("tideVStyleVT", list ("ViewSynth", "Synopsys", "XBLOX"));
vcWv.MakeEnum ("tideRAMTypeVT", list ("TransparentHigh", "TransparentLow",
   "RisingEdge"));
vcWv.MakeEnum ("tideUpDownVT", list ("UpOnly", "DownOnly", "ActiveHigh",
   "ActiveLow"));
vcWv.MakeEnum ("tideSignedVT", list ("TwosComp", "Unsigned"));
vcWv.MakeEnum ("tideCshTypeVT", list ("Barrel", "Arithmetic", "Logical"));
vcWv.MakeEnum ("tideShiftIOVT", list ("Serial", "Parallel"));
/* Caution: changing the order of the gate enumeration will break the change
   cell function, which converts an integer to this type */
vcWv.MakeEnum ("tideGateVT", list ("And","Or","Xor","Nand","Nor","Xnor"));


tide_last_width := 8;
tide_last_input := 4;
tide_last_depth := 16;
tide_last_count := 16;
declare tide_cell_file;
tide_cmd_popup := nil;


/*** Preference database ***/


/* Because using the default prefdb would write preference strings to 
   viewdraw.ini, which would be bad, create unique preference database */
tide_prefs := vlPref.Create ("vdpath");

vlPref.Add (tide_prefs, "vstyle", tideVStyleVT, "Style for VHDL models",
   tideVStyle.ViewSynth);
vlPref.Add (tide_prefs, "logcell", vlYesNoVT, "Log Viewcell output in window",
   vlYesNo.No);
vlPref.Add (tide_prefs, "autoname", vlYesNoVT, "Generate automatic symbol name",
   vlYesNo.Yes);
vlPref.Add (tide_prefs,"gatedel",   vlIntVT, "Generic gate delay",0);
vlPref.Add (tide_prefs,"c2qdel",    vlIntVT, "Generic clock to Q delay",0);
vlPref.Add (tide_prefs,"arithdel",  vlIntVT, "Generic arithmetic bit delay",0);
vlPref.Add (tide_prefs,"buswidth",  vlIntVT, "Maximum bus width",32);
vlPref.Add (tide_prefs,"arithwidth",vlIntVT, "Maximum arithmetic bus width",16);
vlPref.Add (tide_prefs,"inwidth",   vlIntVT, "Maximum number of inputs",8);
vlPref.Add (tide_prefs,"depth",     vlIntVT, "Maximum number of RAM words",256);

/* Get prefs from command line and vdpath.ini file */
path := vscript.vsSrchdirs ("vdpath.ini", vscript.vsFILE);
args := vcWv.args; 
vlPref.Parse (path, &args);


/*** Errors ***/


function tideError (num, text) {
   vlEx.Throw (vlEx.FAILURE,
      vlErr.Create (nil, "vdpath", num, vlErrSev.Error, text)); }


function tideErrorTwoline (num, text1, text2) {
   err := vlErr.Create (nil, "vdpath", num, vlErrSev.Error, text1);
   vlErr.AppendText (err, text2);
   vlEx.Throw (vlEx.FAILURE, err); }


function tideCellError ()
   tideErrorTwoline (1004, "ViewCell failed to create symbol",
      "(Props=>Cell menu, Show generator messages option displays errors)");


function tideArgError (text)
   tideErrorTwoline (1008, "Invalid parameter combination:", text);


/*** Functions to shorten command declarations ***/


function tide_symname (s) {
   prefval := vlPref.Val (tide_prefs, "autoname");
   if (prefval) {
      s1 := "Symbol prefix"; s2 := "Enter the prefix for output symbol name"; }
   else {
      s1 := "Symbol name"; s2 := "Enter the output symbol name"; }
   vcWv.Param (s1, vlStringVT, vcWv.Sticky, s, s2); }

function tide_buswidth () {
   vcWv.Param ("Bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         prefval := vlPref.Val (tide_prefs, "buswidth");
         if (tide_last_width < prefval) tide_last_width;
         else prefval; },
      vcScalebar.fac, "Enter the bus width",
      min <- 1, max <- vlPref.Val (tide_prefs, "buswidth")); }

function tide_arithwidth () {
   vcWv.Param ("Bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         prefval := vlPref.Val (tide_prefs, "arithwidth");
         if (tide_last_width < prefval) tide_last_width;
         else prefval; },
      vcScalebar.fac, "Enter the bus width",
      min <- 2, max <- vlPref.Val (tide_prefs, "arithwidth")); }

function tide_inputwidth () {
   vcWv.Param ("Number of inputs", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         prefval := vlPref.Val (tide_prefs, "inwidth");
         if (tide_last_input < prefval) tide_last_input;
         else prefval; },
      vcScalebar.fac, "Enter the number of inputs",
      min <- 2, max <- vlPref.Val (tide_prefs, "inwidth")); }

function tide_depth () {
   vcWv.Param ("Number of words", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         prefval := vlPref.Val (tide_prefs, "depth");
         if (tide_last_depth < prefval) tide_last_depth;
         else prefval; },
      vcScalebar.fac, "Enter the number of words",
      min <- 2, max <- vlPref.Val (tide_prefs, "depth")); }

function tide_reqpol (s) {
   vcWv.Param (s, tideReqPinPolVT, vcWv.Sticky,
      tideReqPinPol.ActiveHigh, "Select " + s + " polarity: high or low"); }

function tide_optpol (s) {
   vcWv.Param (s, tideOptPinPolVT, vcWv.Sticky, tideOptPinPol.None,
      "Select " + s + " pin presence and polarity"); }

function tide_signed () {
   vcWv.Param ("Representation", tideSignedVT, vcWv.Sticky, tideSigned.Unsigned,
      "Select signed or unsigned representation"); }


/*** Functions to shorten change command declarations ***/


/* Function to debug change command declarations.  Should be the last line
   of every dynamic dialog box function. */
function tide_done (d, funcname, val) {
   block := DisplayGetBlock (d);
   symblock := CompGetBlock (block, BlockGetSelObj (block, vqFlag.Component));
   blockname := ObjGetName (block, symblock, TRUE);
   printf ("VDPATH_DEBUG %s %s %s\n", blockname, funcname, val);
   val; }

/* Construct actual name from partial_name and bus width, and return pin */
function tide_make_bus_name (block, obj, partial_name) {
   width_attr := ObjFindAttr (block, obj, "WIDTH");
   if (null (width_attr))
      width_attr = ObjFindAttr (block, CompGetBlock (block, obj), "WIDTH");
   if (null (width_attr)) tideError (1001,
      "No parameter 'WIDTH' on selected symbol");
   else width := string2number (AttrGetValue (block, width_attr)) - 1;
   if (width == 0) name := sprintf ("%s0", partial_name);
   else name := sprintf ("%s[%s:0]", partial_name, width);

   pin := BlockFindPinByName (CompGetBlock (block, obj), name);
   if (null (pin)) tideError (1003,
      sprintf ("No required pin '%s' on selected symbol", name));
   pin; }

/* Find RAM clocking type */
function tide_ram_find_clock (d) { declare retval;
   block := DisplayGetBlock (d);
   symblock := CompGetBlock (block, BlockGetSelObj (block, vqFlag.Component));

   /* Has pin named clock? */
   pin := BlockFindPinByName (symblock, "CLOCK");
   if (!null (pin)) retval = tideRAMType.RisingEdge;
   else {
      pin = BlockFindPinByName (symblock, "WE");
      if (null (pin)) tideError (1003,
         "No required pin 'WE' on selected symbol");
      pol_attr := ObjFindAttr (block, pin, "LPM_POLARITY");
      if (null (pol_attr)) tideError (1002,
         "No polarity for pin 'WE' on selected symbol");
      if (string_eq (AttrGetValue (block, pol_attr), "INVERT"))
         retval = tideRAMType.TransparentLow;
      else retval = tideRAMType.TransparentHigh; }
   tide_done (d, "trfc", retval); }

/* Find shift type for combinational shift */
function tide_find_shift (d) {
   block := DisplayGetBlock (d);
   symblock := CompGetBlock (block, BlockGetSelObj (block, vqFlag.Component));
   param_attr := ObjFindAttr (block, symblock, "SHIFTTYPE");
   if (null (param_attr))
      tideError (1001, "No parameter 'SHIFTTYPE' on selected symbol");
   param_val := AttrGetValue (block, param_attr);
   if (string_eq (param_val, "ARITHMETIC")) retval := tideCshType.Arithmetic;
   else if (string_eq (param_val, "ROTATE")) retval := tideCshType.Barrel;
   else if (string_eq (param_val, "LOGICAL")) retval := tideCshType.Logical;
   tide_done (d, "cshiftype", retval); }

/* Find up/down type for counter or combinational shift */
function tide_find_up (name, d) {
   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);
   pin := BlockFindPinByName (CompGetBlock (block, obj), name);
   if (null (pin)) retval := tideUpDown.UpOnly;
   else {
      pol_attr := ObjFindAttr (block, pin, "LPM_POLARITY");
      if (null (pol_attr)) tideError (1002,
         sprintf ("No polarity for pin '%s' on selected symbol", name));
      if (string_eq (AttrGetValue (block, pol_attr), "INVERT"))
         retval := tideUpDown.ActiveLow;
      else retval := tideUpDown.ActiveHigh; }
   tide_done (d, "cshiftdir", retval); }

/* Find type of gate (and/or/xor) by looking at LPM_TYPE and output inversion */
function tide_gate_type (d) {

   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);

   /* Extract polarity from symbol */
   pin := tide_make_bus_name (block, obj, "RESULT");
   pol_attr := ObjFindAttr (block, pin, "LPM_POLARITY");
   if (null (pol_attr)) tideError (1002,
      "No polarity for pin 'RESULT' on selected symbol");
   if (string_eq (AttrGetValue (block, pol_attr), "INVERT"))
      index := 3;
   else index := 0;

   /* Extract gate type from symbol */
   type_attr := ObjFindAttr (block, obj, "LPM_TYPE");
   if (null (type_attr))
      type_attr = ObjFindAttr (block, CompGetBlock (block, obj), "LPM_TYPE");
   if (null (type_attr)) tideError (1006,
      "No LPM_TYPE= attribute on selected symbol");
   type_str := AttrGetValue (block, type_attr);
   if (string_eq (type_str, "LPM_OR")) index := index + 1;
   else if (string_eq (type_str, "LPM_XOR")) index := index + 2;

   /* Now index is set to proper value of enumeration: return it */
   tide_done (d, "gatetype", convert (tideGateVT, index)); }

/* If shift register has pin with name "name", it is serial, otherwise it is
   parallel */
function tide_get_serpar (d, name) {
   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);
   if (null (BlockFindPinByName (CompGetBlock (block, obj), name)))
      retval := tideShiftIO.Parallel;
   else retval := tideShiftIO.Serial;
   tide_done (d, "serpar", retval); }

/* Function to get value of some parameter from symbol */
function tide_get_param (d, name) {
   block := DisplayGetBlock (d);
   symblock := CompGetBlock (block, BlockGetSelObj (block, vqFlag.Component));
   param_attr := ObjFindAttr (block, symblock, name);
   if (null (param_attr)) 
      tideError (1001, sprintf ("No parameter '%s' on selected symbol", name));
   else tide_done (d, "param", AttrGetValue (block, param_attr)); }

/* Function to extract symbol name */
function tide_get_symname (d) {
   block := DisplayGetBlock (d);
   compblock := CompGetBlock (block,
      BlockGetSelObj (block, vqFlag.Component));
   tide_done (d, "symname", ObjGetName (compblock, compblock, TRUE)); }

function tide_chgsymname () {
   vcWv.Param ("Symbol name", vlStringVT, vcWv.Dynamic, tide_get_symname,
      "Enter the output symbol name"); }

function tide_chgbuswidth () {
   vcWv.Param ("Bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "WIDTH");,
      vcScalebar.fac, "Enter the bus width",
      min <- 1, max <- vlPref.Val (tide_prefs, "buswidth")); }

function tide_chgarithwidth () {
   vcWv.Param ("Bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "WIDTH");,
      vcScalebar.fac, "Enter the bus width",
      min <- 2, max <- vlPref.Val (tide_prefs, "arithwidth")); }

function tide_chginputwidth () {
   vcWv.Param ("Number of inputs", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "SIZE");,
      vcScalebar.fac, "Enter the number of inputs",
      min <- 2, max <- vlPref.Val (tide_prefs, "inwidth")); }

function tide_chgdepth () {
   vcWv.Param ("Number of words", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "NUMWORDS");,
      vcScalebar.fac, "Enter the number of words",
      min <- 2, max <- vlPref.Val (tide_prefs, "depth")); }

function tide_getoptpol (d, name) {
   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);
   pin := BlockFindPinByName (CompGetBlock (block, obj), name);
   if (null (pin)) retval := tideOptPinPol.None;
   else {
      pol_attr := ObjFindAttr (block, pin, "LPM_POLARITY");
      if (null (pol_attr)) tideError (1002,
         sprintf ("No polarity for pin '%s' on selected symbol", name));
      if (string_eq (AttrGetValue (block, pol_attr), "INVERT"))
         retval := tideOptPinPol.ActiveLow;
      else retval := tideOptPinPol.ActiveHigh; }
   tide_done (d, "optpol", retval); }

function tide_chgoptpol (s, name) {
   vcWv.Param (s, tideOptPinPolVT, vcWv.Dynamic,
      func (d) tide_getoptpol (d, name);,
      "Select " + s + " pin presence and polarity"); }

function tide_getreqpol (d, name) {
   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);
   pin := BlockFindPinByName (CompGetBlock (block, obj), name);
   if (null (pin)) tideError (1003,
      sprintf ("No required pin '%s' on selected symbol", name));
   pol_attr := ObjFindAttr (block, pin, "LPM_POLARITY");
   if (null (pol_attr)) tideError (1002,
      sprintf ("No polarity for pin '%s' on selected symbol", name));
   if (string_eq (AttrGetValue (block, pol_attr), "INVERT"))
      retval := tideOptPinPol.ActiveLow;
   else retval := tideOptPinPol.ActiveHigh;
   tide_done (d, "reqpol", retval); }

function tide_chgreqpol (s, name) {
   vcWv.Param (s, tideReqPinPolVT, vcWv.Dynamic, 
      func (d) tide_getreqpol (d, name);,
      "Select " + s + " polarity: high or low"); }

function tide_chgsigned () {
   vcWv.Param ("Representation", tideSignedVT, vcWv.Dynamic,
      func (d) {
         if (string_eq (tide_get_param (d, "REPRESENTATION"), "UNSIGNED"))
            tideSigned.Unsigned;
         else tideSigned.TwosComp; },
      "Select signed or unsigned representation"); }

function tide_getrpw (d, partial_name) {
   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);

   /* Construct actual name from partial_name and bus width */
   pin := tide_make_bus_name (block, obj, partial_name);

   pol_attr := ObjFindAttr (block, pin, "LPM_POLARITY");
   if (null (pol_attr)) tideError (1002,
      sprintf ("No polarity for pin '%s' on selected symbol",
         partial_name));
   if (string_eq (AttrGetValue (block, pol_attr), "INVERT"))
      retval := tideOptPinPol.ActiveLow;
   else retval := tideOptPinPol.ActiveHigh;
   tide_done (d, "rpwid", retval); }

function tide_chgreqpolwid (s, partial_name) {
   vcWv.Param (s, tideReqPinPolVT, vcWv.Dynamic,
      func (d) tide_getrpw (d, partial_name);,
      "Select " + s + " polarity: high or low"); }


function tide_int2binstr (x) {
   if (x == 0) "0";
   else {
      s := sprintf ("%o", x);
      t := make_string (0);
      for (i:=0; i<string_length(s); ++i) {
         if      (s[i] == '0') t = t + "000";
         else if (s[i] == '1') t = t + "001";
         else if (s[i] == '2') t = t + "010";
         else if (s[i] == '3') t = t + "011";
         else if (s[i] == '4') t = t + "100";
         else if (s[i] == '5') t = t + "101";
         else if (s[i] == '6') t = t + "110";
         else if (s[i] == '7') t = t + "111"; }

      /* Remove leading zeroes: note special case x==0 already covered */
      for (; t[0] == '0';) t = substring (t, 1, 0);
      t; } }


/*** Functions to write the viewcell.cfg file ***/


/* Public function to start the file */
function initcfg () {
   function putstyle (view, synop, xblox, basetype) {
      putintcfg ("S_VIEW", view);
      putintcfg ("S_SYNOP", synop);
      putintcfg ("S_XBLOX", xblox);
      putstrcfg ("BASETYPE", basetype); }
   tide_cell_file = fopen ("viewcell.cfg", "w");
   fputline (tide_cell_file, "int I\nint J\nint K\n");
   prefval := vlPref.Val (tide_prefs, "vstyle");

   /* Note: if the base type is changed, change the dup definition for mux */
   if (prefval == tideVStyle.ViewSynth) putstyle (1,0,0,"vlbit");
   else if (prefval == tideVStyle.Synopsys) putstyle (0,1,0,"std_logic");
   else if (prefval == tideVStyle.XBLOX) putstyle (0,0,1,"vlbit"); }


/* Private function to put pin polarity into file */
function putpincfg (name, ishigh) {
   fputline (tide_cell_file, "int INV_" + name + "\nINV_" + name + " = ");
   if (ishigh) fputline (tide_cell_file, "0\n");
   else fputline (tide_cell_file, "1\n");
   fputline (tide_cell_file, "string LP_" + name + "\nLP_" + name + " = \"");
   if (ishigh) fputline (tide_cell_file, "NORMAL\"\n");
   else fputline (tide_cell_file, "INVERT\"\n"); }


/* Public function to put a required control pin into file */
function putreqcfg (name, reqvt) {
   putpincfg (name, (reqvt == tideReqPinPol.ActiveHigh)); }


/* Public function to put an optional control pin into file */
function putoptcfg (name, optvt) {
   fputline (tide_cell_file, "int HAS_" + name + "\nHAS_" + name + " = ");
   if (optvt == tideOptPinPol.None) fputline (tide_cell_file, "0\n");
   else fputline (tide_cell_file, "1\n");
   putpincfg (name, (optvt == tideOptPinPol.ActiveHigh)); }


/* Public function to put an arbitrary integer name/value pair into file */
function putintcfg (name, val) {
   fputline (tide_cell_file, sprintf ("int %s\n%s = %d\n", name, name, val)); }


/* Public function to put an arbitrary string name/value pair into file */
function putstrcfg (name, val) {
   fputline (tide_cell_file, sprintf ("string %s\n%s = \"%s\"\n",
      name, name, val)); }

/* Public function to put the log2 and log2 - 1 of a value into file */
function putlogcfg (log2name, log2l1name, width) {
   fputline (tide_cell_file, sprintf ("int %s\n%s = conv2lines (%d)\n",
      log2name, log2name, width));
   fputline (tide_cell_file, sprintf ("int %s\n%s = conv2lines (%d) - 1\n",
      log2l1name, log2l1name, width)); }

/* Public function to finish the file */
function finishcfg (d, e, symtemplate, vhdtemplate, symname, ischange, suffix) {

   function completion_func_add (m, c) {
      if ((m != vcSysMode.Exited) || (c != 0)) tideCellError ();
      else {
         if (!vlPref.Val (tide_prefs, "logcell"))
            vscript.vsDelete ("viewcell.cfg", vscript.vsFILE);
         vcCmd.Run (comp_cmd, d, e, list (symname), FALSE); } }

   function completion_func_change (m, c) {
      if ((m != vcSysMode.Exited) || (c != 0)) tideCellError ();
      else {
         if (!vlPref.Val (tide_prefs, "logcell"))
            vscript.vsDelete ("viewcell.cfg", vscript.vsFILE);
         BlockRead (BlockOpen (nil, symname, "1", "Symbol"));

         /* Refresh all the windows: see vdr_nav.vs for source */
         foreach (w; vcWin.Loop ()) {
            if (null (vcWin.Parent (w))) {
               r := vcWin.OwnerRegion (w); 
               if (vcRegion.Fac (r) == vdraw.fac)
                  vrefresh (vcRegion.GetData (r)); } } } }

   function execute_func () {
      complete_func := ischange ? completion_func_change : completion_func_add;
      vcWin.System (vcEvent.Window (e), "viewcell viewcell.cfg -synthesize",
         nil, TRUE, output_type, nil, complete_func, complete_func, nil); }

   /* Determine type of output */
   if (vlPref.Val (tide_prefs, "logcell")) output_type := vcSysOutput.Viewtext;
   else output_type := vcSysOutput.NoOutput;

   /* Optionally add suffix for width to symname */
   if (!ischange && (vlPref.Val (tide_prefs, "autoname")))
      symname = sprintf ("%s%d", symname, suffix);

   fputline (tide_cell_file, sprintf ("string MODNAME\nMODNAME = \"%s\"\n",
      string_upper (symname)));
   fputline (tide_cell_file,sprintf ("\"vdpath:behv/%s.vhd\" \"behv/%s.vhd\"\n",
      vhdtemplate, symname));
   fputline (tide_cell_file, sprintf ("\"vdpath:sym/%s.1\" \"sym/%s.1\"\n",
      symtemplate, symname));
   fclose (tide_cell_file);

   /* Check to make sure template symbol exists: most common error */
   if (!DataObjExists ("vdpath", symtemplate, "1", vqObjClass.Symbol))
      tideError (1009, "Template symbol vdpath:" + symtemplate +
         " not in ViewDraw search path");

   /* Check to see if symbol already exists; run system command */
   if (!ischange && (DataObjExists (nil, symname, "1", vqObjClass.Symbol))) {
      table := vcTable.Create ("OK to overwrite file?", 2, 1);
      vcTable.Add (table, vcRegion.Create ("msg", vcStaticText.fac,
         "Symbol " + symname + " already exists.  Are you sure?"));
      buttons := vcTable.Create ("buttons", 1, 2);
      vcTable.Add (buttons, vcWv.LabelButtonCreate ("yes", "Yes", nil,
         execute_func, list (close_window <- TRUE)));
      no_button := vcWv.LabelButtonCreate ("no", "No", nil, func() {;},
         list (close_window <- TRUE));
      vcTable.Add (buttons, no_button);
      vcTable.Add (table, buttons);

      /* Pop up confirmation dialog box to avoid writing over symbol */
      w := vcEvent.Window (e);
      popup := vcWin.Popup (w, table, no_button,
         vcWin.NoMenus + vcWin.NoCmdline + vcWin.NoStatus);
      vcWin.Forward (w, popup); }
   else execute_func (); }


/*** Command definitions ***/


function tide_add_chk (d, e, name, width, add, cin, cout, ovf) {
   if ((cout != tideOptPinPol.None) && (ovf != tideOptPinPol.None))
      tideArgError ("Carry out and overflow are mutually exclusive"); }


function tide_add_sub_func (d, e, name, width, add, cin, cout, ovf, ischange) {
   tide_last_width = width;

   initcfg ();
   putoptcfg ("ADDSUB", add);
   putoptcfg ("CIN", cin);
   putoptcfg ("COUT", cout);
   putoptcfg ("OVFL", ovf); 
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width-1);
   putintcfg ("N_WP1", width+1);
   putintcfg ("N_DELAY", width * vlPref.Val (tide_prefs, "arithdel"));

   /* Compute X value with length W for assignment when add_sub is X */
   putstrcfg ("X_W", string_fill (make_string (width), 'X'));

   finishcfg (d, e, "add_sub", "add_sub", name, ischange, width); }


tide_add_sub_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, add, cin, cout, ovf)
      tide_add_sub_func (d, e, name, width, add, cin, cout, ovf, FALSE);,
   "_tide_add_sub", "Add/Sub",
   tide_symname ("add_sub"), tide_arithwidth (), 
   tide_optpol ("Add control"), tide_optpol ("Carry in"),
   tide_optpol ("Carry out"), tide_optpol ("Overflow output"));


tide_add_sub_chg := vcWv.Cmd (fac,
   func (d, e, name, width, add, cin, cout, ovf)
      tide_add_sub_func (d, e, name, width, add, cin, cout, ovf, TRUE);,
   "_tide_add_sub_chg",  "Add/Sub",
   tide_chgsymname (), tide_chgarithwidth (), 
   tide_chgoptpol ("Add control", "ADD_SUB"),
   tide_chgoptpol ("Carry in", "CIN"), tide_chgoptpol ("Carry out", "COUT"),
   tide_chgoptpol ("Overflow output", "OVERFLOW"));


vcCmd.SetHelp (tide_add_sub_cmd,
  "Create adder or add/subtract symbol and model");
vcCmd.EnableApply (tide_add_sub_cmd, TRUE);
vcCmd.VerifyParams (tide_add_sub_cmd, tide_add_chk);
vcCmd.VerifyParams (tide_add_sub_chg, tide_add_chk);


/*** Comparator ***/


function tide_compare_chk (d, e, name, width, aeb, alb, agb, fmt) {
   if ((aeb == tideOptPinPol.None) && (alb == tideOptPinPol.None) &&
      (agb == tideOptPinPol.None))
         tideArgError ("One of equal, less or greater pins must be used"); }


function tide_compare_func (d, e, name, width, aeb, alb, agb, fmt, ischange) {
   tide_last_width = width;

   initcfg ();
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width - 1);
   putintcfg ("N_DELAY", width * vlPref.Val (tide_prefs, "arithdel"));
   putoptcfg ("ALB", alb);
   putoptcfg ("AEB", aeb);
   putoptcfg ("AGB", agb);
   if (fmt == tideSigned.Unsigned) putstrcfg ("REPRESENTATION","UNSIGNED");
   else putstrcfg ("REPRESENTATION", "SIGNED");
   if ((alb == tideOptPinPol.None) && (agb == tideOptPinPol.None))
      putintcfg ("HAS_AEB_ONLY", 1);
   else putintcfg ("HAS_AEB_ONLY", 0);

   finishcfg (d, e, "compare", "compare", name, ischange, width); }


tide_compare_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, aeb, alb, agb, fmt)
      tide_compare_func (d, e, name, width, aeb, alb, agb, fmt, FALSE);,
   "_tide_compare",  "Compare",
   tide_symname ("compare"), tide_arithwidth (), tide_optpol ("A = B"),
   tide_optpol ("A < B"), tide_optpol ("A > B"), tide_signed ());


tide_compare_chg := vcWv.Cmd (fac,
   func (d, e, name, width, aeb, alb, agb, fmt)
      tide_compare_func (d, e, name, width, aeb, alb, agb, fmt, TRUE);,
   "_tide_compare_chg",  "Compare",
   tide_chgsymname (), tide_chgarithwidth (), tide_chgoptpol ("A = B", "AEB"),
   tide_chgoptpol ("A < B", "ALB"), tide_chgoptpol ("A > B", "AGB"),
   tide_chgsigned ());


vcCmd.SetHelp (tide_compare_cmd, "Create comparator symbol and model");
vcCmd.EnableApply (tide_compare_cmd, TRUE);
vcCmd.VerifyParams (tide_compare_cmd, tide_compare_chk);
vcCmd.VerifyParams (tide_compare_chg, tide_compare_chk);


/*** Counter ***/


function tide_count_chk (d, e, name, mod, ud, clk, ce, ac, al, sc, sl, tc) {
   if ((al != tideOptPinPol.None) && (ac != tideOptPinPol.None))
      tideArgError ("Asynchronous load and clear are mutually exclusive");
   if ((vlPref.Val (tide_prefs, "vstyle")) == tideVStyle.XBLOX) {
      if (al != tideOptPinPol.None) 
         tideArgError ("XBLOX does not allow asynchronous load");
      s := tide_int2binstr (mod - 1);
      if (string_search (s, "0"))
         tideArgError ("XBLOX only allows powers of two as modulus"); } }


function tide_count_func (d,e,name,mod,ud,clk,ce,ac,al,sc,sl,tc,ischg) {
      
   function put_control (u,d) {
      putintcfg ("S_UP", u); putintcfg ("S_DOWN", d); }

   tide_last_count = mod;

   initcfg ();
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putlogcfg ("N_W",     "N_WL1", mod);
   putintcfg ("N_MOD",   mod);
   putintcfg ("N_MODL1", mod - 1);
   putreqcfg ("CLOCK",   clk);
   putoptcfg ("ENABLE",  ce);
   putoptcfg ("ACLR",    ac);
   putoptcfg ("ALOAD",   al);
   putoptcfg ("SCLR",    sc);
   putoptcfg ("SLOAD",   sl);
   putoptcfg ("TC",      tc);

   /* Put zero and modulus values.  Can only tell length of zero vector by
      looking at length of mod string. */
   s := tide_int2binstr (mod - 1);
   width := string_length (s);
   putstrcfg ("B_MODL1", s);
   putstrcfg ("B_0",     string_fill (make_string (width), '0'));

   /* If synchronous clear is required because of a weird modulus, set MLOAD
      flag.  The modulus is "weird" if it isn't a power of two, that is, if
      the binary representation of (mod less 1) contains any zeroes. */
   if (!string_search (s, "0")) putintcfg ("HAS_MLOAD", 0);
   else putintcfg ("HAS_MLOAD", 1);

   /* Break up control enumeration */
   if (ud == tideUpDown.UpOnly) {
      putoptcfg ("UPDOWN", tideOptPinPol.None); put_control (1,0); }
   else if (ud == tideUpDown.DownOnly) {
      putoptcfg ("UPDOWN", tideOptPinPol.None); put_control (0,1); }
   else if (ud == tideUpDown.ActiveHigh) {
      putoptcfg ("UPDOWN", tideOptPinPol.ActiveHigh); put_control (1,1); }
   else { putoptcfg ("UPDOWN", tideOptPinPol.ActiveLow); put_control (1,1); }

   /* Determine if data pin present */
   if ((al == tideOptPinPol.None) && (sl == tideOptPinPol.None))
      putintcfg ("HAS_DATA", 0);
   else putintcfg ("HAS_DATA", 1);

   putstrcfg ("CLEVEL", (clk == tideReqPinPol.ActiveLow) ? "'0'" : "'1'");
   putstrcfg ("RLEVEL", (ac  == tideOptPinPol.ActiveLow) ? "'0'" : "'1'");
   putstrcfg ("LLEVEL", (al  == tideOptPinPol.ActiveLow) ? "'0'" : "'1'");

   finishcfg (d, e, "count", "count", name, ischg, width); }


tide_count_cmd := vcWv.Cmd (fac,
   func (d,e,name,mod,ud,clk,ce,ac,al,sc,sl,tc)
      tide_count_func (d,e,name,mod,ud,clk,ce,ac,al,sc,sl,tc,FALSE);,
   "_tide_count",  "Counter",
   tide_symname ("count"), 
   vcWv.Param ("Count modulus", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         if (tide_last_count < 65536) tide_last_count;
         else 65536; },
      vcScalebar.fac, "Enter the maximum count value",
      min <- 2, max <- 65536),
   vcWv.Param ("Count-up control", tideUpDownVT, vcWv.Sticky,
      tideUpDown.UpOnly, "Select the count direction control method"),
   tide_reqpol ("Clock"), tide_optpol ("Count enable"),
   tide_optpol ("Async clear"), tide_optpol ("Async load"),
   tide_optpol ("Sync clear"), tide_optpol ("Sync load"),
   tide_optpol ("Terminal count"));


tide_count_chg := vcWv.Cmd (fac,
   func (d,e,name,mod,ud,clk,ce,ac,al,sc,sl,tc)
      tide_count_func (d,e,name,mod,ud,clk,ce,ac,al,sc,sl,tc,TRUE);,
   "_tide_count_chg",  "Counter",
   tide_chgsymname (), 
   vcWv.Param ("Count modulus", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "MODULUS");,
      vcScalebar.fac, "Enter the maximum count value",
      min <- 2, max <- 65536),
   vcWv.Param ("Count-up control", tideUpDownVT, vcWv.Dynamic,
      func (d) tide_find_up ("UPDOWN", d);,
      "Select the count direction control method"),
   tide_chgreqpol ("Clock", "CLOCK"), tide_chgoptpol ("Count enable", "ENABLE"),
   tide_chgoptpol ("Async clear", "ACLR"),
   tide_chgoptpol ("Async load", "ALOAD"),
   tide_chgoptpol ("Sync clear", "SCLR"), tide_chgoptpol ("Sync load", "SLOAD"),
   tide_optpol ("Terminal count"));


vcCmd.SetHelp (tide_count_cmd, "Create up or up/down counter symbol and model");
vcCmd.EnableApply (tide_count_cmd, TRUE);
vcCmd.VerifyParams (tide_count_cmd, tide_count_chk);
vcCmd.VerifyParams (tide_count_chg, tide_count_chk);


/*** Combinational shift ***/


function tide_cshift_func (d, e, name, width, type, control, ischange) {
   function put_type (b,a,l) {
      putintcfg ("S_BARREL", b); putintcfg ("S_ARITH", a);
      putintcfg ("S_LOGICAL", l); }
   function put_control (u,d) {
      putintcfg ("S_UP", u); putintcfg ("S_DOWN", d); }

   tide_last_width = width;
   initcfg ();
   putintcfg ("N_W", width);  putintcfg ("N_WL1", width-1);
   putlogcfg ("N_K", "N_KL1", width);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "gatedel"));
   putintcfg ("DIST_2",  (width >  2) ? 1 : 0);
   putintcfg ("DIST_4",  (width >  4) ? 1 : 0);
   putintcfg ("DIST_8",  (width >  8) ? 1 : 0);
   putintcfg ("DIST_16", (width > 16) ? 1 : 0);

   /* Break up type enumeration */
   if (type == tideCshType.Barrel) {
      put_type (1,0,0); putstrcfg ("SHIFTTYPE", "ROTATE"); }
   else if (type == tideCshType.Arithmetic) {
      put_type (0,1,0); putstrcfg ("SHIFTTYPE", "ARITHMETIC"); }
   else {
      put_type (0,0,1); putstrcfg ("SHIFTTYPE", "LOGICAL"); }

   /* Break up control enumeration */
   if (control == tideUpDown.UpOnly) {
      putoptcfg ("DIRECTION", tideOptPinPol.None); put_control (1,0); }
   else if (control == tideUpDown.DownOnly) {
      putoptcfg ("DIRECTION", tideOptPinPol.None); put_control (0,1); }
   else if (control == tideUpDown.ActiveHigh) {
      putoptcfg ("DIRECTION", tideOptPinPol.ActiveHigh); put_control (1,1); }
   else { putoptcfg ("DIRECTION", tideOptPinPol.ActiveLow); put_control (1,1); }

   finishcfg (d, e, "clshift", "clshift", name, ischange, width); }


tide_cshift_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, type, control)
      tide_cshift_func (d, e, name, width, type, control, FALSE);,
   "_tide_cshift", "Combinational shift",
   tide_symname ("clshift"), tide_arithwidth (),
   vcWv.Param ("Shift method", tideCshTypeVT, vcWv.Sticky, tideCshType.Barrel,
      "Select the type of shifting"),
   vcWv.Param ("Shift-up control", tideUpDownVT, vcWv.Sticky,
      tideUpDown.UpOnly, "Select the shift direction control method"));


tide_cshift_chg := vcWv.Cmd (fac,
   func (d, e, name, width, type, control)
      tide_cshift_func (d, e, name, width, type, control, TRUE);,
   "_tide_cshift_chg", "Combinational shift",
   tide_chgsymname (), tide_chgarithwidth (),
   vcWv.Param ("Shift method", tideCshTypeVT, vcWv.Dynamic, tide_find_shift,
      "Select the type of shifting"),
   vcWv.Param ("Shift-up control", tideUpDownVT, vcWv.Dynamic,
      func (d) tide_find_up ("DIRECTION", d);,
      "Select the shift direction control method"));


vcCmd.SetHelp (tide_cshift_cmd, "Create combinational shift symbol and model");
vcCmd.EnableApply (tide_cshift_cmd, TRUE);


/*** Decoder ***/


function tide_decode_func (d, e, name, width, enable, ischange) {
   tide_last_width = width;
   initcfg ();
   putintcfg ("N_W", width);  putintcfg ("N_WL1", width-1);
   putlogcfg ("N_K", "N_KL1", width);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "gatedel"));
   putoptcfg ("ENABLE", enable);

   putintcfg ("DIST_2",   (width >   1) ? 1 : 0);
   putintcfg ("DIST_4",   (width >   2) ? 1 : 0);
   putintcfg ("DIST_8",   (width >   4) ? 1 : 0);
   putintcfg ("DIST_16",  (width >   8) ? 1 : 0);
   putintcfg ("DIST_32",  (width >  16) ? 1 : 0);

   if      (width <=  2) ceiling :=   1;
   else if (width <=  4) ceiling :=   3;
   else if (width <=  8) ceiling :=   7;
   else if (width <= 16) ceiling :=  15;
   else if (width <= 32) ceiling :=  31;
   putintcfg ("N_CEILING", ceiling);
   putlogcfg ("N_KC", "N_KCL1", ceiling);

   finishcfg (d, e, "decode", "decode", name, ischange, width); }


tide_decode_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, enable)
      tide_decode_func (d, e, name, width, enable, FALSE);,
   "_tide_decode", "Decoder",
   tide_symname ("decode"), tide_arithwidth (), tide_optpol ("Enable"));


tide_decode_chg := vcWv.Cmd (fac,
   func (d, e, name, width, enable)
      tide_decode_func (d, e, name, width, enable, TRUE);,
   "_tide_decode_chg", "Decoder",
   tide_chgsymname (), tide_chgarithwidth (),
   tide_chgoptpol ("Enable", "ENABLE"));


vcCmd.SetHelp (tide_decode_cmd,   "Create decoder symbol and model");
vcCmd.EnableApply (tide_decode_cmd, TRUE);


/*** D flipflop ***/


function tide_dff_chkasc (d, e, name, width, clk, q, clken, ac, as, sc, ss) {
   if ((ac != tideOptPinPol.None) && (as != tideOptPinPol.None))
      if ((vlPref.Val (tide_prefs, "vstyle")) == tideVStyle.XBLOX)
         tideArgError ("XBLOX DFF may not have both async set and clear"); }


function tide_dff_func (d, e, name, width, clk, q, clken, ac, as, sc,
   ss, ischange) {
   tide_last_width = width;

   initcfg ();
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width - 1);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putreqcfg ("CLOCK", clk);
   putreqcfg ("Q", q);
   putoptcfg ("ENABLE", clken);
   putoptcfg ("ACLR", ac);
   putoptcfg ("ASET", as);
   putoptcfg ("SCLR", sc);
   putoptcfg ("SSET", ss);

   /* Synthesis bug #10818 does not allow internal signals as clock/clear */
   putstrcfg ("CLEVEL", (clk == tideReqPinPol.ActiveLow) ? "'0'" : "'1'");
   putstrcfg ("RLEVEL", (ac  == tideOptPinPol.ActiveLow) ? "'0'" : "'1'");
   putstrcfg ("SLEVEL", (as  == tideOptPinPol.ActiveLow) ? "'0'" : "'1'");

   /* Deal correctly with single bit inputs */
   if (width == 1) {
      putstrcfg ("DATANAME", "DATA0"); putstrcfg ("QNAME", "Q0"); }
   else {
      putstrcfg ("DATANAME", sprintf ("DATA[%d:0]", width - 1));
      putstrcfg ("QNAME", sprintf ("Q[%d:0]", width - 1)); }

   /* No way to behaviorally synthesize dff with preset and clear */
   if ((ac != tideOptPinPol.None) && (as != tideOptPinPol.None))
      putintcfg ("HAS_ASETANDCLR", 1);
   else putintcfg ("HAS_ASETANDCLR", 0);

   /* Compute 0,1,X values and produce model */
   putstrcfg ("B_0", string_fill (make_string (width), '0'));
   putstrcfg ("B_1", string_fill (make_string (width), '1'));
   putstrcfg ("B_X", string_fill (make_string (width), 'X'));
   finishcfg (d, e, "dff", "dff", name, ischange, width); }


tide_dff_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, clk, q, clken, ac, as, sc, ss)
      tide_dff_func (d, e, name, width, clk, q, clken, ac, as, sc, ss, FALSE);,
   "_tide_dff",  "D Flipflop",
   tide_symname ("dff"), tide_buswidth (), tide_reqpol ("Clock"),
   tide_reqpol ("Output"), tide_optpol ("Clock enable"),
   tide_optpol ("Async clear"), tide_optpol ("Async set"),
   tide_optpol ("Sync clear"), tide_optpol ("Sync set"));


tide_dff_chg := vcWv.Cmd (fac,
   func (d, e, name, width, clk, q, clken, ac, as, sc, ss)
      tide_dff_func (d, e, name, width, clk, q, clken, ac, as, sc, ss, TRUE);,
   "_tide_dff_chg",  "D Flipflop",
   tide_chgsymname (), tide_chgbuswidth (), tide_chgreqpol ("Clock", "CLOCK"),
   tide_chgreqpolwid ("Output", "Q"), tide_chgoptpol ("Clock enable", "ENABLE"),
   tide_chgoptpol ("Async clear", "ACLR"), tide_chgoptpol ("Async set", "ASET"),
   tide_chgoptpol ("Sync clear", "SCLR"), tide_chgoptpol ("Sync set", "SSET"));


vcCmd.SetHelp (tide_dff_cmd, "Create D flipflop symbol and model");
vcCmd.EnableApply (tide_dff_cmd, TRUE);
vcCmd.VerifyParams (tide_dff_cmd, tide_dff_chkasc);
vcCmd.VerifyParams (tide_dff_chg, tide_dff_chkasc);


/*** Gate ***/


function tide_gate_func (d, e, name, type, width, inputs, ischange) {
   tide_last_width = width; tide_last_input = inputs;
   initcfg ();
   putintcfg ("N_W", width);  putintcfg ("N_WL1", width-1);
   putintcfg ("N_S", inputs); putintcfg ("N_SL1", inputs-1);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "gatedel"));

   /* Determine type and output polarity from enumeration, then set them */
   if ((type == tideGate.And) || (type == tideGate.Nand)) gatetype := "and";
   if ((type == tideGate.Or)  || (type == tideGate.Nor))  gatetype := "or";
   if ((type == tideGate.Xor) || (type == tideGate.Xnor)) gatetype := "xor";
   if ((type == tideGate.And) || (type == tideGate.Or) ||
       (type == tideGate.Xor)) outpol := tideReqPinPol.ActiveHigh;
   else outpol := tideReqPinPol.ActiveLow;
   putstrcfg ("GATETYPE", gatetype);
   putreqcfg ("RESULT", outpol);

   /* Set all input pin polarities high */
   for (i:=0; i<inputs; i=i+1)
      putreqcfg (sprintf ("%d", i), tideReqPinPol.ActiveHigh);

   /* Suffix models with "b" if single bit wide */
   if (width == 1) finishcfg (d, e, sprintf ("%s%db", gatetype, inputs),
      "gateb", name, ischange, inputs);
   else finishcfg (d, e, sprintf ("%s%d", gatetype, inputs), "gate",
      name, ischange, inputs); }


tide_gate_cmd := vcWv.Cmd (fac,
   func (d, e, name, type, width, inputs)
      tide_gate_func (d, e, name, type, width, inputs, FALSE);,
   "_tide_gate", "Gate",
   tide_symname ("gate"),
   vcWv.Param ("Gate type", tideGateVT, vcWv.Sticky, tideGate.And,
      "Select the gate type"),
   tide_buswidth (), tide_inputwidth());


tide_gate_chg := vcWv.Cmd (fac,
   func (d, e, name, type, width, inputs)
      tide_gate_func (d, e, name, type, width, inputs, TRUE);,
   "_tide_gate_chg", "Gate",
   tide_chgsymname (),
   vcWv.Param ("Gate type", tideGateVT, vcWv.Dynamic, tide_gate_type,
      "Select the gate type"),
   tide_chgbuswidth (), tide_chginputwidth());


vcCmd.SetHelp (tide_gate_cmd, "Create AND,OR,XOR,etc symbol and model");
vcCmd.EnableApply (tide_gate_cmd, TRUE);


/*** Latch ***/


function tide_latch_func (d, e, name, width, gate, q, ac, as, ischange) {
   tide_last_width = width;

   initcfg ();
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width - 1);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putreqcfg ("GATE", gate);
   putreqcfg ("Q", q);
   putoptcfg ("ACLR", ac);
   putoptcfg ("ASET", as);
   putstrcfg ("B_0", string_fill (make_string (width), '0'));
   putstrcfg ("B_1", string_fill (make_string (width), '1'));

   /* Deal correctly with single bit inputs */
   if (width == 1) {
      putstrcfg ("DATANAME", "DATA0"); putstrcfg ("QNAME", "Q0"); }
   else {
      putstrcfg ("DATANAME", sprintf ("DATA[%d:0]", width - 1));
      putstrcfg ("QNAME", sprintf ("Q[%d:0]", width - 1)); }

   /* Produce model */
   finishcfg (d, e, "dlatch", "dlatch", name, ischange, width); }


tide_latch_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, gate, q, ac, as)
      tide_latch_func (d, e, name, width, gate, q, ac, as, FALSE);,
   "_tide_latch",  "D Latch",
   tide_symname ("dlatch"), tide_buswidth (), tide_reqpol ("Gate"),
   tide_reqpol ("Output"), tide_optpol ("Async clear"),
   tide_optpol ("Async set"));


tide_latch_chg := vcWv.Cmd (fac,
   func (d, e, name, width, gate, q, ac, as)
      tide_latch_func (d, e, name, width, gate, q, ac, as, TRUE);,
   "_tide_latch_chg",  "D Latch",
   tide_chgsymname (), tide_chgbuswidth (), tide_chgreqpol ("Gate", "GATE"),
   tide_chgreqpolwid ("Output", "Q"),
   tide_chgoptpol ("Async clear", "ACLR"),
   tide_chgoptpol ("Async set", "ASET"));


vcCmd.SetHelp (tide_latch_cmd,    "Create D latch symbol and model");
vcCmd.EnableApply (tide_latch_cmd, TRUE);


/*** Multiplier ***/


function tide_mult_func (d, e, name, awidth, bwidth, fmt, ischange) {
   width := awidth + bwidth;

   initcfg ();
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width - 1);
   putintcfg ("N_A", awidth);
   putintcfg ("N_AL1", awidth - 1);
   putintcfg ("N_AL2", awidth - 2);
   putintcfg ("N_B", bwidth);
   putintcfg ("N_BL1", bwidth - 1);
   putintcfg ("N_BL2", bwidth - 2);
   putintcfg ("N_DELAY", width * vlPref.Val (tide_prefs, "arithdel"));
   if (fmt == tideSigned.Unsigned) putstrcfg ("REPRESENTATION", "UNSIGNED");
   else putstrcfg ("REPRESENTATION", "SIGNED");

   finishcfg (d, e, "mult", "mult", name, ischange, width); }


tide_mult_cmd := vcWv.Cmd (fac,
   func (d, e, name, awidth, bwidth, fmt)
      tide_mult_func (d, e, name, awidth, bwidth, fmt, FALSE);,
   "_tide_mult",  "Multiply",
   tide_symname ("mult"),
   vcWv.Param ("A bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         prefval := vlPref.Val (tide_prefs, "arithwidth");
         if (tide_last_width < prefval) tide_last_width;
         else prefval; },
      vcScalebar.fac, "Enter the A bus width",
      min <- 2, max <- vlPref.Val (tide_prefs, "arithwidth")),
   vcWv.Param ("B bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (a) {
         prefval := vlPref.Val (tide_prefs, "arithwidth");
         if (tide_last_width < prefval) tide_last_width;
         else prefval; },
      vcScalebar.fac, "Enter the B bus width",
      min <- 2, max <- vlPref.Val (tide_prefs, "arithwidth")),
   tide_signed ());


tide_mult_chg := vcWv.Cmd (fac,
   func (d, e, name, awidth, bwidth, fmt)
      tide_mult_func (d, e, name, awidth, bwidth, fmt, TRUE);,
   "_tide_mult_chg",  "Multiply",
   tide_chgsymname (),
   vcWv.Param ("A bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "WIDTHA");,
      vcScalebar.fac, "Enter the A bus width",
      min <- 2, max <- vlPref.Val (tide_prefs, "arithwidth")),
   vcWv.Param ("B bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (d) tide_get_param (d, "WIDTHB");,
      vcScalebar.fac, "Enter the B bus width",
      min <- 2, max <- vlPref.Val (tide_prefs, "arithwidth")),
   tide_chgsigned ());


vcCmd.SetHelp (tide_mult_cmd, "Create multiplier symbol and model");
vcCmd.EnableApply (tide_mult_cmd, TRUE);


/*** Mux ***/


function tide_mux_func (d, e, name, width, inputs, ischange) {
   function putexpand (n) {
      if (n == 0) { putintcfg ("S_EXPAND", 0); putintcfg ("N_CL1", inputs-1); }
      else { putintcfg ("S_EXPAND", 1); putintcfg ("N_CL1", n); } }

   tide_last_width = width; tide_last_input = inputs;

   initcfg ();
   putintcfg ("N_W", inputs);  putintcfg ("N_WL1", inputs-1);
   putintcfg ("N_B", width);   putintcfg ("N_BL1", width-1);
   putlogcfg ("N_K", "N_KL1", inputs);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "gatedel"));

   /* Put marker for proper top level */
   if (inputs >   4)      putintcfg ("LEVEL_TOP", 2);
   else if (inputs >   2) putintcfg ("LEVEL_TOP", 1);
   else                   putintcfg ("LEVEL_TOP", 0);

   /* Cascade level values */
   putintcfg ("LEVEL_2", (inputs >   4) ?   7 : 0);
   putintcfg ("LEVEL_1", (inputs >   2) ?   3 : 0);

   /* Determine number of expansion bits */
   if      (inputs ==   2) putexpand   (0);
   else if (inputs ==   3) putexpand   (3);
   else if (inputs ==   4) putexpand   (0);
   else if (inputs  <   8) putexpand   (7);
   else if (inputs ==   8) putexpand   (0);

   /* Make data vector, or single bit; create symbol name */
   /* Viewcell doesn't allow "" so can't switch between "" and "_vector" */
   symname := sprintf ("mux%d", inputs);
   basebit := "vlbit";
   if (vlPref.Val (tide_prefs, "vstyle") == tideVStyle.Synopsys)
      basebit := "std_logic";
   if (width == 1) {
      symname = symname + "b";
      putstrcfg ("BASEBIT", basebit); }
   else putstrcfg ("BASEBIT", sprintf ("%s_vector (%d downto 0)",
      basebit, width-1));

   finishcfg (d, e, symname, "mux", name, ischange, inputs); }


tide_mux_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, inputs)
      tide_mux_func (d, e, name, width, inputs, FALSE);,
   "_tide_mux", "Multiplexer",
   tide_symname ("mux"), tide_buswidth (), tide_inputwidth ());


tide_mux_chg := vcWv.Cmd (fac,
   func (d, e, name, width, inputs)
      tide_mux_func (d, e, name, width, inputs, TRUE);,
   "_tide_mux_chg", "Mux",
   tide_chgsymname (), tide_chgbuswidth (), tide_chginputwidth ());


vcCmd.SetHelp (tide_mux_cmd, "Create multiplexer symbol and model");
vcCmd.EnableApply (tide_mux_cmd, TRUE);


/*** Ram D/Q ***/


function tide_ram_chkdepth (depth) {
   if (depth > 32)
      if ((vlPref.Val (tide_prefs, "vstyle")) != tideVStyle.XBLOX)
         tideArgError ("Maximum depth for non-XBLOX RAM is 32"); }
      

function tide_ramdq_func (d, e, name, width, depth, clocking, enable, ischg) {
   function putclocking (h,l,c) {
      putintcfg ("S_HILATCH", h); putintcfg ("S_LOLATCH", l);
      putintcfg ("S_CLOCK",  c); }

   tide_last_width = width; tide_last_depth = depth;
   initcfg ();

   /* Parameter values */
   putintcfg ("N_B", width);  putintcfg ("N_BL1", width-1);
   putintcfg ("N_W", depth);  putintcfg ("N_WL1", depth-1);
   putlogcfg ("N_K", "N_KL1", depth);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putoptcfg ("ENABLE", enable);

   /* Variables for decoder */
   putintcfg ("DIST_2",   (depth >   1) ? 1 : 0);
   putintcfg ("DIST_4",   (depth >   2) ? 1 : 0);
   putintcfg ("DIST_8",   (depth >   4) ? 1 : 0);
   putintcfg ("DIST_16",  (depth >   8) ? 1 : 0);
   putintcfg ("DIST_32",  (depth >  16) ? 1 : 0);

   if      (depth <=   2) ceiling :=   1;
   else if (depth <=   4) ceiling :=   3;
   else if (depth <=   8) ceiling :=   7;
   else if (depth <=  16) ceiling :=  15;
   else                   ceiling :=  31;
   putintcfg ("N_CEILING", ceiling);
   putlogcfg ("N_KC", "N_KCL1", ceiling);

   /* Clocking variables */
   if (clocking == tideRAMType.TransparentHigh) {
      putclocking (1,0,0);
      putstrcfg ("CLOCKNAME", "WE");
      putstrcfg ("LP_CLOCK", "NORMAL");
      putintcfg ("INV_CLOCK", 0); }
   else if (clocking == tideRAMType.TransparentLow) {
      putclocking (0,1,0);
      putstrcfg ("CLOCKNAME", "WE");
      putstrcfg ("LP_CLOCK", "INVERT");
      putintcfg ("INV_CLOCK", 1); }
   else {
      putstrcfg ("CLOCKNAME", "CLOCK");
      putintcfg ("INV_CLOCK", 0);
      putstrcfg ("LP_CLOCK", "NORMAL");
      putclocking (0,0,1); }

   /* Deal correctly with single bit inputs */
   if (width == 1) {
      putstrcfg ("DATANAME", "DATA0"); putstrcfg ("QNAME", "Q0"); }
   else {
      putstrcfg ("DATANAME", sprintf ("DATA[%d:0]", width - 1));
      putstrcfg ("QNAME", sprintf ("Q[%d:0]", width - 1)); }

   putstrcfg ("B_0", string_fill (make_string (width), '0'));
   finishcfg (d, e, "ramdq", "ramdq", name, ischg, width); }


tide_ramdq_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, depth, clocking, enable)
      tide_ramdq_func (d, e, name, width, depth, clocking, enable, FALSE);,
      "_tide_ramdq", "Ram with D,Q ports",
      tide_symname ("ramdq"), tide_buswidth (), tide_depth (),
   vcWv.Param ("Clocking scheme", tideRAMTypeVT, vcWv.Sticky,
      tideRAMType.TransparentHigh, "Select the clock scheme"),
   tide_optpol ("Mem enable"));


tide_ramdq_chg := vcWv.Cmd (fac,
   func (d, e, name, width, depth, clocking, enable)
      tide_ramdq_func (d, e, name, width, depth, clocking, enable, TRUE);,
      "_tide_ramdq_chg", "Ram with D,Q ports",
      tide_chgsymname (), tide_chgbuswidth (), tide_chgdepth (),
   vcWv.Param ("Clocking scheme", tideRAMTypeVT, vcWv.Dynamic,
      tide_ram_find_clock, "Select the clock scheme"),
   tide_chgoptpol ("Mem enable", "MEMENAB"));


vcCmd.SetHelp (tide_ramdq_cmd, "Create RAM with D and Q port symbol and model");
vcCmd.EnableApply (tide_ramdq_cmd, TRUE);
vcCmd.VerifyParams (tide_ramdq_cmd,
   func (d, e, n, w, depth, c, en) tide_ram_chkdepth (depth););
vcCmd.VerifyParams (tide_ramdq_chg,
   func (d, e, n, w, depth, c, en) tide_ram_chkdepth (depth););


/*** RAM IO ***/


function tide_ramio_func (d, e, name, width, depth, clocking, enable,
   outenab, ischange) {

   function putclocking (h,l,c) {
      putintcfg ("S_HILATCH", h); putintcfg ("S_LOLATCH", l);
      putintcfg ("S_CLOCK",  c); }

   tide_last_width = width; tide_last_depth = depth;
   initcfg ();

   /* Parameter values */
   putintcfg ("N_B", width);  putintcfg ("N_BL1", width-1);
   putintcfg ("N_W", depth);  putintcfg ("N_WL1", depth-1);
   putlogcfg ("N_K", "N_KL1", depth);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putoptcfg ("ENABLE", enable); putreqcfg ("OUTENAB", outenab);

   /* Variables for decoder */
   putintcfg ("DIST_2",   (depth >   1) ? 1 : 0);
   putintcfg ("DIST_4",   (depth >   2) ? 1 : 0);
   putintcfg ("DIST_8",   (depth >   4) ? 1 : 0);
   putintcfg ("DIST_16",  (depth >   8) ? 1 : 0);
   putintcfg ("DIST_32",  (depth >  16) ? 1 : 0);

   if      (depth <=   2) ceiling :=   1;
   else if (depth <=   4) ceiling :=   3;
   else if (depth <=   8) ceiling :=   7;
   else if (depth <=  16) ceiling :=  15;
   else                   ceiling :=  31;
   putintcfg ("N_CEILING", ceiling);
   putlogcfg ("N_KC", "N_KCL1", ceiling);

   /* Clocking variables */
   if (clocking == tideRAMType.TransparentHigh) {
      putclocking (1,0,0);
      putstrcfg ("CLOCKNAME", "WE");
      putstrcfg ("LP_CLOCK", "NORMAL");
      putintcfg ("INV_CLOCK", 0); }
   else if (clocking == tideRAMType.TransparentLow) {
      putclocking (0,1,0);
      putstrcfg ("CLOCKNAME", "WE");
      putstrcfg ("LP_CLOCK", "INVERT");
      putintcfg ("INV_CLOCK", 1); }
   else {
      putstrcfg ("CLOCKNAME", "CLOCK");
      putstrcfg ("LP_CLOCK", "NORMAL");
      putintcfg ("INV_CLOCK", 0);
      putclocking (0,0,1); }

   /* Deal correctly with single bit inputs */
   if (width == 1) putstrcfg ("DIONAME", "DIO0");
   else putstrcfg ("DIONAME", sprintf ("DIO[%d:0]", width - 1));

   putstrcfg ("B_0", string_fill (make_string (width), '0'));
   putstrcfg ("B_Z", string_fill (make_string (width), 'Z'));
   finishcfg (d, e, "ramio", "ramio", name, ischange, width); }


tide_ramio_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, depth, clocking, enable, outenab)
      tide_ramio_func (d,e,name,width,depth,clocking,enable,outenab,FALSE);,
   "_tide_ramio", "Ram with IO port",
   tide_symname ("ramio"), tide_buswidth (), tide_depth (),
   vcWv.Param ("Clocking scheme", tideRAMTypeVT, vcWv.Sticky,
      tideRAMType.TransparentHigh, "Select the clock scheme"),
   tide_optpol ("Mem enable"), tide_reqpol ("Output enable"));


tide_ramio_chg := vcWv.Cmd (fac,
   func (d, e, name, width, depth, clocking, enable, outenab)
      tide_ramio_func (d,e,name,width,depth,clocking,enable,outenab,TRUE);,
   "_tide_ramio_chg", "Ram with IO port",
   tide_chgsymname (), tide_chgbuswidth (), tide_chgdepth (),
   vcWv.Param ("Clocking scheme", tideRAMTypeVT, vcWv.Dynamic,
      tide_ram_find_clock, "Select the clock scheme"),
   tide_chgoptpol ("Mem enable", "MEMENAB"),
   tide_chgreqpol ("Output enable", "OUTENAB"));


vcCmd.SetHelp (tide_ramio_cmd, "Create RAM with IO port symbol and model");
vcCmd.EnableApply (tide_ramio_cmd, TRUE);
vcCmd.VerifyParams (tide_ramio_cmd,
   func (d, e, n, w, depth, c, en, o) tide_ram_chkdepth (depth););
vcCmd.VerifyParams (tide_ramio_chg,
   func (d, e, n, w, depth, c, en, o) tide_ram_chkdepth (depth););


/*** Shift register ***/


function tide_shiftreg_func (d,e,name,width,clk,shiften,ac,din,dout,ischg) {
   function putpaironoff (on_name, off_name, pin_name, width) {
      putoptcfg (on_name,  tideOptPinPol.ActiveHigh);
      putoptcfg (off_name, tideOptPinPol.None);
      if (width == 0) putstrcfg (pin_name, on_name);
      else {
         full_name := sprintf ("%s[%d:0]", on_name, width - 1);
         putstrcfg (pin_name, full_name); } }

   tide_last_width = width;

   initcfg ();
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width - 1);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putreqcfg ("CLOCK", clk);
   putoptcfg ("SHIFTEN", shiften);
   putoptcfg ("ACLR", ac);
   putstrcfg ("B_0", string_fill (make_string (width), '0'));

   if (din == tideShiftIO.Serial) putpaironoff ("SHIFTIN", "DATA",
      "INPUTNAME", 0);
   else putpaironoff ("DATA", "SHIFTIN", "INPUTNAME", width);
   if (dout == tideShiftIO.Serial) putpaironoff ("SHIFTOUT", "Q",
      "OUTPUTNAME", 0);
   else putpaironoff ("Q", "SHIFTOUT", "OUTPUTNAME", width);
   finishcfg (d, e, "shiftreg", "shiftreg", name, ischg, width); }


tide_shiftreg_cmd := vcWv.Cmd (fac,
    func (d,e,name,width,clk,shiften,ac,din,dout)
       tide_shiftreg_func (d,e,name,width,clk,shiften,ac,din,dout,FALSE);,
   "_tide_shiftreg", "Shift register",
   tide_symname ("shreg"), tide_buswidth (), tide_reqpol ("Clock"),
   tide_optpol ("Shift enable"), tide_optpol ("Async clear"), 
   vcWv.Param ("Data in", tideShiftIOVT, vcWv.Sticky, tideShiftIO.Serial,
      "Select the input format"),
   vcWv.Param ("Data out", tideShiftIOVT, vcWv.Sticky, tideShiftIO.Parallel,
      "Select the output format"));


tide_shiftreg_chg := vcWv.Cmd (fac,
    func (d,e,name,width,clk,shiften,ac,din,dout)
       tide_shiftreg_func (d,e,name,width,clk,shiften,ac,din,dout,TRUE);,
   "_tide_shiftreg_chg", "Shift register",
   tide_chgsymname (), tide_chgbuswidth (), tide_chgreqpol ("Clock", "CLOCK"),
   tide_chgoptpol ("Shift enable", "SHIFTEN"),
   tide_chgoptpol ("Async clear", "ACLR"), 
   vcWv.Param ("Data in", tideShiftIOVT, vcWv.Dynamic, 
      func (d) tide_get_serpar (d, "SHIFTIN");, "Select the input format"),
   vcWv.Param ("Data out", tideShiftIOVT, vcWv.Dynamic,
      func (d) tide_get_serpar (d, "SHIFTOUT");, "Select the output format"));


vcCmd.SetHelp (tide_shiftreg_cmd, "Create shift register symbol and model");
vcCmd.EnableApply (tide_shiftreg_cmd, TRUE);


/*** TFF ***/


function tide_tff_func (d, e, name, width, clk, q, clken, ac, as, sc,
   ss, ischange) {
   tide_last_width = width;

   initcfg ();
   putintcfg ("N_W", width);
   putintcfg ("N_WL1", width - 1);
   putintcfg ("N_WL2", width - 2);
   putintcfg ("N_DELAY", vlPref.Val (tide_prefs, "c2qdel"));
   putreqcfg ("CLOCK", clk);
   putreqcfg ("Q", q);
   putoptcfg ("ENABLE", clken);
   putoptcfg ("ACLR", ac);
   putoptcfg ("ASET", as);
   putoptcfg ("SCLR", sc);
   putoptcfg ("SSET", ss);

   /* Synthesis bug #10818 does not allow internal signals as clock/clear */
   putstrcfg ("CLEVEL", (clk == tideReqPinPol.ActiveLow) ? "'0'" : "'1'");
   putstrcfg ("RLEVEL", (ac  == tideOptPinPol.ActiveLow) ? "'0'" : "'1'");
   putstrcfg ("SLEVEL", (as  == tideOptPinPol.ActiveLow) ? "'0'" : "'1'");

   /* No way to behaviorally synthesize tff with preset and clear */
   if ((ac != tideOptPinPol.None) && (as != tideOptPinPol.None))
      putintcfg ("HAS_ASETANDCLR", 1);
   else putintcfg ("HAS_ASETANDCLR", 0);

   /* Same as above, setting flag if both sync set and clr are used */
   if ((sc != tideOptPinPol.None) && (ss != tideOptPinPol.None))
      putintcfg ("HAS_SSETANDCLR", 1);
   else putintcfg ("HAS_SSETANDCLR", 0);

   /* Deal correctly with single bit inputs */
   if (width == 1) {
      putstrcfg ("DATANAME", "DATA0"); putstrcfg ("QNAME", "Q0"); }
   else {
      putstrcfg ("DATANAME", sprintf ("DATA[%d:0]", width - 1));
      putstrcfg ("QNAME", sprintf ("Q[%d:0]", width - 1)); }

   /* Compute 0,1,X values and produce model */
   putstrcfg ("B_0", string_fill (make_string (width), '0'));
   putstrcfg ("B_1", string_fill (make_string (width), '1'));
   putstrcfg ("B_X", string_fill (make_string (width), 'X'));
   finishcfg (d, e, "tff", "tff", name, ischange, width); }


tide_tff_cmd := vcWv.Cmd (fac,
   func (d, e, name, width, clk, q, clken, ac, as, sc, ss)
      tide_tff_func (d, e, name, width, clk, q, clken, ac, as, sc, ss, FALSE);,
   "_tide_tff",  "T Flipflop",
   tide_symname ("tff"), tide_buswidth (), tide_reqpol ("Clock"),
   tide_reqpol ("Output"), tide_optpol ("Clock enable"),
   tide_optpol ("Async clear"), tide_optpol ("Async set"),
   tide_optpol ("Sync clear"), tide_optpol ("Sync set"));


tide_tff_chg := vcWv.Cmd (fac,
   func (d, e, name, width, clk, q, clken, ac, as, sc, ss)
      tide_tff_func (d, e, name, width, clk, q, clken, ac, as, sc, ss, TRUE);,
   "_tide_tff_chg",  "T Flipflop",
   tide_chgsymname (), tide_chgbuswidth (), tide_chgreqpol ("Clock", "CLOCK"),
   tide_chgreqpolwid ("Output", "Q"), tide_chgoptpol ("Clock enable", "ENABLE"),
   tide_chgoptpol ("Async clear", "ACLR"), tide_chgoptpol ("Async set", "ASET"),
   tide_chgoptpol ("Sync clear", "SCLR"), tide_chgoptpol ("Sync set", "SSET"));


vcCmd.SetHelp (tide_tff_cmd, "Create T flipflop symbol and model");
vcCmd.EnableApply (tide_tff_cmd, TRUE);


/*** Tristate ***/


function tide_tristate_func (d, e, name, width, enable, output, ischange) {
   tide_last_width = width;

   initcfg ();

   putreqcfg ("ENABLE", enable);
   putreqcfg ("RESULT", output);
   putintcfg ("N_WL1", width-1);
   putintcfg ("N_W", width);
   putintcfg ("N_DELAY", vlPref.Val(tide_prefs, "gatedel"));

   /* Compute X value with length W for X assignment  */
   putstrcfg ("X_W", string_fill (make_string (width), 'X'));

   /* Compute Z value with length W for Z assignment */
   putstrcfg ("Z_W", string_fill (make_string (width), 'Z'));

   /* Deal correctly with single bit inputs */
   if (width == 1) {
      putstrcfg ("RESULTNAME", "RESULT0");
      putstrcfg ("TRIDATANAME", "TRIDATA0"); }
   else {
      putstrcfg ("RESULTNAME", sprintf ("RESULT[%d:0]", width - 1));
      putstrcfg ("TRIDATANAME", sprintf ("TRIDATA[%d:0]", width - 1)); }

   finishcfg (d, e, "tristate","tristate", name, ischange, width); }  


tide_tristate_cmd := vcWv.Cmd (fac, 
   func(d, e, name, width, enable, output) 
      tide_tristate_func(d, e, name, width, enable, output,FALSE);,
   "_tide_tristate", "Tristate",
   tide_symname ("tri"), tide_buswidth (), tide_reqpol ("Enable"),
   tide_reqpol ("Output"));
vcCmd.EnableApply (tide_tristate_cmd, TRUE); 


tide_tristate_chg := vcWv.Cmd (fac,
   func(d, e, name, width, enable, output)
      tide_tristate_func(d, e, name, width, enable, output, TRUE);, 
   "_tide_tristate_chg", "Tristate",
   tide_chgsymname (), tide_chgbuswidth (),
   tide_chgreqpol ("Enable", "ENABLETR"),
   tide_chgreqpolwid ("Output", "RESULT"));


vcCmd.SetHelp (tide_tristate_cmd,
   "Create tristate symbol and simulation model");


/*** Add => Cell command ***/


function tide_add_func (d, e) {
   if (tide_cmd_popup == nil) {
      table := vcTable.Create ("", 2, 1);
      picklist := vcRegion.Create ("", vcPicklist.fac, list (
         list (list ("Gate", tide_gate_cmd),
               list ("Tristate", tide_tristate_cmd),
               list ("Multiplexer", tide_mux_cmd),
               list ("Decode", tide_decode_cmd)),
         list (list ("Add, add/sub", tide_add_sub_cmd),
               list ("Compare", tide_compare_cmd),
               list ("Multiply", tide_mult_cmd),
               list ("Comb shift", tide_cshift_cmd)),
         list (list ("D flipflop", tide_dff_cmd),
               list ("D latch", tide_latch_cmd),
               list ("T flipflop", tide_tff_cmd),
               list ("Shift register", tide_shiftreg_cmd)),
         list (list ("Up, u/d counter", tide_count_cmd),
               list ("RAM with D/Q", tide_ramdq_cmd),
               list ("RAM with IO", tide_ramio_cmd))));

      vcTable.Add (table, picklist);
      vcPicklist.Notify (picklist, func (data, onoff)
         if (onoff) vcCmd.Run (car (data), d, e, nil, FALSE););

      buttons := vcRegion.Create ("", vcButnRow.fac, list (
         list ("Dismiss", "Dismiss", func ();, list (close_window <- TRUE))));
      vcTable.Add (table, buttons);
      tide_cmd_popup = vcWin.Popup (vcEvent.Window (e), table,
         vcButnRow.GetDefault (buttons),
         vcWin.NoCmdline+vcWin.NoStatus+vcWin.NoBindings);
      vcWin.SetTitle (tide_cmd_popup, "Add Cell");
      vcWin.DestroyNotify (tide_cmd_popup, func (w) tide_cmd_popup = nil;); }
   else vcWin.Show (tide_cmd_popup); }


tide_add_cmd := vcWv.Cmd (fac, tide_add_func, "addcell", "Add Cell");


vcMenu.AppendEntry (add_menu, "Cell...", tide_add_cmd);


/*** Change => Cell command ***/


function tide_change_func (d, e) {
   /* 51: just throw alert */
   block := DisplayGetBlock (d);
   obj := BlockGetSelObj (block, vqFlag.Component);
   if (null (obj)) tideError (1005, "No symbol selected");
   symblock := CompGetBlock (block, obj);
   type_attr := ObjFindAttr (block, symblock, "LPM_TYPE");
   if (null (type_attr)) tideError (1006, 
      "No LPM_TYPE= attribute on selected symbol");
   str := AttrGetValue (block, type_attr);

   /* Some extra work is required to distinguish between DFF and shift reg.
      It is shift reg if it has SHIFTEN, SHIFTIN or SHIFTOUT pins. */
   if (string_eq (str, "DFF")) { declare pin;
      pin = BlockFindPinByName (symblock, "SHIFTEN");
      if (null (pin)) pin = BlockFindPinByName (symblock, "SHIFTIN");
      if (null (pin)) pin = BlockFindPinByName (symblock, "SHIFTOUT");
      if (!null (pin)) str = "SHREG"; }

   /* Go from string to name of command */
   cmdpair := assoc (str,
      list (cons ("LPM_ADD_SUB", tide_add_sub_chg),
            cons ("LPM_AND",     tide_gate_chg),
            cons ("LPM_BUSTRI",  tide_tristate_chg),
            cons ("LPM_COUNTER", tide_count_chg),
            cons ("LPM_COMPARE", tide_compare_chg),
            cons ("LPM_CLSHIFT", tide_cshift_chg),
            cons ("LPM_DECODE",  tide_decode_chg),
            cons ("LPM_DFF",     tide_dff_chg),
            cons ("LPM_MUX",     tide_mux_chg),
            cons ("LPM_LATCH",   tide_latch_chg),
            cons ("LPM_MULT",    tide_mult_chg),
            cons ("LPM_OR",      tide_gate_chg),
            cons ("LPM_SHREG",   tide_shiftreg_chg),
            cons ("LPM_RAM_DQ",  tide_ramdq_chg),
            cons ("LPM_RAM_IO",  tide_ramdq_chg),
            cons ("LPM_XOR",     tide_gate_chg)));
   if (cmdpair) vcCmd.Run (cdr (cmdpair), d, e, nil, TRUE);

   /* Error if value of LPM_TYPE attribute not in above list */
   else tideError (1007,
      sprintf ("Invalid value '%s' for LPM_TYPE on selected symbol", str)); }


tide_change_cmd := vcWv.Cmd (fac, tide_change_func, "chgcell", "_tide_chg");
vcCmd.SetHelp (tide_change_cmd, "Change selected cell");
vcMenu.AppendEntry (change_menu, "Cell", tide_change_cmd);


/*** Prop => Cell command ***/


function tide_prop_func (d, e, vstyle, logcell, autoname, gatedel, c2qdel,
   arithdel, buswidth, arithwidth, inputwidth, depth) {

   vlPref.Setval (tide_prefs, "vstyle",     vstyle);
   vlPref.Setval (tide_prefs, "logcell",    logcell);
   vlPref.Setval (tide_prefs, "autoname",   autoname);
   vlPref.Setval (tide_prefs, "gatedel",    gatedel);
   vlPref.Setval (tide_prefs, "c2qdel",     c2qdel);
   vlPref.Setval (tide_prefs, "arithdel",   arithdel);
   vlPref.Setval (tide_prefs, "buswidth",   buswidth);
   vlPref.Setval (tide_prefs, "arithwidth", arithwidth);
   vlPref.Setval (tide_prefs, "inwidth",    inputwidth);
   vlPref.Setval (tide_prefs, "depth",      depth);

   vlPref.WriteIni (tide_prefs, "vdpath.ini", TRUE); }


tide_prop_cmd := vcWv.Cmd (fac, tide_prop_func, "_tide_prop", "Cell Properties",
   vcWv.Param ("VHDL Style", tideVStyleVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "vstyle");,
      "Select the synthesis style for generated VHDL"),
   vcWv.Param ("Show generator messages", vlYesNoVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "logcell");,
      "Capture generator messages in log window"),
   vcWv.Param ("Append width to symbol name", vlYesNoVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "autoname");,
      "Create automatic name from your prefix and bus width"),
   vcWv.Param ("Generic gate delay (ns)", vlIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "gatedel");,
      "Enter generic gate delay value for simulation models"),
   vcWv.Param ("Generic clock to Q delay (ns)", vlIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "c2qdel");,
      "Enter generic clock to Q delay value for simulation models"),
   vcWv.Param ("Generic arithmetic bit delay (ns)", vlIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "arithdel");,
      "Enter generic arithmetic bit delay value for simulation models"),
   vcWv.Param ("Maximum bus width", vlPositiveIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "buswidth");,
      "Enter maximum value for bus width sliders"),
   vcWv.Param ("Maximum arithmetic width", vlPositiveIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "arithwidth");,
      "Enter maximum value for arithmetic width sliders"),
   vcWv.Param ("Maximum number of inputs", vlPositiveIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "inwidth");,
      "Enter maximum number of combinational gate inputs"),
   vcWv.Param ("Maximum number of RAM words", vlPositiveIntVT, vcWv.Dynamic,
      func (a) vlPref.Val (tide_prefs, "depth");,
      "Enter maximum number of words for RAMs"));


vcMenu.AppendEntry (prop_menu, "Cell", tide_prop_cmd);
