-- Package to define alu procedures.
-- procedures are used by components generated by ViewArchitect
--      typical use:
--              LIBRARY alucons;
--              USE alucons.alu_procs.ALL;
--              entity add_10 is
--                port (
--                   A , B : in    std_ulogic_vector ( 9 downto 0 ) ;
--                   CI : in    std_ulogic ;
--                   S : out    std_ulogic_vector ( 9 downto 0 ) ;
--                   CO : out    std_ulogic
--                ) ;
--              end add_10 ;
--              architecture behv of add_10 is
--              begin
--                vgn_add (a,b,ci,s,co) ;
--              end behv ;
--
-- procedures may be replaced with technology specific VHDL netlist
--      by AluCons
-- procedures not so replaced are synthesizable by ViewSynth
--------------------
-- Created 10/27/94     - jrs
--------------------

PACKAGE alu_procs IS

  procedure vgn_add (
    signal ain, bin : in    vlbit_vector;
    signal cin      : in    vlbit;
    signal sum      : out   vlbit_vector;
    signal cout     : out   vlbit );

  procedure vgn_inc (
    signal a  : in  vlbit_vector;
    signal s  : out vlbit_vector;
    signal co : out vlbit );

  procedure vgn_dec (
    signal a  : in  vlbit_vector;
    signal s  : out vlbit_vector;
    signal nbo : out vlbit );

  procedure vgn_eq (
    signal a, b         : in    vlbit_vector;
    signal eq           : out   vlbit );

  procedure vgn_gt (
    signal a, b         : in    vlbit_vector;
    signal agb          : out   vlbit );

  procedure vgn_ge (
    signal a, b         : in    vlbit_vector;
    signal ageb          : out   vlbit );

  procedure vgn_alu (
    signal a, b         : in    vlbit_vector;
    signal ci, add      : in    vlbit;
    signal s            : out   vlbit_vector;
    signal co           : out   vlbit );

  procedure vgn_sub (        -- nbi/nbo == !borrow
    signal a, b         : in    vlbit_vector;
    signal nbi          : in    vlbit;
    signal s            : out   vlbit_vector;
    signal nbo          : out   vlbit ) ;

  procedure vw_sub (        -- nbi/nbo == !borrow
    signal a, b         : in    vlbit_vector;
    signal s            : out   vlbit_vector;
    signal nbo          : out   vlbit ) ;

  procedure vgn_bisdee (
    signal a, b         : in    vlbit_vector;
    signal ci, minus      : in    vlbit;
    signal s            : out   vlbit_vector;
    signal co           : out   vlbit );

  procedure vgn_cnt (
    signal clk, rst, ld, en, up : in vlbit;
    signal d : in vlbit_vector;
    signal q : inout vlbit_vector );

end alu_procs;

package body alu_procs is

---------------------------------------------------------
   procedure vgn_cnt (
     signal clk, rst, ld, en, up : in vlbit;
     signal d : in vlbit_vector;
     signal q : inout vlbit_vector )
   is

   variable qplus : vlbit_vector (1+q'left downto 0) ;
   begin
    loop
    wait until ((clk'event and clk = '1') OR (rst = '1'));
    if (rst = '1') then
            q <= extendum("0",1+q'left);
--    elsif (clk'event and clk = '1') then
    else
      qplus := extendum("0",2+q'left);
      if (ld = '1') then
            q <= d;
      elsif (en = '1') then
         if (up = '1') then
              qplus := addum(q,"01");
              q <= qplus(q'left downto 0);
         else
              q <= q-1;
         end if;
      end if;
    end if;
   end loop;
   end;

---------------------------------------------------------
  procedure vgn_add (
    signal ain, bin : in    vlbit_vector;
    signal cin      : in    vlbit;
    signal sum      : out   vlbit_vector;
    signal cout     : out   vlbit )
  is

  variable rip :    vlbit := '0' ;

  begin
    rip := cin;
    for i in 0 to ain'left loop
        sum(i) <= (ain(i) XOR bin(i)) XOR rip;
        rip    := (ain(i) AND bin(i)) OR (rip AND (ain(i) XOR bin(i)));
    end loop;
    cout <= rip;
  end;

---------------------------------------------------------
  procedure vgn_inc (
    signal a    : in    vlbit_vector;
    signal s    : out   vlbit_vector;
    signal co   : out   vlbit )
  is

  variable rip :    vlbit := '0' ;

  begin
    rip := '1';
    for i in 0 to a'left loop
        s(i) <= a(i) XOR rip;
        rip  := a(i) AND rip;
    end loop;
    co <= rip;
  end;

---------------------------------------------------------
  procedure vgn_dec (
    signal a    : in    vlbit_vector;
    signal s    : out   vlbit_vector;
    signal nbo  : out   vlbit )
  is

  variable rip :    vlbit := '0' ;

  begin
    rip := '0';
    for i in 0 to a'left loop
        s(i) <= a(i) XOR (NOT rip);
        rip  := a(i) OR rip;
    end loop;
    nbo <= rip;
  end;

---------------------------------------------------------
  procedure vgn_bisdee (        -- minus => ci/co == !borrow
    signal a, b         : in    vlbit_vector;
    signal ci, minus    : in    vlbit;
    signal s            : out   vlbit_vector;
    signal co           : out   vlbit )
  is

  variable rip :    vlbit := '0' ;
  variable nb  :    vlbit_vector (b'left downto 0) ;

  begin
    rip := ci;
    nb  := b;
    if (minus = '1') then
        nb := NOT b;
    end if;
    for i in 0 to a'left loop
        s(i) <= (a(i) XOR nb(i)) XOR rip;
        rip  := (a(i) AND nb(i)) OR (rip AND (a(i) XOR nb(i)));
    end loop;
    co <= rip;
  end;

---------------------------------------------------------
  procedure vgn_sub (        -- nbi/nbo == !borrow
    signal a, b         : in    vlbit_vector;
    signal nbi          : in    vlbit;
    signal s            : out   vlbit_vector;
    signal nbo          : out   vlbit )
  is

  variable rip :    vlbit := '0' ;
  variable nb  :    vlbit_vector (b'left downto 0) ;

  begin
    rip := nbi;
    nb  := NOT b;
    for i in 0 to a'left loop
        s(i) <= (a(i) XOR nb(i)) XOR rip;
        rip  := (a(i) AND nb(i)) OR (rip AND (a(i) XOR nb(i)));
    end loop;
    nbo <= rip;
  end;

---------------------------------------------------------
  procedure vw_sub (        -- nbo == !borrow out
    signal a, b         : in    vlbit_vector;
    signal s            : out   vlbit_vector;
    signal nbo          : out   vlbit )
  is

  variable rip :    vlbit := '0' ;
  variable nb  :    vlbit_vector (b'left downto 0) ;

  begin
    rip := '1';
    nb  := NOT b;
    for i in 0 to a'left loop
        s(i) <= (a(i) XOR nb(i)) XOR rip;
        rip  := (a(i) AND nb(i)) OR (rip AND (a(i) XOR nb(i)));
    end loop;
    nbo <= rip;
  end;

---------------------------------------------------------
  procedure vgn_alu (
    signal a, b         : in    vlbit_vector;
    signal ci, add      : in    vlbit;
    signal s            : out   vlbit_vector;
    signal co           : out   vlbit )
  is

  variable rip :    vlbit := '0' ;
  variable nb  :    vlbit_vector (b'left downto 0) ;

  begin
    rip := ci;
    nb  := b;
    if (add = '0') then
        nb := NOT b;
        rip := NOT rip;
    end if;
    for i in 0 to a'left loop
        s(i) <= (a(i) XOR nb(i)) XOR rip;
        rip  := (a(i) AND nb(i)) OR (rip AND (a(i) XOR nb(i)));
    end loop;
    if (add = '0') then
        rip := NOT rip;
    end if;
    co <= rip;
  end;

---------------------------------------------------------
  procedure vgn_eq (
    signal a, b         : in    vlbit_vector;
    signal eq           : out   vlbit )
  is

  begin
    eq <= (a = b);
  end;

---------------------------------------------------------
  procedure vgn_gt (
    signal a, b : in    vlbit_vector;
    signal agb  : out   vlbit )
  is

  begin
    agb <= (a > b) ;
  end;

---------------------------------------------------------
  procedure vgn_ge (
    signal a, b : in    vlbit_vector;
    signal ageb  : out   vlbit )
  is

  begin
    ageb <= (a >= b) ;
  end;

---------------------------------------------------------

end alu_procs;

