-- ***************************************************************
-- Viewlogic has created the following VHDL models to provide
-- practical VHDL source code examples. The VHDL models contained
-- in this listing may be: used, displayed, copied, modified and 
-- have modified, prepared and have prepared derivative works of,
-- operated, processed and distributed without license from or
-- royalty to Viewlogic. 
--
-- Viewlogic by supporting the use of the VHDL standard 
-- is promoting the utilization of VHDL for library
-- building and hardware design. As design size/complexity grow,
-- standards such as VHDL will become increasing important
-- to allow designers to develop at a high level of abstraction
-- and insure the transportability of their designs. Viewlogic,
-- in maintaining its technology leadership role in the EDA 
-- market, provides VHDL synthesis with VHDL Designer and VHDL
-- simulation through System VHDL.
-- ***************************************************************

-- ASIC support functions
--
-- Version 0.1   Revised by: Don Gay  On: 10/02/90
--                   This is the initial writing.
--
-- Version 0.2   Revised by: Don Gay  On:  1/23/91
--                   fixed X delay calculation
--
-- Version 0.3   Revised by: Don Gay  On:  2/10/91
--                   Fixed Pulse width error message
--
-- **********************************************************
-- Define the interface to these routines
-- **********************************************************
--
PACKAGE timing IS

-- Delay selection
--
--    Selects and returns the correct delay
function DELAY (
                  out_val  :  vlbit;
                  last_val :  vlbit;
                  TPHL     :  time;
                  TPLH     :  time;
                  TPHX     :  time;
                  TPLX     :  time)
               return time;


-- MAX TIME selection
--
--    Returns the largest of two time values
function MAXTIME (
                  time1    : time;
                  time2    : time)
               return time;


-- MIN TIME selection
--
--    Returns the smallest of two time values
function MINTIME (
                  time1    : time;
                  time2    : time)
               return time;


-- DELTA TIME selection
--
--    Returns a positive time value that is the difference between
--    the two time values input.
function DELTATIME (
                  time1    : time;
                  time2    : time)
               return time;


-- Setup Time Checking
--
--    Returns TRUE if setup time violation is detected; FALSE if no
--    violation is detected.
--
--    Called with Setup time, Reference time (Clock change time),
--    Data change time, and the setup being tested for the error report
function CHKSETUP (
                  setup    : time;
                  clk      : time;
                  data     : time;
                  setname  : string)
               return BOOLEAN;


-- Hold Time Checking
--
--    Returns TRUE if a hold time violation; FALSE if no
--    violation is detected.
--
--    Called with hold time, Reference time (Clock change time),
--    Data change time, and the hold being tested for the error report
function CHKHOLD (
                  hold     : time;
                  clk      : time;
                  data     : time;
                  holdname : string)
               return BOOLEAN;


-- Pulse Width Checking
--
--    Returns TRUE if a pulse width time violation is detected; FALSE if no
--    violation is detected.
--
--    Called with pulse width, First change time, Second change time
--    Data change time, and the signal name being tested for the error report
function CHKPULSEW (
                  width       : time;
                  edge1       : time;
                  edge2       : time;
                  pulsename   : string)
               return BOOLEAN;


-- Release time Checking
--
--    Returns TRUE if a release time violation is detected; FALSE if no
--    violation is detected.
--
--    Called with release time, Released time, Clock time,
--    and the signal name being tested for the error report
function CHKRELEASET (
                  Spec     : time;
                  edge    : time;
                  clock       : time;
                  valuename   : string)
               return BOOLEAN;


END timing;


-- **********************************************************
-- Define the functions
-- **********************************************************
--
PACKAGE BODY timing IS
--

-- Delay selection
--
--    Selects and returns the correct delay
function DELAY (
                  out_val  :  vlbit;
                  last_val :  vlbit;
                  TPHL     :  time;
                  TPLH     :  time;
                  TPHX     :  time;
                  TPLX     :  time)
   return time is
   
   begin

       case out_val is
         when '0' => return TPHL;
         when '1' => return TPLH;
         when others =>
            case last_val is
               when '0' => return TPLX;
               when '1' => return TPHX;
               -- When others returns shortest delay to X
               when others =>
                  if (TPLX > TPHX) then return TPHX;
                     else return TPLX;
                  end if;
            end case;
     end case;

end DELAY;


-- MAX TIME selection
--
--    Returns the largest of two time values
function MAXTIME (
                  time1    : time;
                  time2    : time)
   return time is

begin

   if (time1 > time2) then return time1;
      else return time2;
   end if;

end MAXTIME;


-- MIN TIME selection
--
--    Returns the smallest of two time values
function MINTIME (
                  time1    : time;
                  time2    : time)
   return time is

begin

   if (time1 > time2) then return time2;
      else return time1;
   end if;

end MINTIME;



-- DELTA TIME selection
--
--    Returns a positive time value that is the difference between
--    the two time values input.
function DELTATIME (
                  time1    : time;
                  time2    : time)
   return time is

begin

   if (time1 > time2) then return (time1 - time2);
      else return (time2 - time1);
   end if;

end DELTATIME;


-- Setup Time Checking
--
--    Returns TRUE if a setup time violation; FALSE if no
--    violation is detected.
--
--    Called with Setup time, Reference time (Clock change time),
--    Data change time, and the setup name for the error report
function CHKSETUP (
                  setup    : time;
                  clk      : time;
                  data     : time;
                  setname  : string)


   return BOOLEAN is

begin

   -- Suppress Errors at time 0
   if ((now <= 0 ns) or

   -- Check the setup time
   ( (data + setup) <= clk )) then return( false );
   else
      putline("Error: Setup time violation at time ",now);
          put("       For clk at ",clk);
      putline(" and Data at ",data);
          put("       Setup of ",(clk - data));
          put(" checked against ",setup);
      putline(" for ",setname);
      return( True );
   end if;

return( FALSE );

end CHKSETUP;


-- hold Time Checking
--
--    Returns TRUE if a hold time violation is detected; FALSE if no
--    violation is detected.
--
--    Called with hold time, Reference time (Clock change time),
--    Data change time, and the hold name for the error report
function CHKHOLD (
                  hold     : time;
                  clk      : time;
                  data     : time;
                  holdname : string)


   return BOOLEAN is

begin

   -- Suppress Errors at time 0
   if ((now <= 0 ns) or

   -- Check the hold time
   ( data >= (clk + hold))) then return( False );
   else
      putline("Error: Hold time violation at time ",now);
          put("       For clk at ",clk);
      putline(" and Data at ",data);
          put("       Hold of ",(data - clk));
          put(" checked against ",hold);
      putline(" for ",holdname);
      return( TRUE );
   end if;

return( FALSE );

end CHKhold;


-- Pulse Width Checking
--
--    Returns TRUE if a pulse width time violation; FALSE if no
--    violation is detected.
--
--    Called with pulse width, First change time, Second change time
--    Data change time, and the signal name being tested for the error report
function CHKPULSEW (
                  width       : time;
                  edge1       : time;
                  edge2       : time;
                  pulsename   : string)

   return BOOLEAN is

begin

   -- Suppress Errors at time 0
   if (now <= 0 ns) then return( False );
   end if;

   -- Check the pulse width
   if ( width <= (edge2 - edge1)) then return( False );
   else
      putline("Error: Pulse Width violation at time ",now);
      putline("       For signal: ",pulsename);
      return( TRUE );
   end if;

return( FALSE );

end CHKPULSEW;



-- Release time Checking
--
--    Returns TRUE if a release time violation is detected; FALSE if no
--    violation is detected.
--
--    Called with release time, Released time, Clock time,
--    and the signal name being tested for the error report
function CHKRELEASET (
                  Spec     : time;
                  edge    : time;
                  clock       : time;
                  valuename   : string)

   return BOOLEAN is

begin

   -- Suppress Errors at time 0
   if (now <= 0 ns) then return( False );

   -- Check the release time
   elsif ( (Spec + edge) < clock ) then return( False );
   else
      putline("Error: Release time violation at time ",now);
          put("       For signal: ",valuename);
      return( TRUE );
   end if;

return( FALSE );

end CHKRELEASET;

END timing;

