/* $Id: vhdlpush.vs,v 1.2 93/12/20 17:59:24 josh Stab $ */

/*
 * This module implements level-push-vhdl, looking at the Vantage
 * VHDL directory hiearchy to find the source code for a particular
 * architecture, entity, or configuration.  It uses the new vhdlDu
 * API to traverse the Vantage directories, which in theory isolates
 * this code from the details of how Vantage sets up their directories.
 * Of course, this code is dependent on the ability to get from a
 * library to an entity, a library to a configuration, an entity to
 * an architecture, etc.
 */

/*
 * This functions returns a 3 element list, corresponding to objects
 * for an entity, architecture, and configuration.  Any of these may
 * be nil.  The user of this function may only make use of one of
 * these, but once I've got one, it's easy to get them all, and it's
 * easy to return all of them in an array.
 *
 * vhdl_attr        is of the form [lib:]entity[(arch)]
 * vhdl_config_attr is of the form [lib:]configuration
 */
function vhdl_push_info(vhdl_attr, vhdl_config_attr) {
    entity := nil;
    arch := nil;
    cfg := nil;

    lib_name := nil;
    declare lib;

    /* If the VHDL attribute was specified, look for entity and arch */
    if (vhdl_attr != nil) {
        /* Pick out the library name if one is specified. */
        colon := lindex (vhdl_attr, ':');

        if (colon) {
            lib_name = substring (vhdl_attr, 0, colon);
            vhdl_attr = substring (vhdl_attr, colon + 1, 0);
        } /* if */
        lib = vhdlDu.FindLibrary ("vsslib.ini", lib_name);
        if (lib != nil) {
            paren := lindex (vhdl_attr, '(');
            if (paren) {
                entity = vhdlDu.FindChild (lib, substring (vhdl_attr, 0,
                                                           paren),
                                           vhdlDuType.Entity);
                if (entity != nil)
                    arch = vhdlDu.FindChild (entity,
                                             substring (vhdl_attr,
                                                        paren + 1, -1),
                                             vhdlDuType.Architecture);
            } /* if */

            /* If there is no architecture specified, find last compiled */ 
            else {
                entity = vhdlDu.FindChild (lib, vhdl_attr, vhdlDuType.Entity);
                if (entity != nil)
                    arch = vhdlDu.EntityLatestArch (entity);
            } /* else */
        } /* if */
    } /* if */

    /* Parse the configuration -- we can get entity and arch from that. */
    else if (vhdl_config_attr != nil) {
        /* Pick out the library name if one is specified. */
        colon := lindex (vhdl_config_attr, ':');
        if (colon) {
            lib_name = substring (vhdl_config_attr, 0, colon);
            vhdl_config_attr = substring (vhdl_config_attr, colon + 1, 0);
        } /* if */
        lib = vhdlDu.FindLibrary ("vsslib.ini", lib_name);
        if (lib != nil) {
            cfg = vhdlDu.FindChild (lib, vhdl_config_attr,
                                    vhdlDuType.Configuration);
            arch = vhdlDu.FindChild (cfg, "", vhdlDuType.Architecture);
            entity = vhdlDu.FindChild (cfg, "", vhdlDuType.Entity);
        } /* if */
    } /* else if */

    /* Return all 3 in an array */
    [entity, arch, cfg];
} /* function vhdl_push_info */

/*
 * Test to see whether a component with ACCEL= attribute has a Vantage
 * architecture or entity.  We can get these from the configuration
 * or from the entity(arch).
 */
function is_vhdl(display, event, attval, data, ext)
    (string_eqi (attval, "VANTAGE") &&
     ((FindAttr (data.block, data.comp, "VHDL") != nil) ||
      (FindAttr (data.block, data.comp, "VHDL_CONFIGURATION") != nil)));

/*
 * Test to see whether a component with ACCEL= attribute has a Vantage
 * configuration.
 */
function is_vhdl_cfg(display, event, attval, data, ext)
    (string_eqi (attval, "VANTAGE") &&
     (FindAttr (data.block, data.comp, "VHDL_CONFIGURATION") != nil));


/*
 * This is used for the VHD prompt to avoid showing it if it's a Vantage
 * component.
 */
function not_vantage(display, event, attval, data, ext) {
    accel := FindAttr (data.block, data.comp, "ACCEL");
    ((accel == nil) || !string_eqi (AttrGetValue (data.block, accel), "VANTAGE"));
} /* function not_vantage */


/*
 * This function-generator is used for doing a level-push-config, -entity,
 * and -arch.  The index n corresponds to the array returned from
 * vhdl_push_info.
 */
function show_func(n, type) {
    func(display, event, attr, attval, data, ext, editor) {
        function v(name) {
            attr := FindAttr (data.block, data.comp, name);
            (attr == nil) ? nil : AttrGetValue (data.block, attr);
        } /* function v */

        pi := vhdl_push_info (v ("VHDL"), v ("VHDL_CONFIGURATION"));
        if (pi[n] == nil)
            vcErr.ReportFromString (vcEvent.Window (event), nil,
                                    sprintf ("vdraw: Error: Unable to find VHDL %s for selected component\nPlease compile your model first",
                                             type));
        else {
            bufname := "VHDL " + convert (vlStringVT, vhdlDu.Type (pi[n]))
                + " " + vhdlDu.Name (pi[n]);
            vt := vcRegion.Create (bufname, vtx.fac, bufname,
                                   load_file <- FALSE);
            foreach (str; vhdlDu.Source (pi[n]))
                vtx.AppendString (vt, str);
            vtx.ReadOnly (vt, TRUE);
            vcWin.Create (vt);
        } /* else */
    };
} /* function show_func */

show_entity := show_func (0, "Entity");
show_arch := show_func (1, "Architecture");
show_cfg := show_func (2, "Configuration");
