/*
 *  $Revision: 1.16.5.1 $
 *
 *  Copyright (c) 1992 by Viewlogic Systems, Inc.
 */

/*
 * Here is a full fledged drawing editor written in Viewscript!
 * The data representation for the drawing is a list of objects.
 * Each object is a lexical closure which takes one flag,
 * which must be the symbol `draw or the symbol `mbb.
 *
 * This example makes heavy use of lexical closures.  Dive in and
 * have fun!  There is relatively little code here, but the file
 * is long due to excessive commenting.
 */

/*
 * We will build up the "Add" menu each time we define a new kind of
 * object.  The entry list is built up as a lisp list using cons.
 * The define_object call below is used to set up the user interface
 * for creating an object.  It adds to this list.
 */
add_menu_entries := nil;

/*
 * The representation of a drawing, in this implementation, is a
 * single cons cell whose car points to the object list, and whose
 * cdr points to the viewport.  The following functions are used to
 * access the object.  Note that rather than defining new functions,
 * we can just give a new name to an existing function.  Thus there
 * is no extra function-call overhead for the abstraction.
 */
make_drawing    := list;
object_list     := car;
canvas          := cadr;
mbb             := caddr;
set_object_list := set_car;
function set_mbb(drawing, new_mbb) set_car (cddr (drawing), new_mbb);

function viewport(objs_vp)
    vcVPCanvas.Viewport (canvas (objs_vp));

function add_object(objs_vp, new_obj) {
    set_object_list (objs_vp, cons (new_obj, object_list (objs_vp)));
    reset_canvas := FALSE;

    old_mbb := mbb (objs_vp);
    new_obj_mbb := new_obj (`mbb, viewport (objs_vp));
    if (old_mbb == nil) {
        set_mbb (objs_vp, new_obj_mbb);
        reset_canvas = TRUE;
    } /* if */
    else {
        new_mbb := vgRect.Create (0, 0, 0, 0);
        vgRect.Max (new_obj_mbb, old_mbb, new_mbb);
        if (!vgRect.Equal (old_mbb, new_mbb)) {
            set_mbb (objs_vp, new_mbb);
            reset_canvas = TRUE;
        } /* if */
    } /* else */

    if (reset_canvas)
        vcRegion.PrefSet (canvas (objs_vp),
                          list (full <- mbb (objs_vp)));
    new_obj (`draw, viewport (objs_vp));
} /* function add_object */

/*
 * Define an object type by creating a command to create that object
 * for the user, binding a key, and setting a menu to set the "mode".
 */
function define_object(key, class_name, p1func, p2func, create) {

    function create_cmd_func(objs_vp, event, args)
        add_object (objs_vp, apply (create, args));
        
    create_cmd := vcWv.Cmd (fac, create_cmd_func, class_name, class_name,
                            vcWv.Param ("args", vscmVT, vcWv.Constant, nil));

    /*
     * This function is called to initiate drawing an object.
     * It sets the current "mode" so that subsequent middle-mouse
     * hits run this function.  Then it calls vcDrag.create, which
     * is Core's built-in rubber-banding functionality, which can
     * draw circles, boxes, lines, and ellipses.  The "func" passed
     * is called after the rubber-banding is successfully completed
     * by the user.  Different objects (e.g. ellipses, lines, circles),
     * have their action functions called with different numbers of
     * arguments.  Hence the func (args...) in the vcDrag.create call
     * takes a variable number of arguments.  The create function, which
     * is passed into define_object above, is passed those arguments,
     * plus an additional argument, the viewport, tacked onto the beginning.
     * This enables the create function to translate from pixel coordinates
     * to user coordinates.
     */
    function drag(objs_vp, event) {
        set_mode (objs_vp, event);

        d := vcDrag.Create (event,
                       vcRegion.Area (vcEvent.Region (event)),
                       vgColor.White,
                       func (args...) {
                           vcWin.AddStatus (vcEvent.Window (event),
                                            "done.");
                           vcCmd.Run (create_cmd, objs_vp, event,
                                      list (args), TRUE);
                       },

                       /*
                        * The next argument to vcDrag.create directs it
                        * as to what kind of rubber-band object to draw.
                        * The Viewscript system can convert the string,
                        * "Ellipse", "Line", "Circle", or "Box", into
                        * the correct enum value to direct the dragging
                        * software to do the right thing.
                        */
                       class_name,
                       
                       (p1func == nil) ? 0 : p1func (event),
                       (p2func == nil) ? 0 : p2func (event),

                       func(event) {
                           vp := viewport (objs_vp);
                           x := vcEvent.X (event);
                           y := vcEvent.Y (event);
                           vgU.PixelToUser (vp, &x, &y);
                           x = round (x / 50) * 50;
                           y = round (y / 50) * 50;
                           vgU.UserToPixel (vp, &x, &y);
                           vcEvent.SetX (event, x);
                           vcEvent.SetY (event, y);
                       });
        vcDrag.SetViewport (d, viewport (objs_vp));
        vcWin.ClearStatus (vcEvent.Window (event));
        vcWin.AddStatus (vcEvent.Window (event), "Starting drawing...");
        vcDrag.SetCancel (d, func ()
                          vcWin.AddStatus (vcEvent.Window (event),
                                           "done."););
    } /* function drag */

    /*
     * Create a command object out of the drag function, and bind it to
     * the specified key.
     */
    drag_cmd := vcWv.Cmd (fac, drag);
    vcBind.Set (bindings, key, drag_cmd);

    /*
     * Define the set_mode function which allows the user to begin an
     * operation in the current mode by hitting the middle button.
     */
    function set_mode(objs_vp, event) {
        vcBind.Set (bindings, '\middledown', drag_cmd);
        vcWin.ClearStatus (vcEvent.Window (event));
        vcWin.AddStatus (vcEvent.Window (event),
                         sprintf ("%s", class_name) + " mode.");
    } /* function set_mode */

    set_mode_cmd := vcWv.Cmd (fac, set_mode);

    /*
     * Add in the menu entry.
     */
    add_menu_entries = cons (list (class_name, set_mode_cmd),
                             add_menu_entries);
} /* function define_object */

/*
 * Create a generic object that can receive two messages "draw" and "mbb".
 * The object is sent the message by calling it as a function.  A new
 * class of objects is invented by defining a create_<class> function that
 * calls create_object, passing in methods for drawing and finding the
 * minimum bounding box.  Other methods can be defined by adding new
 * arguments to the create_object routine and changing all the objects to
 * include the new method in their calls to create_object.
 */
function create_object(draw, mbb)
    func (msg, vp)
        if (msg == `draw)       draw (vp);
        else if (msg == `mbb)   mbb ();
    ;

function create_line(x1, y1, x2, y2)
    create_object (func (vp) vgU.Line (vp, x1, y1, x2, y2);,
                   func () vgRect.Create (x1, y1, x2, y2););
define_object ('\_l', "Line", nil, nil, create_line);

function create_circle(center_x, center_y, radius)
    create_object (func (vp) vgU.Circle (vp, center_x, center_y, radius);,
                   func () vgRect.Create (center_x - radius,
                                          center_y - radius,
                                          center_x + radius,
                                          center_y + radius););
define_object ('\_c', "Circle", nil, nil, create_circle);

function create_ellipse(cx, cy, rx, ry)
    create_object (func (vp) vgU.Ellipse (vp, cx, cy, rx, ry);,
                   func () vgRect.Create (cx - rx, cy - ry, cx + rx, cy +ry););
define_object ('\_e', "Ellipse", nil, nil, create_ellipse);

function create_box(rect)
    create_object (func (vp) vgU.Box (vp, rect);,
                   func () rect;);
define_object ('\_b', "Box", nil, nil, create_box);

function create_hbar(x1, x2)
    create_object (func (vp);,
                   func () vgRect.Create (x1, 0, x2, 0););
define_object ('\_h', "Hbar",
               func(event) vcEvent.Y (event) - 10;,
               func(event) vcEvent.Y (event) + 10;,
               create_hbar);

function create_vbar(y1, y2)
    create_object (func (vp);,
                   func () vgRect.Create (0, y1, 0, y2););
define_object ('\_h', "Vbar",
               func(event) vcEvent.X (event) - 10;,
               func(event) vcEvent.X (event) + 10;,
               create_vbar);

define_object ('\_m', "MBox",
               func(event) vcEvent.X (event) + 100;,
               func(event) vcEvent.Y (event) + 100;,
               create_box);

/*
 * The _Create call is the primary entry point for facilities.  In
 * this case, the drawing created as no objects and no viewport yet.
 * The objects are created manually by the user, and the viewport is
 * filled in at the first refresh event.
 */
function _Create(arg, region) {
    objs := nil;
    canvas := vcRegion.Create ("", vcVPCanvas.fac, region,
                               width <- 500,
                               full <- vgRect.Create (0, 0, 1000, 1000),
                               bounds <- vgRect.Create (-1000, -1000,
                                                        2000, 2000));
    vcRegion.FocusInhibit (canvas);
    objs_vp := make_drawing (objs, canvas, nil);
    scroll := vcVPCanvas.Scroll (vcRegion.Name (region), canvas,
                                 list (vertical <- 2,
                                       horizontal <- 2));
    vcRegion.AddChild (region, scroll);
    vcVPCanvas.RefreshProc (canvas, func(event) do_refresh (objs_vp, event););
    objs_vp;
} /* function create */
    
function do_refresh(objs_vp, event) {
    region := vcEvent.Region (event);
    vp := viewport (objs_vp);

/*    vgU.EraseRect (vp, vgU.User (vp)); */
/*    vgU.DrawGrid (vp, 50, 50, 0, 0, 0, 0); */

    /* 
     * Go through the object list and send a "draw" message to each
     * object.
     */
    foreach (obj; vscript.list_iter (object_list (objs_vp)))
        obj (`draw, vp);
} /* function refresh */

/*
 * For now, the "add" menu is the top level, because we don't have any
 * other operations defined.
 */
menus := apply (vcWv.MakeMenu, cons ("DrawMenu", cons (vcWin.RedSquareEntry,
                                                        add_menu_entries)));
vcFac.SetMenus (fac, menus);

/*
 * Set the size of the facility to something reasonable for a drawing editor.
 */
/*
function _Size(objs_vp, size, region, wv)
    vcSize.Set (size, 640, 480, 10, 20, 10000, 10000);
*/
