----------------------------------------------------------------------------
-- Behavioral Model for Simple 16-Bit ALU.                                  
--                                                                          
-- Inputs: A leg  (16 bits)                                                 
--         B leg  (16 bits)                                                 
--         Carry in (1 bit)                                                 
--         Result register clock                                            
--         Function Select (2 bits)                                         
--                                                                          
--             00 ==> ADD                                                   
--             01 ==> OR                                                    
--             10 ==> AND                                                   
--             11 ==> XOR                                                   
--                                                                          
--                                                                          
-- Outputs: Data out (16 bits, latched)                                     
--          Carry out (1 bit, unlatched)                                    
--                                                                          
----------------------------------------------------------------------------

----------------------------------------------------------------------------
-- Interface declaration:
----------------------------------------------------------------------------

entity alu is

-- Generic delays, with default values

   generic (cout_delay: time := 12300 PS;   -- Carry out delay
            reg_delay:  time := 4200 PS);   -- Register delay

-- IO ports:

   port (signal dout:       out vlbit_vector(0 to 15);
         signal cout:       out vlbit;
         signal func_sel:   in  vlbit_vector(0 to 1);
         signal a, b:       in  vlbit_vector(0 to 15);
         signal cin:        in  vlbit;
         signal clk:        in  vlbit);

end alu;

----------------------------------------------------------------------------
-- Architecture body:
----------------------------------------------------------------------------

architecture behavior of alu is

signal ALUout: vlbit_vector(0 to 16);     -- ALU output: cout & dout
signal ResReg: vlbit_vector(0 to 15);     -- Result register

   begin

   -------------------------------------------------------------------------
   -- Concurrent ALU process statement:
   -- When any input signal changes,
   -- compute new result.
   -------------------------------------------------------------------------

   alu_process: process(func_sel, a, b, cin)

   variable res_18:  vlbit_1d(-1 to 16);  -- 18-bit temporary result

   constant OPADD:   integer := 0;        -- OPcodes
   constant OPOR:    integer := 1;
   constant OPAND:   integer := 2;
   constant OPXOR:   integer := 3;

   constant XOUT:    vlbit_1d(0 to 16) := ('X',
                                           'X', 'X', 'X', 'X',
                                           'X', 'X', 'X', 'X',
                                           'X', 'X', 'X', 'X',
                                           'X', 'X', 'X', 'X');

      begin

      -- Compute the ALU output.    

      assert not vecunknown (func_sel)
         report "Unknown ALU function select line";

      case v1d2int (func_sel) is
         when OPADD  => res_18 := addum (addum (a, b), '0' & cin);
                        ALUout <= res_18(0 to 16);
         when OPOR   => ALUout <= '0' & (a OR b);           
         when OPAND  => ALUout <= '0' & (a AND b);          
         when OPXOR  => ALUout <= '0' & (a XOR b);          
         when OTHERS => ALUout <= XOUT;
                        putline ("Invalid ALU function select line");
      end case;
   end process;

   -------------------------------------------------------------------------
   -- Concurrent Register process statement:
   -- Load up the result register, but only on rising edge of clock. 
   -------------------------------------------------------------------------

   register_process: process
      begin
      wait until clk = '1';
      ResReg <= ALUout(1 to 16);
   end process;

   -------------------------------------------------------------------------
   -- Concurrent signal assignments:
   -- When ALU output or result register changes,
   -- schedule new values on data out pins. 
   -------------------------------------------------------------------------

   cout <= ALUout(0) after cout_delay;

   dout <= ResReg after reg_delay;

end behavior;
